#include "qpsd_utils.h"
#include <QApplication>

// Constructor

PSD_Parameters::PSD_Parameters(DvObject_var &obj, int direction_in) : init_ok(false),  phi2rad(1.), theta2rad(1.)
{
    tetra.clear();
    _PI = 4.*atan(1.0);
    stop = false;
	
	v_sign = (double)direction_in;
    
	_x = 0;
	_y = 0;
	_z = 0;
	_vel = 0;
	
	_phi_centre = 0;
	_phi_plus = 0;
	_phi_minus = 0;
	
	_theta_centre = 0;
	_theta_plus = 0;
	_theta_minus = 0;

   if(!init_data(obj)) return;
	
    init_ok=true;
}

// ------------------

double PSD_Parameters::getData(int time, int r, int phi, int theta)
{
    int index[3];
    index[indexE]=r;
    index[indexPhi]=phi;
    index[indexTheta]=theta;

    return data->dbl(time,index[0],index[1],index[2]);
}


// ------------------

bool PSD_Parameters::init_PSDcentre(DvObject_var &centreOfPSD)
{
	if(centreOfPSD.is_nil()) {
		vector <size_t> dims;
		dims.push_back(3);
		centre_PSD = new DvObject(0.0, dims, data->seqSize());
		return true;
	}
	
	// check target frame
	if(rotn_obj.is_nil() ){
		message = "No rotation matrix";
		return false;
	}
	
	// will test ToFrame
	if( !centreOfPSD->sameFrame(*rotn_obj)){
		message = "PSD bulk velocity must be in target frame";
		return false;
	}
	
	// get in correct units
	
	centre_PSD = centreOfPSD->changeUnitsTo("1.>m s^-1", "m s^-1"); // Work in SI units 

	if(centre_PSD->getError().contains("change Units") )
	{
		message = "cannot get units (SI_CONVERSION) for slice normal";
		return false;
	}

  
	if( !centre_PSD->isJoined(*data) )
	{
		if(centre_PSD->seqSize() == 1){
			// create sequence from single object
			centre_PSD = new DvObject(centre_PSD, data->seqSize());
		}
		else{
			centre_PSD =  centre_PSD->Join(data);

			if(centre_PSD.is_nil()) {
				message = "centre of PSD and Data join failed.";
				return false;
			}
		}	
	}

    return true;

}

// ------------------

bool PSD_Parameters::init_sliceNormal(DvObject_var &normal_in)
{
	if(normal_in.is_nil()) {
		message = "cannot get slice normal";
		return false;
	}
	
	// check target frame
	if(rotn_obj.is_nil() ){
		message = "No rotation matrix";
		return false;
	}
	DvString targetFrame = rotn_obj->getXrefText(TOFRAME);
	if( !normal_in->sameFrame(targetFrame)){
	
		// note, normal without frame returns true for sameFrame()
		message = "Normal vector for slice must be in target frame";
		return false;
	}
	
    if(normal_in.is_ok())
    {
		// This object is modified, so must use copy
		sliceNormal = new DvObject(normal_in);
		
		if(sliceNormal->seqSize() == 1){
			// create sequence from single object
			sliceNormal = new DvObject(sliceNormal, data->seqSize());
		}
		else{
		
	   		// join will make a copy. 
			sliceNormal =  sliceNormal->Join(data);

			if(sliceNormal.is_nil()) {
				message = "slice normal and Data join failed.";
			}
		}	
    }    

    return true;

}

bool PSD_Parameters::init_AdvsliceNormal(DvObject_var &Avec, DvObject_var &Bvec, int choice)
{
	if(Avec.is_nil() || Bvec.is_nil()) {
		message = "cannot get slice A and B vectors";
		return false;
	}
	
	// frame check done before calling this function so not needed here
	
	// Make copies of objects, join and calc normal, x, y unit vectors
	
    DvObject_var Avec_copy = new DvObject(Avec);
	DvObject_var Bvec_copy = new DvObject(Bvec);
	
	if(Avec_copy->seqSize() == 1){
		Avec_copy = new DvObject(Avec_copy, data->seqSize());
	}
	else{
		Avec_copy = Avec_copy->Join(data);
	}
	if(Bvec_copy->seqSize() == 1){
		Bvec_copy = new DvObject(Bvec_copy, data->seqSize());
	}
	else{
		Bvec_copy = Bvec_copy->Join(data);
	}
	if (Avec_copy.is_nil() || Bvec_copy.is_nil()){
		message = "Avec and/or Bvec failed to join to Data.";
		return false;
	}

	// now calc cross products and fill sliceNormal, sliceX, sliceY
	// choice=0 has A = normal; choice = 1 has n=AxB
	if (choice==0){  // normal ||A, Y || AxB, then X || Y x normal
		sliceNormal = new DvObject(Avec_copy);
		sliceY      = Avec_copy->vec(Bvec_copy);
		sliceX      = sliceY->vec(sliceNormal);
		if (sliceX->is_nil())
		{
			message="SliceY: sliceY x normal failed"; 
			return false;
		}
		if (sliceY->is_nil())
		{
			message="SliceY: Avec x Bvec failed"; 
			return false;
		}
	}
	else if (choice == 1) // normal || AxB, Y || normal x A, X || A
	{
		sliceNormal  = Avec_copy->vec(Bvec_copy);
		sliceX       = new DvObject(Avec_copy);
		sliceY       = sliceNormal->vec(sliceX);
	}
	if (sliceNormal->is_nil())
	{
		message="SliceNormal: Avec x Bvec failed"; 
		return false;
	}
	if (sliceY->is_nil())
	{
		message="SliceY: sliceNormal x sliceX failed"; 
		return false;
	}

	sliceNormal /= sliceNormal->abs();
	sliceX      /= sliceX->abs();
	sliceY      /= sliceY->abs();
	
    return true;
}

bool PSD_Parameters::init_PADfielddir(DvObject_var &field_in)
{
	if(field_in.is_nil()) {
		message = "cannot get field direction";
		return false;
	}
	
	// check target frame
	if(rotn_obj.is_nil() ){
		message = "No rotation matrix";
		return false;
	}
	DvString targetFrame = rotn_obj->getXrefText(TOFRAME);
	if( !field_in->sameFrame(targetFrame)){
	
		// note, normal without frame returns true for sameFrame()
		message = "Field vector must be in target science frame";
		return false;
	}
	
    if(field_in.is_ok())
    {
		// This object is modified, so must use copy
		field_dir = new DvObject(field_in);
		
		if(field_dir->seqSize() == 1){
			// create sequence from single object
			field_dir = new DvObject(field_dir, data->seqSize());
		}
		else{
		
	   		// join will make a copy. 
			field_dir =  field_dir->Join(data);

			if(field_dir.is_nil()) {
				message = "Field and PSD Data join failed.";
			}
		}	
    }    
	
    return true;

}

// ------------------

bool PSD_Parameters::init_data(DvObject_var &obj)
{
    data = new DvObject(obj);

	int n_recs = data->seqSize();
	int dimTheta = 0;
	int dimPhi = 0;
	
	if(_phi_centre) {
		delete _phi_centre; 
		_phi_centre = 0;
	}
	if(_phi_plus) {
		delete _phi_plus; 
		_phi_plus = 0;
	} 
	if(_phi_minus)  {
		delete _phi_minus; 
		_phi_minus = 0;
	} 
			
	if(_theta_centre)  {
		delete _theta_centre; 
		_theta_centre = 0;
	} 
	if(_theta_plus)  {
		delete _theta_plus; 
		_theta_plus = 0;
	} 
	if(_theta_minus)  {
		delete _theta_minus; 
		_theta_minus = 0;
	} 
 	
    data = data->changeUnitsTo("1>s^3 m^-6");
	
	if(data->getError().contains("change Units") )    {
        message = "Failed to convert distribution to SI units.";
        return false;
    }
    
	DvObject_var FillXref = data->get_xref(FILL_VALUE);
    if(FillXref.is_ok()) dataFillVal = FillXref->asDouble();
	else dataFillVal = DvNaN;
	
    chargeUnit=1.60217646e-19; //C
    massUnit=1.67262158e-27; // kg
      
    // Associating the xrefs to the coordinates
    
    bool initE(false), initPhi(false), initTheta(false);
    DvString strDepend;
    char strError[200];
    
    for(int i=1; i<=3; i++)
    {
        strDepend = "DEPEND_";
		strDepend += i;
        DvObject_var o = obj->get_xref(strDepend);
        
		bool isDeg, isAngle;

        isAngle = o->isAngle();
		isDeg = o->isDeg();
			
		if( isAngle )
		{
			double conv2rad=1.;
			if(isDeg) conv2rad=_PI/180.;

			DvObject_var AngFillXref = o->get_xref(FILL_VALUE);
    		double fillVal = FillXref->asDouble();

			double minVal;
			int kStart = 0, kEnd; // not used by qpsd
			bool isPhi = o->isPhiAngle(isDeg, i, minVal, kStart, kEnd);
				
			int n_dim = 0;
			int i_num = 0;
				
			double *i_centre = 0;
			double *i_plus = 0;
			double *i_minus = 0;
		
			if( n_recs != o->seqSize() && o->seqSize() != 1){
				sprintf(strError, "Depend seq not same length as data seq");
				message = strError;
				return false;
			}
					
			n_dim = o->arraySize();
			i_num = n_recs*n_dim;
			i_centre = new double[i_num];
			i_plus = new double[i_num];
			i_minus = new double[i_num];
					
			int count = 0;

			if(o->seqSize() > 1){
				for(int j=0; j<n_recs; j++)
				{
					for(int k=0; k<n_dim; k++)
					{
						double i_val = o->dbl(j,k);

						if( notFill(i_val, fillVal) )
						{
							i_val *= conv2rad;

						}
						i_centre[count] = i_val;
						count++;
							
					}
				}
			}
			else{
				for(int j=0; j<n_recs; j++)
				{
					for(int k=0; k<n_dim; k++)
					{
						double i_val = o->dbl(0,k); // use const value

						if( notFill(i_val, fillVal) )
						{
							i_val *= conv2rad;

						}
						i_centre[count] = i_val;
						count++;
							
					}
				}
			}
                
				
			// Get deltas
				
			DvObject_var oPlus = o->get_xref(DELTA_PLUS);

			if(oPlus->arraySize() > 1 && oPlus->seqSize() > 1) {
				for(size_t count=0; count < n_recs; count++) {
					for(size_t j=0; j<n_dim; j++){
						i_plus[count*n_dim + j] = i_centre[count*n_dim + j] + conv2rad * oPlus->asDouble(count, j);
					}
				}
			}
			else if(oPlus->arraySize() > 1 && oPlus->seqSize() == 1) {
				for(size_t count=0; count < n_recs; count++) {
					for(size_t j=0; j<n_dim; j++){
						i_plus[count*n_dim + j] = i_centre[count*n_dim + j] + conv2rad * oPlus->asDouble(j);
					}
				}
			}
			else if(oPlus->arraySize() == 1 && oPlus->seqSize() > 1) {
				for(size_t count=0; count < n_recs; count++) {
					double delta = oPlus->asDouble(count) * conv2rad;
					for(size_t j=0; j<n_dim; j++){
						i_plus[count*n_dim + j] = i_centre[count*n_dim + j] + delta;
					}
				}
			}
			else if(oPlus->arraySize() == 1 && oPlus->seqSize() == 1) {
				double delta = oPlus->asDouble(0) * conv2rad;
				for(size_t count=0; count < i_num; count++) {
					i_plus[count] = i_centre[count] + delta;
				}
			}
			else{
				message = "Invalid DELTA_PLUS for angle bin";
				return false;
			}
				
			DvObject_var oMinus=o->get_xref(DELTA_MINUS);

			if(oMinus->arraySize() > 1 && oMinus->seqSize() > 1) {
				for(size_t count=0; count < n_recs; count++) {
					for(size_t j=0; j<n_dim; j++){
						i_minus[count*n_dim + j] = i_centre[count*n_dim + j] - conv2rad * oMinus->asDouble(count, j);
					}
				}
			}
			else if(oMinus->arraySize() > 1 && oMinus->seqSize() == 1) {
				for(size_t count=0; count < n_recs; count++) {
					for(size_t j=0; j<n_dim; j++){
						i_minus[count*n_dim + j] = i_centre[count*n_dim + j] - conv2rad * oMinus->asDouble(j);
					}
				}
			}
			else if(oMinus->arraySize() == 1 && oMinus->seqSize() > 1) {
				for(size_t count=0; count < n_recs; count++) {
					double delta = oMinus->asDouble(count) * conv2rad;
					for(size_t j=0; j<n_dim; j++){
						i_minus[count*n_dim + j] = i_centre[count*n_dim + j] - delta;
					}
				}
			}
			else if(oMinus->arraySize() == 1 && oMinus->seqSize() == 1) {
				double delta = oMinus->asDouble(0) * conv2rad;
				for(size_t count=0; count < i_num; count++) {
					i_minus[count] = i_centre[count] - delta;
				}
			}
			else{
				message = "Invalid DELTA_MINUS for angle bin";
				return false;
			}
					
			// set angle to centre of bin
			for(size_t count=0; count < i_num; count++) {
				i_centre[count] = 0.5 * (i_plus[count] +  i_minus[count]);
			}

			bool offsetTheta=false;
                
			if( !isPhi)
			{
				if(minVal<0.)
				{
					offsetTheta=true;
				}
			}
                
                 
			if(isPhi)
			{
				_phi_centre = i_centre;
				_phi_plus = i_plus;
				_phi_minus = i_minus;
                    
				dimPhi = n_dim;
                    
				phiFillVal=fillVal;
				phi2rad=conv2rad;
				initPhi=true;
				indexPhi=i-1;
                    
				sprintf(strError, "Using \"%s\" as phi coordinate",  o->getXrefText("FIELDNAM").c_str());
				message = strError;
			}
			else
			{
				_theta_centre = i_centre;
				_theta_plus   = i_plus;
				_theta_minus  = i_minus;
				
				dimTheta = n_dim;
				// Theta <=> every value is between 0 and Pi
					
				for(int count=0; count < i_num; count++)
				{
					if(offsetTheta)
					{
						_theta_centre[count] = _PI/2. - _theta_centre[count];
						_theta_plus[count] = _PI/2. - _theta_plus[count];
						_theta_minus[count] = _PI/2. - _theta_minus[count];
					}
				}
                    
                                        
				thetaFillVal=fillVal;
				theta2rad=conv2rad;
				initTheta=true;
				indexTheta=i-1;
                    
				sprintf(strError, "Using \"%s\" as theta coordinate", o->getXrefText("FIELDNAM").c_str());
				message = strError;
			}
		}
		else
		{
			// We suppose that it is E
            e_seq = new DvObject(o);
			e_seq->ensureSIC(); // propagates SIC to depends if needed
			
			size_t eLen = e_seq->arraySize();
			
			if(e_seq->seqSize() == 1){
				e_seq = new DvObject(e_seq, data->seqSize());
			}
				    
			if(e_seq.is_ok())
			{
				e_seq = e_seq->changeUnitsTo("1>J");
				
				if(e_seq->getError().contains("change Units") )
				{
					message = "Wrong units for energy?";
					return false;
				}
				DvObject_var EnFillXref = e_seq->get_xref(FILL_VALUE);
    			eFillVal = FillXref->asDouble();
				
			}

			ePlus = e_seq->get_xref(DELTA_PLUS);
			
			if(ePlus.is_nil() )
			{
				message = "Invalid DELTA_PLUS for energy";
				return false;
			}
			if(ePlus->arraySize() == 1) {
				// single value (assume not t-dependent)
				vector <size_t> dims;
				dims.push_back(eLen);
				ePlus = new DvObject(ePlus->dbl(), dims, data->seqSize());
			}
			else if(ePlus->seqSize() == 1) ePlus = new DvObject(ePlus, data->seqSize()); // single array
              
			eMinus = e_seq->get_xref(DELTA_MINUS);
			if(eMinus.is_nil())
			{
				message = "Invalid DELTA_MINUS for energy";
				return false;
			}			
			if(eMinus->arraySize() == 1) {
				// single value (assume not t-dependent)
				vector <size_t> dims;
				dims.push_back(eLen);
				eMinus = new DvObject(eMinus->dbl(), dims, data->seqSize());
			}
			else if(eMinus->seqSize() == 1) eMinus = new DvObject(eMinus, data->seqSize()); // single array

			initE=true;
			indexE = i-1;
                
			sprintf(strError, "Using \"%s\" as energy coordinate", o->getXrefText("FIELDNAM").c_str());
			message = strError;
		}
	}
    
    if(!initE || !initPhi || !initTheta)
    {
        message = "Failed finding the 3 coordinates";
        return false;
    }
    
    if(data->dims[indexE] != e_seq->dims[0])
    { 
        message = "Dimensions for data and energy coordinates do not match";
        return false;
    }
    if(data->dims[indexTheta] != dimTheta)
    {
        message = "Dimensions for data and theta coordinates do not match";
        return false;
    }
    if(data->dims[indexPhi] != dimPhi)
    {
        message = "Dimensions for data and Phi coordinates do not match";
        return false;
    }
    
    maxTime=data->seqSize()-1;
    maxE=data->dims[indexE]-1;
    maxTheta=data->dims[indexTheta]-1;
    maxPhi=data->dims[indexPhi]-1;
   
    return true;
}



// ------------------

bool PSD_Parameters::init_cartesian()
{
	// set up space for cartesian coords and other versions of values used often

	
	if(_x) {delete _x; _x=NULL;}
	if(_y) {delete _y; _y=NULL;}
	if(_z) {delete _z; _z=NULL;}
	if(_vel) {delete _vel; _vel=NULL;}

    _x = new double[(maxE+1) * (maxTheta+1) * (maxPhi+1)];
    _y = new double[(maxE+1) * (maxTheta+1) * (maxPhi+1)];
    _z = new double[(maxE+1) * (maxTheta+1) * (maxPhi+1)];
    _vel = new double[(maxE+1)];
		
	
	return true;
	
}
// ------------------

bool PSD_Parameters::fill_cartesian(int time)
{
	// set up cartesian coords and other versions of values used often
	if(data.is_nil()) return false;
	
	int n_seq = data->seqSize();
	double _theta, _phi;
	double x, y, z;
	
	// loops must be in this order
	int j=0;
 	
	for(int e=0; e<=maxE; e++){
		_vel[e] = getV(time, e);
	}
	
	for(int phi=0; phi<=maxPhi; phi++){
		_phi = getPhi(time, phi);
				
		for(int theta=0; theta<=maxTheta; theta++){
			_theta = getTheta(time, theta);

			for(int e=0; e<=maxE; e++){
				double r = _vel[e];
				if(r >= 0.)
				{
					r *= v_sign;   // v_sign performs look to flow if needed
					x = r*sin(_theta)*cos(_phi);
					y = r*sin(_theta)*sin(_phi);
					z = r*cos(_theta);
					
					// rotate
/* post-mult mtx
					if(!rotn_const.is_nil()) {
						_x[j] = x*(*rotn_const)(0,0) + y*(*rotn_const)(1,0) + z*(*rotn_const)(2,0);
						_y[j] = x*(*rotn_const)(0,1) + y*(*rotn_const)(1,1) + z*(*rotn_const)(2,1);
						_z[j] = x*(*rotn_const)(0,2) + y*(*rotn_const)(1,2) + z*(*rotn_const)(2,2);
					}
					else {
						_x[j] = x*(*rotn_seq)[time](0,0) + y*(*rotn_seq)[time](1,0) + z*(*rotn_seq)[time](2,0);
						_y[j] = x*(*rotn_seq)[time](0,1) + y*(*rotn_seq)[time](1,1) + z*(*rotn_seq)[time](2,1);
						_z[j] = x*(*rotn_seq)[time](0,2) + y*(*rotn_seq)[time](1,2) + z*(*rotn_seq)[time](2,2);
					}
*/
					if(rotn_obj->seqSize() == 1) {
						_x[j] = x * rotn_obj->dbl(0,0,0) + y * rotn_obj->dbl(0,0,1) + z * rotn_obj->dbl(0,0,2);
						_y[j] = x * rotn_obj->dbl(0,1,0) + y * rotn_obj->dbl(0,1,1) + z * rotn_obj->dbl(0,1,2);
						_z[j] = x * rotn_obj->dbl(0,2,0) + y * rotn_obj->dbl(0,2,1) + z * rotn_obj->dbl(0,2,2);
					}
					else {
						_x[j] = x * rotn_obj->dbl(time, 0,0) + y * rotn_obj->dbl(time, 0,1) + z * rotn_obj->dbl(time, 0,2);
						_y[j] = x * rotn_obj->dbl(time, 1,0) + y * rotn_obj->dbl(time, 1,1) + z * rotn_obj->dbl(time, 1,2);
						_z[j] = x * rotn_obj->dbl(time, 2,0) + y * rotn_obj->dbl(time, 2,1) + z * rotn_obj->dbl(time, 2,2);
					}
				}
				else{
					_x[j] = 0;
					_y[j] = 0;
					_z[j] = 0;
				}
				j++;
			}
		}
	}

	return true;
}


// ------------------

bool PSD_Parameters::init_potential(DvObject_var &obj)
{
    potential = new DvObject(obj);
	
    if(potential.is_ok())
    {
        potential = potential->changeUnitsTo("1>V");
        
        if( potential->seqSize() > 1  )
        {
			potential =  potential->Join(data);

			if(potential.is_nil()) {
				message = "Potential and Data join failed.";
            	return false;
			}
			
        }
		else{
			potential = new DvObject(potential, data->seqSize());
		}
        //TODO potFillVal
    }
	else
	{
		potential = new DvObject(0.0, data->seqSize());
	}
    
    return true;
}

// ------------------

bool PSD_Parameters::init_threshold(DvObject_var &obj)
{
    if(obj.is_ok() && obj->seqSize() > 1)
    {
        threshold = obj->changeUnitsTo("1>J");
        if(threshold->getError().contains("change Units") )
        {
            message = "Threshold conversion to Joules failed, assuming eV";
			*threshold *= 1.60217646e-19;  // convert eV to J
        }
        
        if( !threshold->isJoined(*data)  )
        {
			threshold = threshold->Join(data);

			if(threshold.is_nil()) {
				message = "Potential and Data join failed.";
            	return false;
			}			
        }
    }
    else
    {
        if(obj.is_ok())
        {
			threshold = obj->changeUnitsTo("1>J");
			
			if(threshold->getError().contains("change Units")){
				  *threshold *= 1.60217646e-19; // assume eV and convert to J
			}
			// convert to sequence
            threshold = new DvObject(threshold->asDouble(0), data->seqSize());
        }
        else
        {
            threshold = new DvObject(0.0, data->seqSize());
        }
    }
    
    return true;
}

// ------------------


bool PSD_Parameters::init_rotn(DvObject_var& rotMatx)
{
	rotn_obj = rotMatx;
	
    if(rotn_obj.is_ok() && rotn_obj->seqSize() > 1)
    {
       
		// join rotnMatrix to data tt
		rotn_obj =  rotn_obj->Join(data);

		if(rotn_obj.is_nil() ) {
			message = "rotation sequence and Data not joined.";
			return false;
		}
     }  
	  
	if(rotn_obj->nDims() != 2)
	{
		message = "Wrong dimensions for rotation matrix.";
		return false;
	}
        
	if(rotn_obj->dims[0] != 3 || rotn_obj->dims[1] != 3)
	{
		message = "Wrong dimensions for rotation matrix.";
		return false;
	}
    
    return true;
}


// ------------------


DvObject_var PSD_Parameters::computeValuesAlongVector(int time, DvObject_var &CoM, DvObject_var &direction, double v_min, double v_max, int nb_values)
{
// NOT FULLY IMPLEMENTED YET
    if(CoM.is_nil() || direction.is_nil())
    {
        cout << "MomentsParameters::computeValuesAlongVector: nil centerOfMass and/or direction" << endl;
        return DvObject_var();
    }
    
    if(CoM->dims[0] != 3 || direction->dims[0] != 3 )
    {
        cout << "MomentsParameters::computeValuesAlongVector: wrong dimension for centerOfMass and/or direction" << endl;
        return DvObject_var();
    }
    
    double val;
    

	buildTetra(time);
    
    vector <size_t> dim;
	dim.push_back(nb_values);
    DvObject_var res = new DvObject(0.0, dim);
	

    double element_size = (v_max-v_min)/nb_values;
    
    vector <size_t> dim33;
	dim33.push_back(3);
	dim33.push_back(3);
	DvObject_var rotation = new DvObject(0.0, dim33);
	
//    getRotationMatrix(*direction, rotation);
    
    if(!tetra.empty())
    {
        list<Point3D> points;
        for(int i=0; i<nb_values; i++)
        {
            double _z = v_min+(i+0.5)*element_size;
            double x = rotation->dbl(0,2) * _z + CoM->dbl(0);
            double y = rotation->dbl(1,2) * _z + CoM->dbl(1);
            double z = rotation->dbl(2,2) * _z + CoM->dbl(2);
            
            Point3D temp(x, y, z, -1.);
            temp.orig_x=i;
            points.push_front(temp);
        }
        
        list<Element3D> tetra_temp(tetra);
        interpolation(points, tetra_temp);
        
        for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        {
            res->dbl((int)(iter->orig_x)) = iter->val;
        }
        
        // Creation of the metadata
        res->change_xref("UNITS", "s^3/m^6");
        DvString fieldnam = data->getXrefText(FIELDNAM);
        fieldnam += " - Values along vector";
        res->change_xref("FIELDNAM", fieldnam);
        res->change_xref("FRAME", "scalar>na");
        res->change_xref("PARAMETER_TYPE", "Data");
        res->change_xref("SI_CONVERSION", "1>s^3 m^-6");
		res->change_xref("DISPLAYTYPE", "Spectrogram");
        
        DvObject_var axis = new DvObject(0.0, dim);
        for(int i=0; i<nb_values; i++)
        {
            axis->dbl(i) = (i+0.5)*element_size+v_min;
        }
		DvObject_var doubleFill = new DvObject(-1.e+31);
        axis->change_xref(FILL_VALUE, doubleFill );
        axis->change_xref("UNITS", "m s-1");
        fieldnam = data->getXrefText(FIELDNAM);
		fieldnam += " - Values along vector - Axis";
        axis->change_xref("FIELDNAM", fieldnam.c_str());
        axis->change_xref("FRAME", "scalar>na");
        axis->change_xref("PARAMETER_TYPE", "Data");
        axis->change_xref("SI_CONVERSION", "1>m s^-1");
        
        DvObject_var delta_axis = new DvObject((double)element_size/2.);
        delta_axis->change_xref(FILL_VALUE, doubleFill );
        delta_axis->change_xref("UNITS", "m s-1");
        fieldnam = data->getXrefText(FIELDNAM);
		fieldnam +=  " - Values along vector - Axis Delta";
        delta_axis->change_xref("FIELDNAM", fieldnam.c_str());
        delta_axis->change_xref("FRAME", "scalar>na");
        delta_axis->change_xref("PARAMETER_TYPE", "Data");
        delta_axis->change_xref("SI_CONVERSION", "1>m s^-1");
        
        axis->change_xref("DELTA_PLUS", delta_axis);
        axis->change_xref("DELTA_MINUS", delta_axis);
        res->change_xref("DEPEND_1", axis);
    }
    else
    {
        return DvObject_var();
    }

    return res;
}

//----------------------------

// computePAD here is legacy code, unfinished (I think)
// replaced by getPAD which borrows much of the logic.
// getPAD borrows much from getSlice

DvObject_var PSD_Parameters::computePAD(int time, DvObject_var &CoM, DvObject_var &direction, double v_min, double v_max, int nb_v, int nb_theta)
{
    if(CoM.is_nil() || direction.is_nil())
    {
        cout << "PSD_Parameters::computePAD: nil centerOfMass and/or direction" << endl;
        return DvObject_var();
    }
    
    if(CoM->dims[0] != 3 || direction->dims[0] != 3)
    {
        cout << "PSD_Parameters::computePAD: wrong dimension for centerOfMass and/or direction" << endl;
        return DvObject_var();
    }
    

	buildTetra(time);
    
	vector <size_t> dim;
	dim.push_back(nb_v);
	dim.push_back(nb_theta);
    DvObject_var res = new DvObject(0.0, dim);
	
    double element_size_v=(v_max-v_min)/nb_v;
    
    // Find rotation matrix
    
    vector <size_t> dim33;
	dim33.push_back(3);
	dim33.push_back(3);
	DvObject_var rotation = new DvObject(0.0, dim33); // DOES NOT seem to be needed

 //   getRotationMatrix(*direction, rotation);
    
    // End find rotation matrix
    
    if(!tetra.empty())
    {
        list<Point3D> points;
        for(int i=0; i<nb_v; i++)
        {
            double v=v_min+(i+0.5)*element_size_v;
            for(int j=0; j<nb_theta; j++)
            {
                double theta=(_PI/nb_theta)*(j+0.5);
                double total=0.;
                bool foundOnce=false;
                
                for(int k=0; k<360; k+=10)
                {
                    double phi=_PI/180.*(k+5.);
                    double x=v*sin(theta)*cos(phi);
                    double y=v*sin(theta)*sin(phi);
                    double z=v*cos(theta);
                    
                    Point3D temp(x, y, z, -1.);
                    temp.orig_x=i;
                    temp.orig_y=j;
                    points.push_front(temp);
                }
               
            }
        }
        
        list<Element3D> tetra_temp(tetra);
        interpolation(points, tetra_temp);
                
        for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        {
            if(iter->val>=0.)
            {
				res->dbl((int)(iter->orig_x), (int)(iter->orig_y)) += iter->val*10.;
               
            }
        }
        
        // Creation of the metadata

        res->change_xref("UNITS", "s^3 rad/m^6");
        DvString fieldnam = data->getXrefText(FIELDNAM);
		fieldnam += " - PAD";
        res->change_xref("FIELDNAM", fieldnam);
        res->change_xref("FRAME", "scalar>na");
        res->change_xref("PARAMETER_TYPE", "Data");
        res->change_xref("SI_CONVERSION", "1>s^3 rad m^-6");
        res->change_xref("DISPLAYTYPE", "Spectrogram");
        
		vector <size_t> dimv;
		dimv.push_back(nb_v);
        DvObject_var v_axis = new DvObject(0.0, dimv);
        for(int i=0; i<nb_v; i++)
        {
            v_axis->dbl(i) = v_min+(i+0.5)*element_size_v;
        }
        v_axis->change_xref("UNITS", "m s-1");
        fieldnam = data->getXrefText(FIELDNAM);
		fieldnam += " - PAD - V Axis";
        v_axis->change_xref("FIELDNAM", fieldnam.c_str());
        v_axis->change_xref("FRAME", "scalar>na");
        v_axis->change_xref("PARAMETER_TYPE", "Data");
        v_axis->change_xref("SI_CONVERSION", "1>m s^-1");
        
        DvObject_var delta_v_axis = new DvObject((double)element_size_v/2.);
        delta_v_axis->change_xref("UNITS", "m s-1");
        fieldnam = data->getXrefText(FIELDNAM);
		fieldnam += " - PAD - V Axis Delta";
        delta_v_axis->change_xref("FIELDNAM", fieldnam);
        delta_v_axis->change_xref("FRAME", "scalar>na");
        delta_v_axis->change_xref("PARAMETER_TYPE", "Data");
        delta_v_axis->change_xref("SI_CONVERSION", "1>m s^-1");
        
		vector <size_t> dimt;
		dimt.push_back(nb_theta);
        DvObject_var t_axis = new DvObject(0.0, dimt);
        for(int i=0; i<nb_theta; i++)
        {
            t_axis->dbl(i) = (_PI/nb_theta)*(i+0.5);
        }
        t_axis->change_xref("UNITS", "rad");
        fieldnam = data->getXrefText(FIELDNAM);
		fieldnam += " - PAD - Theta Axis";
        t_axis->change_xref("FIELDNAM", fieldnam);
        t_axis->change_xref("FRAME", "scalar>na");
        t_axis->change_xref("PARAMETER_TYPE", "Data");
        t_axis->change_xref("SI_CONVERSION", "1>rad");
        
        DvObject_var delta_t_axis = new DvObject(5./180.*_PI);
        delta_t_axis->change_xref("UNITS", "rad");
        fieldnam = data->getXrefText(FIELDNAM);
		fieldnam += " - PAD - Theta Axis Delta";
        delta_t_axis->change_xref("FIELDNAM", fieldnam);
        delta_t_axis->change_xref("FRAME", "scalar>na");
        delta_t_axis->change_xref("PARAMETER_TYPE", "Data");
        delta_t_axis->change_xref("SI_CONVERSION", "1>rad");
        
        v_axis->change_xref("DELTA_PLUS", delta_v_axis);
        v_axis->change_xref("DELTA_MINUS", delta_v_axis);
        t_axis->change_xref("DELTA_PLUS", delta_t_axis);
        t_axis->change_xref("DELTA_MINUS", delta_t_axis);
        res->change_xref("DEPEND_1", v_axis);
        res->change_xref("DEPEND_2", t_axis);
    }
    else
    {
        return DvObject_var();
    }
    
    return res;
}



// ------------------

void PSD_Parameters::buildTetra(int time)
{

    tetra.clear();
    list<Point3D> points;
    double val, vol, ePThis, eMThis;
	double x, y, z;
	
	// for finding Vmax and min
	
	int e_of_max = -1;
	int e_of_min = -1;	
	
	double max_E = 0;
	double min_E = 1.e31;
	for(int e=0; e<=maxE; e++)
	{
		if(_vel[e] >= 0.){
			ePThis = getEPlus(time, e);	
			eMThis = getEMinus(time, e);
			if(ePThis > max_E)
			{
				max_E=ePThis;
				e_of_max = e;
			}
			if(eMThis > 0. && eMThis < min_E)
			{
				min_E=eMThis;
				e_of_min = e;
			}
		}
	}
	
	// find points
	
	int j = 0;
	
    for(int phi=0; phi<=maxPhi; phi++)
    {		
        for(int theta=0; theta<=maxTheta; theta++)
        {
            for(int e=0; e<=maxE; e++)
            {
                val = getData(time, e, phi, theta);
				
                if( notFill(val, dataFillVal) )
                {
					vol = getElementVolume(time, e, phi, theta);
                    if(_vel[e] >= 0. && vol > 0.)
                    {
						// set point of bin centre
						Point3D temp(_x[j], _y[j], _z[j], val, vol);
                        points.push_front(temp);
						
						// Max and min points...
						// direction is same for E and EPlus/minus, so just renormalise
						if(e == e_of_max)
                        {
							double max_v = E2V(max_E)/_vel[e];
			
							x = max_v * _x[j];
            				y = max_v * _y[j];
            				z = max_v * _z[j];
               
							Point3D temp_max(x, y, z, val, vol);
							points.push_front(temp_max);
							
                        }
                        if(e == e_of_min)
                        {
							double min_v = E2V(min_E)/_vel[e];
			
							x = min_v * _x[j];
            				y = min_v * _y[j];
            				z = min_v * _z[j];
               
							Point3D temp_min(x, y, z, val, vol);
							points.push_front(temp_min);
                        }
						
						
                    }
                }
				j++;
            }
			
        }
    }

    if(points.size()>4)
    {
         triangulation(points, tetra);
    }
}

//-----------------------------

void PSD_Parameters::getRotationMatrix(double direction[], double rotation[][3])
{
	// NOTE this modifies the direction matrix, so it must be a copy of WL obj
	// See notes below. The resulting rotation matrix takes a pt (u,v,w) in the
	// coord sys which has dirn=(0,0,1) and gives the coords (x,y,z) in the
	// system in which the input dirn is expressed. It is used, for example,
	// to lay down an arbitrary cartesian plane with its natural coords and
	// then find the GSE coords of the grid-points of that plane.
    double norm=sqrt(direction[0]*direction[0]+direction[1]*direction[1]+direction[2]*direction[2]);
    direction[0]/=norm;
    direction[1]/=norm;
    direction[2]/=norm;
    
    if(fabs(direction[0])<0.01 && fabs(direction[1])<0.01)
    {
        if(direction[2]>0.)
        {
            rotation[0][0]=1.;
            rotation[1][1]=1.;
            rotation[2][2]=1.;
            
            rotation[0][1]=0.;
            rotation[0][2]=0.;
            rotation[1][0]=0.;
            rotation[1][2]=0.;
            rotation[2][0]=0.;
            rotation[2][1]=0.;
        }
        else
        {
            rotation[0][0]=-1.;
            rotation[1][1]=-1.;
            rotation[2][2]=-1.;
            
            rotation[0][1]=0.;
            rotation[0][2]=0.;
            rotation[1][0]=0.;
            rotation[1][2]=0.;
            rotation[2][0]=0.;
            rotation[2][1]=0.;
        }
    }
    else
    {
        double normal[3]; // Normal to z axis and direction vector
        normal[0]=-(double)direction[1];
        normal[1]=direction[0];
        normal[2]=0.;
        double normal_norm=sqrt(normal[0]*normal[0]+normal[1]*normal[1]+normal[2]*normal[2]);
        normal[0]/=normal_norm;
        normal[1]/=normal_norm;
        normal[2]/=normal_norm;
        double angle=acos(direction[2]);// Rotation angle around the normal so that z matches the direction 
        // Change angle sign if necessary
        //if ((normal X z).direction >0, sin>0, no sign change)
		// SJS - following line always leaves angle  +ve. Let dir=(X,Y,Z)
		// then norm = (-Y,X,0) and "if" expression below is (X^2 + Y^2).
		// Investigation suggests +ve angle always correct, and gives matrix
		// s.t. rot (0,0,1) = (X,Y,Z)
        if(normal[1]*direction[0]-normal[0]*direction[1] <0.) angle=-angle;

        rotation[0][0]=1. + (1.-cos(angle))*(normal[0]*normal[0]-1);
        rotation[0][1]=-normal[2]*sin(angle)+(1.-cos(angle))*normal[0]*normal[1];
        rotation[0][2]=normal[1]*sin(angle)+(1.-cos(angle))*normal[0]*normal[2];
        rotation[1][0]=normal[2]*sin(angle)+(1.-cos(angle))*normal[0]*normal[1];
        rotation[1][1]=1. + (1.-cos(angle))*(normal[1]*normal[1]-1);
        rotation[1][2]=-normal[0]*sin(angle)+(1.-cos(angle))*normal[1]*normal[2];
        rotation[2][0]=-normal[1]*sin(angle)+(1.-cos(angle))*normal[0]*normal[2];
        rotation[2][1]=normal[0]*sin(angle)+(1.-cos(angle))*normal[1]*normal[2];
        rotation[2][2]=1. + (1.-cos(angle))*(normal[2]*normal[2]-1);
    }
}

// ------------------

void PSD_Parameters::getSlice(DvObject_var &PSD_slice, QLabel *percLabl)
{
    double scalemin(1.e30), scalemax(-1.);
    int seq_size = data->seqSize();

	vector <size_t> dim;
	dim.push_back(resolution);
	dim.push_back(resolution);
	PSD_slice = new DvObject (0.0, dim, seq_size);
	
	// get V scale
	double max_v=0.;
   	for(int time=0; time<seq_size; time++){
		for(int e=0; e<=maxE; e++)
    	{
        	double v=getVPlus(time, e);
        	if(v>max_v) max_v=v;
    	}
    }
	double element_size=2.*max_v/resolution;
	
	// get slices
    for(int time=0; time<seq_size; time++)
    {
		int perc = (100 * time) / seq_size;

		if(percLabl) percLabl->setText(QString("%1%").arg(perc));
		qApp->processEvents();	
		
		if(this->stop) {
			seq_size = time;
			continue; // allows clean interrupt (keep what's done)
		}
		
		// get points in rotated cartesian frame
		
		fill_cartesian(time);
        
		// build the teselating tetrahedra

		buildTetra(time);

		// get the actual slice
		
		
    
    	// Find rotation matrix into normal
    
    	double rotation[3][3];
		
		double dirn[3];
		size_t atT = time;
		if(sliceNormal->seqSize() == 1) atT=0;
		
		dirn[0] = sliceNormal->dbl(atT,0);
		dirn[1] = sliceNormal->dbl(atT,1);
		dirn[2] = sliceNormal->dbl(atT,2);
		
    	getRotationMatrix(dirn, rotation);


    	if(!tetra.empty())
    	{
        	list<Point3D> points;
			for(int i=0; i<resolution; i++)
        	{
            	for(int j=0; j<resolution; j++)
            	{
                	double _x=(i+0.5)*element_size-max_v;
                	double _y=(j+0.5)*element_size-max_v;
                	double _z=0.;
                	double x=rotation[0][0]*_x+rotation[0][1]*_y+rotation[0][2]*_z+centre_PSD->dbl(time,0);
                	double y=rotation[1][0]*_x+rotation[1][1]*_y+rotation[1][2]*_z+centre_PSD->dbl(time,1);
                	double z=rotation[2][0]*_x+rotation[2][1]*_y+rotation[2][2]*_z+centre_PSD->dbl(time,2);
               
                	Point3D temp(x, y, z, -1.);
                	temp.orig_x=i;
                	temp.orig_y=j;
                	points.push_front(temp);
                
            	}
        	}

        	interpolation(points, tetra);
        
        	for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        	{
            	PSD_slice->dbl(time,(int)(iter->orig_x),(int)(iter->orig_y)) = iter->val;
				
				// update scale min/max
				if( iter->val > 0) // not fill
				{
					if( iter->val < scalemin) scalemin= iter->val;
					if( iter->val > scalemax) scalemax= iter->val;					
				}
				
        	}
        }		


	}
	
	//  get xrefs and modify DEPENDS
	
	PSD_slice->copy_xrefs_from(data);

	DvObject_var unitFill = new DvObject(-1.);
	PSD_slice->change_xref(FILL_VALUE, unitFill );
	PSD_slice->change_xref("UNITS", "s^3/m^6");
	DvString fieldnam = data->getXrefText(FIELDNAM);
	fieldnam +=  " - Slice";
	PSD_slice->change_xref("FIELDNAM", fieldnam);
	PSD_slice->change_xref("PARAMETER_TYPE", "Data");
	PSD_slice->change_xref("SI_CONVERSION", "1>s^3 m^-6");
	PSD_slice->change_xref("DISPLAYTYPE", "Spectrogram");
	
	
	vector <size_t> dimr;
	dimr.push_back(resolution);	
	DvObject_var axis = new DvObject(0.0, dimr);
	for(int i=0; i<resolution; i++)
	{
		axis->dbl(i) = (i+0.5)*element_size-max_v;
	}
	axis->change_xref(FILL_VALUE, unitFill );
	axis->change_xref("UNITS", "m s-1");
	fieldnam = data->getXrefText(FIELDNAM);
	fieldnam +=   " - Slice - Axis";
	axis->change_xref("FIELDNAM", fieldnam);
	axis->change_xref("FRAME", "scalar>na");
	axis->change_xref("PARAMETER_TYPE", "Data");
	axis->change_xref("SI_CONVERSION", "1>m s^-1");
        
	DvObject_var delta_axis = new DvObject(element_size/2.);
	delta_axis->change_xref(FILL_VALUE, unitFill );
	delta_axis->change_xref("UNITS", "m s-1");
	fieldnam = data->getXrefText(FIELDNAM);
	fieldnam +=   " - Slice - Axis Delta";
	delta_axis->change_xref("FIELDNAM", fieldnam);
	delta_axis->change_xref("FRAME", "scalar>na");
	delta_axis->change_xref("PARAMETER_TYPE", "Data");
	delta_axis->change_xref("SI_CONVERSION", "1>m s^-1");
        
	axis->change_xref("DELTA_PLUS", delta_axis);
	axis->change_xref("DELTA_MINUS", delta_axis);
	PSD_slice->change_xref("DEPEND_1", axis);
	PSD_slice->change_xref("DEPEND_2", axis);
	if (PSD_slice->xref_exists("DEPEND_3")) PSD_slice->delete_xref("DEPEND_3");
     
	DvObject_var smin = new DvObject(scalemin);
	DvObject_var smax = new DvObject(scalemax);
	
    PSD_slice->change_xref("SCALEMIN", smin);
    PSD_slice->change_xref("SCALEMAX", smax);

}


void PSD_Parameters::getAdvSlice(DvObject_var &PSD_slice, int choice, QLabel *percLabl)
{
    double scalemin(1.e30), scalemax(-1.);
    int seq_size = data->seqSize();

	vector <size_t> dim;
	// need even number of velocity bins (same +axis, -axis)
	if (resolution % 2 == 1) resolution +=1;  
	dim.push_back(resolution);
	dim.push_back(resolution);
	PSD_slice = new DvObject (0.0, dim, seq_size);
	
	// get V scale
	double max_v=0.;
	double min_v= 1.e31;
   	for(int time=0; time<seq_size; time++){
		for(int e=0; e<=maxE; e++)
    	{
        	double v=getVPlus(time, e);
        	if(v>max_v) max_v=v;
			v=getVMinus(time, e);
			if(v>0. && v<min_v) min_v=v;
    	}
    }
	if (max_v < min_v){  //shouldn't get to here given above but ...
		printf("Failed to find valid velocity range in Adv slice\n");
		return;
	}
	// indices for different things
	//Edges -hr   -hr-1     -2    -1    0    1    2      hr-2     hr-1     hr
	//       |--------|------|-----|----|----|----|----|------|-----|-------|
	//V          0      1           hr-1  hr  hr+1 hr+2        2hr-2   2hr-1   
	int half_res = (resolution)/2;
	int n_bins_nr_zero = 0.161*resolution;   
				// total bins with |v| < min_v and lin spaced to v=0
				// half this for one side; must be even!
				// hard-wired; could be made user-settable
	if (n_bins_nr_zero % 2 == 1) n_bins_nr_zero += 1;
	int n_half_nr_zero = n_bins_nr_zero/2;
	if(half_res <= n_half_nr_zero)  //safety check; 
	{
		n_half_nr_zero =1;  // drop back to single bin
	}
	// transition from lin to log-spaced at min_v <-> nr0
	// Ex has nr0=3, hr=7 with nr0=n_half_nr_zero, hr=half_res
	//
	//Edges  -1    0    1    2   nr0  nr0+1  nr0+2    hr-1     hr
	//      --|----|----|----|----|-----|------|--------|--------|
	//V        hr-1  hr  hr+1 hr+2        2hr-3   2hr-2   2hr-1   
	
	double dlogV = (log(max_v/min_v))/(half_res - n_half_nr_zero);
	double V_fac = exp(dlogV);
	valarray <double> Vedges(half_res+1);
	valarray <double> _vx(resolution); // output arrays
	valarray <double> _vy(resolution);
	valarray <double> delta_V(resolution); // output deltas; note: _v centred in bin
	Vedges[0]=0;
	double dV_lin_nr_zero = min_v/n_half_nr_zero;
	for (int i=1; i<n_half_nr_zero+1; i++)
	{
		Vedges[i] = i*dV_lin_nr_zero;	
	}
	for(int i=n_half_nr_zero+1; i<half_res+1; i++)
	{
		Vedges[i] = V_fac*Vedges[i-1];
	}
	// We have all the edges, so now compute bins, both +ve and -ve
	for(int i=0; i<half_res; i++)
	{
		double dV = Vedges[i+1] - Vedges[i];
		double V = 0.5*(Vedges[i+1] + Vedges[i]);
		_vx[half_res-1-i]     = -V;
		_vy[half_res-1-i]     = -V;
		_vx[half_res+i]       = +V;
		_vy[half_res+i]       = +V;
		delta_V[half_res-1-i] = dV/2.;
		delta_V[half_res+i]   = dV/2.;
	}
		
	// get slices
    for(int time=0; time<seq_size; time++)
    {
		int perc = (100 * time) / seq_size;

		if(percLabl) percLabl->setText(QString("%1%").arg(perc));
		qApp->processEvents();	
		
		if(this->stop) {
			seq_size = time;
			continue; // allows clean interrupt (keep what's done)
		}
		
		// get points in rotated cartesian frame
		
		fill_cartesian(time);
        
		// build the teselating tetrahedra

		buildTetra(time);
    
    	// Find rotation matrix into normal; rot takes a vector in slice
		// coords and rotates it to be in target science frame, so for example
		// rot(0,0,1) = sliceNormal. Thus the COLUMNS of S are the vectors
		// sliceY, slicenormal
    
    	double rotation[3][3];
		
		size_t atT = time;
		if(sliceNormal->seqSize() == 1) atT=0;  
		// note sliceX, sliceY, sliceNormal already joined
		
		for(int row=0; row<3; row++)
		{
			rotation[row][0] = sliceX->dbl(atT,row);
			rotation[row][1] = sliceY->dbl(atT,row);
			rotation[row][2] = sliceNormal->dbl(atT,row);			
		}
		
    	if(!tetra.empty())
    	{
        	list<Point3D> points;
			// work through _vx, _vy grid, rotate each pt to target sci frame
			// then push onto list of pts
			for(int i=0; i<resolution; i++)
        	{
            	for(int j=0; j<resolution; j++)
            	{
//                	double _vz=0.;  no need - just remove 3rd element below
                	double x=rotation[0][0]*_vx[i]+rotation[0][1]*_vy[j]+centre_PSD->dbl(time,0);
                	double y=rotation[1][0]*_vx[i]+rotation[1][1]*_vy[j]+centre_PSD->dbl(time,1);
                	double z=rotation[2][0]*_vx[i]+rotation[2][1]*_vy[j]+centre_PSD->dbl(time,2);
               
                	Point3D temp(x, y, z, -1.);
                	temp.orig_x=i;
                	temp.orig_y=j;
                	points.push_front(temp);
                
            	}
        	}

        	interpolation(points, tetra);
        
        	for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        	{
            	PSD_slice->dbl(time,(int)(iter->orig_x),(int)(iter->orig_y)) = iter->val;
				
				// update scale min/max
				if( iter->val > 0) // not fill
				{
					if( iter->val < scalemin) scalemin= iter->val;
					if( iter->val > scalemax) scalemax= iter->val;					
				}
				
        	}
        }		


	}
	
	//  get xrefs and modify DEPENDS
	
	PSD_slice->copy_xrefs_from(data);

	DvObject_var unitFill = new DvObject(-1.);
	PSD_slice->change_xref(FILL_VALUE, unitFill );
	PSD_slice->change_xref("UNITS", "s^3/m^6");
	DvString textstr = data->getXrefText(FIELDNAM);
	textstr +=  " - Slice";
	PSD_slice->change_xref("FIELDNAM", textstr);
	PSD_slice->change_xref("PARAMETER_TYPE", "Data");
	PSD_slice->change_xref("SI_CONVERSION", "1>s^3 m^-6");
	PSD_slice->change_xref("DISPLAYTYPE", "Spectrogram");
	
	
	vector <size_t> dimr;
	dimr.push_back(resolution);	
	DvObject_var xaxis = new DvObject(_vx,dimr);
	DvObject_var yaxis = new DvObject(_vy,dimr);
	DvObject_var delta_axis = new DvObject(delta_V,dimr);
// X-axis
	textstr = sliceX->getXrefText("FIELDNAM");
	xaxis->change_xref(FILL_VALUE, unitFill );
	xaxis->change_xref("UNITS", "m s-1");
	if (textstr.blank())
	{
		textstr = "Slice vector V1";
	}
	xaxis->change_xref("LABLAXIS", textstr);
	xaxis->change_xref("FIELDNAM", textstr);
	xaxis->change_xref("FRAME", "scalar>na");
	xaxis->change_xref("PARAMETER_TYPE", "Data");
	xaxis->change_xref("SI_CONVERSION", "1>m s^-1");
// Y-axis
	textstr = sliceY->getXrefText("FIELDNAM");
	yaxis->change_xref(FILL_VALUE, unitFill );
	yaxis->change_xref("UNITS", "m s-1");
	if (textstr.blank())
	{
		textstr = "Slice perp part of vector V2";
	}
	yaxis->change_xref("FIELDNAM", textstr);
	yaxis->change_xref("LABLAXIS", textstr);
	yaxis->change_xref("FRAME", "scalar>na");
	yaxis->change_xref("PARAMETER_TYPE", "Data");
	yaxis->change_xref("SI_CONVERSION", "1>m s^-1");
	
	delta_axis->change_xref(FILL_VALUE, unitFill );
	delta_axis->change_xref("UNITS", "m s-1");
	textstr = data->getXrefText(FIELDNAM);
	textstr +=   " - Slice - Axis Delta";
	delta_axis->change_xref("FIELDNAM", textstr);
	delta_axis->change_xref("FRAME", "scalar>na");
	delta_axis->change_xref("PARAMETER_TYPE", "Data");
	delta_axis->change_xref("SI_CONVERSION", "1>m s^-1");
        
	xaxis->change_xref("DELTA_PLUS", delta_axis);
	xaxis->change_xref("DELTA_MINUS", delta_axis);
	yaxis->change_xref("DELTA_PLUS", delta_axis);
	yaxis->change_xref("DELTA_MINUS", delta_axis);
	PSD_slice->change_xref("DEPEND_1", xaxis);
	PSD_slice->change_xref("DEPEND_2", yaxis);
	if (PSD_slice->xref_exists("DEPEND_3")) PSD_slice->delete_xref("DEPEND_3");
     
	DvObject_var smin = new DvObject(scalemin);
	DvObject_var smax = new DvObject(scalemax);
	
    PSD_slice->change_xref("SCALEMIN", smin);
    PSD_slice->change_xref("SCALEMAX", smax);

}



void PSD_Parameters::getPAD(DvObject_var &PSD_PAD, int npa, int nphi, QLabel *percLabl)
{
    double scalemin(1.e30), scalemax(-1.);
    int seq_size = data->seqSize();

	vector <size_t> dim;
	dim.push_back(resolution);  // Number of V-bins
	dim.push_back(npa);         // number of pitch angle bins
	PSD_PAD = new DvObject (0.0, dim, seq_size);
	DvObject_var PAD_FRAC_STD = new DvObject (0.0, dim, seq_size);
	
	// get V scale - all E's in Joules, so vels in m/s
	// centre_PSD already converted to SI
	double max_v=0.;
	double min_v=1.e30;
   	for(int time=0; time<seq_size; time++){
		for(int e=0; e<=maxE; e++)
    	{
        	double v=getVPlus(time, e); //getVPLus sets v to upper edge of vel bin
        	if(v>max_v) max_v=v;
			v=getVMinus(time,e);
			if(v>0.&& v< min_v) min_v=v;
    	}
    }
	if (max_v < min_v){  //shouldn't get to here given above but ...
		printf("Failed to find valid velocity range in PAD\n");
		return;
	}
	
	// set up (_v, _pa, _phi) arrays on which to interpolate;
	// do once outside loop as these don't change with time
	
	// Want final product in energy bins, but interpololation needs to be
	// done in velocity.
	// log-spaced energy bin edges, with centre energies at algebraic centre
	// E0    E1         E2      ...        EN
	// |-----|----------|------ ... -------|
	//    e1      e2        e3      e(N-1)
	double dlogE = (log(V2E(max_v)/V2E(min_v)))/resolution;
	double E_fac = exp(dlogE);
	valarray <double> Eedges(resolution+1);
	valarray <double> _E(resolution);  // E-bin centres
	valarray <double> _v(resolution);  // interpolation pts
	Eedges[0]=V2E(min_v);
	for(int i=1; i<(resolution+1);i++)
	{
		Eedges[i] = E_fac*Eedges[i-1];
		_E[i-1] = 0.5*(Eedges[i]+Eedges[i-1]);
		_v[i-1] = E2V(_E[i-1]);
	}
	
	// pitch angles
	double dpa_rad = _PI/npa;
	valarray <double> _pa(npa);
	for(int j=0; j<npa; j++)
	{
		_pa[j]= (j+0.5)*dpa_rad;
	}
	
	// gyrophase
	double dphi_rad = 2.*_PI/nphi;
	valarray <double> _phi(nphi);
	for(int k=0; k<nphi; k++)
	{
		_phi[k] = (k+0.5)*dphi_rad;
	}
	
	vector <size_t> dim3d;
	dim3d.push_back(resolution);
	dim3d.push_back(npa);
	dim3d.push_back(nphi);
		
	// get PADs
    for(int time=0; time<seq_size; time++)
    {
		int perc = (100 * time) / seq_size;

		if(percLabl) percLabl->setText(QString("%1%").arg(perc));
		qApp->processEvents();	
		
		if(this->stop) {
			seq_size = time;
			continue; // allows clean interrupt (keep what's done)
		}
		
		// fill(_x,_y,_z) of v-space points of PSD in rotated cartesian frame
		// (rotated to be in target science frame, e.g. GSE)
		
		fill_cartesian(time);
        
		// build the teselating tetrahedra

		buildTetra(time);

		// Set up PAD output points. Need to set them up in field-aligned coords
  
    	// Find rotation matrix to make field direction be along +Z
    
    	double rotation[3][3];
		
		double dirn[3];
		size_t atT = time;
		if(field_dir->seqSize() == 1) atT=0;
		
		dirn[0] = field_dir->dbl(atT,0);
		dirn[1] = field_dir->dbl(atT,1);
		dirn[2] = field_dir->dbl(atT,2);
		
    	getRotationMatrix(dirn, rotation);
		// (rotation)(0,0,1) = dirn
		//  Lay down points in pad space r, pa, phi with pa=0 corresponding to dirn
		//  Then transform back into science frame (e.g., GSE), i.e.,
		//  same frame as dirn and PSD (after applying conversion to PSD from
		//  instrument or whatever frame to science frame using GUI rotation
		//  information. I.e.,
		//  v_vec_pad = rot^-1 (v_sci - Vbulk_sci)
		//  so
		//  v_sci = rot(v_vec_pad) + Vbulk_sci

    	if(!tetra.empty())
    	{
        	list<Point3D> points;
			for(int i=0; i<resolution; i++)
        	{
            	for(int j=0; j<npa; j++)
            	{
					for(int k=0; k<nphi; k++)
					{
					// calc cartesion version of pad points
					double _xx = _v[i]*sin(_pa[j])*cos(_phi[k]);
					double _yy = _v[i]*sin(_pa[j])*sin(_phi[k]);
					double _zz = _v[i]*cos(_pa[j]);
					
					//  transform to science frame
					double x=rotation[0][0]*_xx+rotation[0][1]*_yy+rotation[0][2]*_zz+centre_PSD->dbl(time,0);
					double y=rotation[1][0]*_xx+rotation[1][1]*_yy+rotation[1][2]*_zz+centre_PSD->dbl(time,1);
					double z=rotation[2][0]*_xx+rotation[2][1]*_yy+rotation[2][2]*_zz+centre_PSD->dbl(time,2);
					
					Point3D temp(x, y, z, -1.);
					temp.orig_x=i;  // these obviously hold info for pts
					temp.orig_y=j;
					temp.orig_z=k;  // need to hold phi index for later avg
					points.push_front(temp);
					}
            	}
        	}

        	interpolation(points, tetra);
			// Hold interpolated pts in zero-initialised array
			DvObject_var PSD_PAD3D = new DvObject(0.0,dim3d);  

        
        	for(list<Point3D>::iterator iter=points.begin(); iter!=points.end(); iter++)
        	{
				// unpack interpolation into 3D array (v, pa, phi)
				PSD_PAD3D->dbl(0,(int)(iter->orig_x),(int)(iter->orig_y),(int)(iter->orig_z)) = iter->val;
								
				// update scale min/max
				if( iter->val > 0) // not fill
				{
					if( iter->val < scalemin) scalemin= iter->val;
					if( iter->val > scalemax) scalemax= iter->val;					
				}
				
        	}
			// perform average over phi
			for (int i=0; i<resolution; i++)
			{
				for (int j=0; j<npa; j++)
				{
					double value;
					double temp=0.0;
					double temp2=0.0;
					int nfound=0;
					for(int k=0; k<nphi; k++) 
					{
						value =  PSD_PAD3D->dbl(0,i,j,k);
						if(value > 0.0){  // skip fills
							temp +=value;
							temp2 += value*value;
							nfound++;
						}
					}
					if(nfound>0)  // phi avg over good vals; interp fill=-1.
					{
						temp  /= (float)nfound;  //mean f(v)
						temp2 /= (float)nfound;  //mean f^2
						PSD_PAD->dbl(time,i,j)=temp;
						PAD_FRAC_STD->dbl(time,i,j) = sqrt((temp2/(temp*temp)) - 1.0);
					} else {
						PSD_PAD->dbl(time,i,j)=-1.;
						PAD_FRAC_STD->dbl(time,i,j)= -1.;
					}
				}
			}
        }		


	}
	
	//  get xrefs and modify DEPENDS
	
	PSD_PAD->copy_xrefs_from(data);

	DvObject_var unitFill = new DvObject(-1.); // interp returns -1 as fill
	PSD_PAD->change_xref(FILL_VALUE, unitFill ); 
	PSD_PAD->change_xref("UNITS", "s^3/m^6");
	DvString fieldnam = data->getXrefText(FIELDNAM);
	fieldnam +=  "-PAD";
	PSD_PAD->change_xref("FIELDNAM", fieldnam);
	PSD_PAD->change_xref("PARAMETER_TYPE", "Data");
	PSD_PAD->change_xref("SI_CONVERSION", "1>s^3 m^-6");
	PSD_PAD->change_xref("DISPLAYTYPE", "Spectrogram");
	
	vector <size_t> dimr;
	dimr.push_back(resolution);	
	DvObject_var Eaxis = new DvObject(0.0, dimr);
	DvObject_var Edelta = new DvObject(0.0,dimr);
	for(int i=0; i<resolution; i++)
	{
		Eaxis->dbl(i) = _E[i]/1.6021e-19;  // in eV
		Edelta->dbl(i) = (Eedges[i+1] - Eedges[i])/3.2042e-19;
	}
	Eaxis->change_xref(FILL_VALUE, unitFill );
	Eaxis->change_xref("UNITS", "eV");
	fieldnam = "speed";
	Eaxis->change_xref("FIELDNAM", fieldnam);
	Eaxis->change_xref("FRAME", "scalar>na");
	Eaxis->change_xref("PARAMETER_TYPE", "Data");
	Eaxis->change_xref("SI_CONVERSION", "1.6021e-19>J");
	Eaxis->change_xref("LABLAXIS","Energy");
	Eaxis->change_xref("SCALETYP","log");
	
	Edelta->change_xref(FILL_VALUE, unitFill );
	Edelta->change_xref("UNITS", "m s-1");
	fieldnam = "deltaV";
	Edelta->change_xref("FIELDNAM", fieldnam);
	Edelta->change_xref("FRAME", "scalar>na");
	Edelta->change_xref("PARAMETER_TYPE", "Data");
	Edelta->change_xref("SI_CONVERSION", "1.6021e-19>J");

	Eaxis->change_xref("DELTA_PLUS", Edelta);
	Eaxis->change_xref("DELTA_MINUS", Edelta);
 	PSD_PAD->change_xref("DEPEND_1", Eaxis);
	
	valarray <double> pa_vals(npa);
	for(int i=0; i<npa; i++)
	{
		pa_vals[i] = (i+0.5)*180./npa;
	}
	vector <size_t> dimpa;
	dimpa.push_back(npa);
	DvObject_var paaxis = new DvObject(pa_vals, dimpa);
	paaxis->change_xref(FILL_VALUE, unitFill );
	paaxis->change_xref("UNITS", "deg");
	fieldnam = "pitch_angle";
	paaxis->change_xref("FIELDNAM", fieldnam);
	paaxis->change_xref("FRAME", "scalar>na");
	paaxis->change_xref("PARAMETER_TYPE", "Data");
	paaxis->change_xref("SI_CONVERSION", "1>degree");
	paaxis->change_xref("LABLAXIS","pitch angle");
	double padelta = 90./npa;  // half delta
	paaxis->change_xref("DELTA_PLUS", padelta);
	paaxis->change_xref("DELTA_MINUS", padelta);

	PSD_PAD->change_xref("DEPEND_2", paaxis);
	if (PSD_PAD->xref_exists("DEPEND_3")) PSD_PAD->delete_xref("DEPEND_3");
	
	// add PAD_STD
	
	PAD_FRAC_STD->copy_xrefs_from(PSD_PAD);
	fieldnam = PAD_FRAC_STD->getXrefText(FIELDNAM);
	fieldnam += "_normalised STDev";
	PAD_FRAC_STD->change_xref("FIELDNAM",fieldnam);
	PAD_FRAC_STD->change_xref("SI_CONVERSION","1>(unitless)");
	PAD_FRAC_STD->change_xref("UNITS","");
	PAD_FRAC_STD->change_xref("LABLAXIS","Fractional STDev over gyrophase");
	PAD_FRAC_STD->change_xref("FRAME","scalar>na");
	PAD_FRAC_STD->change_xref(FILL_VALUE, unitFill ); 

	
	PSD_PAD->change_xref("STDev_over_gyrophase",PAD_FRAC_STD);
     
	DvObject_var smin = new DvObject(scalemin);
	DvObject_var smax = new DvObject(scalemax);
	
    PSD_PAD->change_xref("SCALEMIN", smin);
    PSD_PAD->change_xref("SCALEMAX", smax);

}


// ------------------

void PSD_Parameters::getMomentsSummed(int flags, DvObject_var &density, DvObject_var &velocity, DvObject_var &pressure, DvObject_var &heat, DvObject_var &temperature, QLabel *percLabl)
{
	vector <size_t> dim3;
	dim3.push_back(3);	// vector

    int seq_size=data->seqSize();
	
    density = new DvObject(0.0, seq_size);
    temperature = new DvObject(0.0, seq_size);

    velocity = new DvObject(0.0, dim3, seq_size);
    heat = new DvObject(0.0, dim3, seq_size);
	
	vector <size_t> dim33;
	dim33.push_back(3);	// tensor
	dim33.push_back(3);	// tensor

    pressure = new DvObject(0.0, dim33, seq_size);
    
    double tot_density, tot_numberflux[3], tot_pressure[3][3], tot_heat[3];
    double v_centre[3], v_centre_norm;
    double val, vol, _phi, _theta;
    bool density_ok, velocity_ok, pressure_ok, heat_ok;
    double w;
	
    
    for(int time=0; time<seq_size; time++)
    {
		int j=0;
		
		int perc = (100 * time) / seq_size;

		if(percLabl) percLabl->setText(QString("%1%").arg(perc));
		qApp->processEvents();	
		
		if(this->stop) {
			seq_size = time;
			continue; // allows clean interrupt (keep what's done)
		}

		fill_cartesian(time);
		
		tot_density=0.; 
		density_ok=false;
        tot_numberflux[0]=tot_numberflux[1]=tot_numberflux[2]=0.; 
		velocity_ok=false;
		
        // Computation of density and velocity
		// loops must be in this order
        for(int phi=0; phi<=maxPhi; phi++)
        {
            for(int theta=0; theta<=maxTheta; theta++)
            {
                for(int e=0; e<=maxE; e++)
                {
                    val = getData(time, e, phi, theta);
                    if(val>=0.)
                    {
						vol = getElementVolume(time, e, phi, theta);
						
						if(vol > 0.){
							w = val*vol;
                        	tot_density += w;

                       	 	density_ok=true;
                        
                       	 	if(flags & MOM_VELOCITY)
                       	 	{
								tot_numberflux[0] += _x[j]*w;
								tot_numberflux[1] += _y[j]*w;
								tot_numberflux[2] += _z[j]*w;
								velocity_ok=true;
                       
                        	}
						}
                    }
					j++;
                }
            }
        }
		
        // re-use j counter
		j=0;

        if(density_ok) density->dbl(time) = tot_density;
        else density->dbl(time) = -1.;
        
        if(flags & MOM_VELOCITY)
        {
            if(velocity_ok && density_ok)
            {
                velocity->dbl(time,0) = tot_numberflux[0]/(density->dbl(time));
                velocity->dbl(time,1) = tot_numberflux[1]/(density->dbl(time));
                velocity->dbl(time,2) = tot_numberflux[2]/(density->dbl(time));
            }
            else
            {
                velocity->dbl(time,0) = -1.e+31;
                velocity->dbl(time,1) = -1.e+31;
                velocity->dbl(time,2) = -1.e+31;
            }
        }

        if(((flags & MOM_PRESSURE) || (flags & MOM_HEATFLUX)) && (flags & MOM_DENSITY) && (flags & MOM_VELOCITY) && velocity->dbl(time,0) > -.9e+31)
        {
            tot_pressure[0][0]=tot_pressure[1][0]=tot_pressure[2][0]=0.;
            tot_pressure[0][1]=tot_pressure[1][1]=tot_pressure[2][1]=0.;
            tot_pressure[0][2]=tot_pressure[1][2]=tot_pressure[2][2]=0.;
            pressure_ok=false;
            
            tot_heat[0]=tot_heat[1]=tot_heat[2]=0.;
            heat_ok=false;

            for(int phi=0; phi<=maxPhi; phi++)
            {
                for(int theta=0; theta<=maxTheta; theta++)
                {
                    for(int e=0; e<=maxE; e++)
                    {
                        val = getData(time, e, phi, theta);

                        if(val>=0.)
                        {
                        	vol = getElementVolume(time, e, phi, theta);
							if (vol > 0.){ 
								w = vol*val;
						
								v_centre[0] = _x[j] - velocity->dbl(time,0);
								v_centre[1] = _y[j] - velocity->dbl(time,1);
								v_centre[2] = _z[j] - velocity->dbl(time,2);
                                
								v_centre_norm=v_centre[0]*v_centre[0]+v_centre[1]*v_centre[1]+v_centre[2]*v_centre[2];
    
								if(flags & MOM_PRESSURE)
								{
									for(int i=0; i<3; i++)
									{
										for(int ii=0; ii<3; ii++)
										{
											tot_pressure[i][ii] += v_centre[i]*v_centre[ii]*w;
										}
									}
									pressure_ok=true;
								}
                                
								if(flags & MOM_HEATFLUX)
								{
									for(int i=0; i<3; i++)
									{
										tot_heat[i] += v_centre_norm*v_centre[i]*w;
									}
									heat_ok=true;
								}
							}
						}
						j++;

                    } // end for e
                } // end for theta
            } // end for phi
            
        } // end computation of pressure and heat

        if(flags & MOM_PRESSURE)
        {
            if(pressure_ok)
            {
                pressure->dbl(time,0,0)=tot_pressure[0][0]*mass;
                pressure->dbl(time,1,0)=tot_pressure[1][0]*mass;
                pressure->dbl(time,2,0)=tot_pressure[2][0]*mass;
                pressure->dbl(time,0,1)=tot_pressure[0][1]*mass;
                pressure->dbl(time,1,1)=tot_pressure[1][1]*mass;
                pressure->dbl(time,2,1)=tot_pressure[2][1]*mass;
                pressure->dbl(time,0,2)=tot_pressure[0][2]*mass;
                pressure->dbl(time,1,2)=tot_pressure[1][2]*mass;
                pressure->dbl(time,2,2)=tot_pressure[2][2]*mass;
                
                temperature->dbl(time) = (pressure->dbl(time,0,0)+pressure->dbl(time,1,1)+pressure->dbl(time,2,2))/(3.*density->dbl(time)*1.3806503e-23);
            }
            else
            {
                pressure->dbl(time,0,0)=-1e+31;
                pressure->dbl(time,1,0)=-1e+31;
                pressure->dbl(time,2,0)=-1e+31;
                pressure->dbl(time,0,1)=-1e+31;
                pressure->dbl(time,1,1)=-1e+31;
                pressure->dbl(time,2,1)=-1e+31;
                pressure->dbl(time,0,2)=-1e+31;
                pressure->dbl(time,1,2)=-1e+31;
                pressure->dbl(time,2,2)=-1e+31;
                temperature->dbl(time)=-1e+31;
            }
        }
        
        if(flags & MOM_HEATFLUX)
        {
            if(heat_ok)
            {
                heat->dbl(time,0)=tot_heat[0]*mass/2.;
                heat->dbl(time,1)=tot_heat[1]*mass/2.;
                heat->dbl(time,2)=tot_heat[2]*mass/2.;
            }
            else
            {
				heat->dbl(time,0)=-1e+31;
                heat->dbl(time,1)=-1e+31;
                heat->dbl(time,2)=-1e+31;
            }
        }

    } // end loop over records (time)

    // Metadata
	DvObject_var unitFill = new DvObject(-1.);
	DvObject_var doubleFill = new DvObject(-1.e31);
    
	DvObject_var tt = data->getDep0();

    if(flags & MOM_DENSITY)
    {
		density->change_xref(DEPEND_0, tt);

		if(density.is_ok()){
       		for(int i=0; i<seq_size; i++)
        	{
            	if(density->dbl(i) != -1.) density->dbl(i) /= 1000000.;
        	}
		}
		
		density->change_xref(FILL_VALUE, unitFill );
        density->change_xref(UNITS, "cm^-3");
        density->change_xref(FIELDNAM, "Density - Sum");
        density->change_xref(FRAME, "scalar>na");
        density->change_xref(PARAMETER_TYPE, "Data");
        density->change_xref(SI_CONVERSION, "1.0e6>m^-3");
        density->change_xref(DISPLAYTYPE, "Time_Series");
    }
    
    if(flags & MOM_VELOCITY)
    {
		velocity->change_xref(DEPEND_0, tt);
		
		if(velocity.is_ok()){
        	for(int i=0; i<seq_size; i++)
       		{
            	if(velocity->dbl(i,0) > -.9e+31)
            	{
                	velocity->dbl(i,0) /= 1000.;
                	velocity->dbl(i,1) /= 1000.;
                	velocity->dbl(i,2) /= 1000.;
            	}
        	}
		}
        velocity->change_xref(UNITS, "km s^-1");
        velocity->change_xref(FIELDNAM, "Bulk Velocity");
		velocity->change_xref(FILL_VALUE, doubleFill );
        velocity->change_xref(SI_CONVERSION, "1.0e3>m s^-1");
        velocity->change_xref(DISPLAYTYPE, "Time_Series");
        velocity->change_xref(FRAME, "vector>gse_xyz");
    }
    
    if(flags & MOM_PRESSURE)
    {
		pressure->change_xref(DEPEND_0, tt);

		if(pressure.is_ok()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if(pressure->dbl(i,0,0) > -.9e+31)
            	{
                	for(int j=0; j<3; j++)
                	{
                    	for(int k=0; k<3; k++)
                    	{
                        	pressure->dbl(i,j,k) *= 1000000000.;
                    	}
                	}
            	}
			}
        }
        
		DvObject_var rep = new DvObject(DvString(""), dim3);
		rep->str(0) = "x";
		rep->str(1) = "y";
		rep->str(2) = "z";
		pressure->change_xref(REPRESENTATION_1, rep);
		pressure->change_xref(REPRESENTATION_2, rep);
		pressure->change_xref(LABEL_1, rep);
		pressure->change_xref(LABEL_2, rep);
		
		
        pressure->change_xref(UNITS, "nPa");
        pressure->change_xref(FIELDNAM, "Pressure Tensor");
		pressure->change_xref(FILL_VALUE, doubleFill );
        pressure->change_xref(SI_CONVERSION, "1.0e-9>Pa");
        pressure->change_xref(DISPLAYTYPE, "Time_Series");
        pressure->change_xref(FRAME, "vector>gse_xyz");
		pressure->change_xref(TENSOR_ORDER, "2");
    }
    
    if((flags & MOM_HEATFLUX) || (flags & MOM_PRESSURE))
    {
		heat->change_xref(DEPEND_0, tt);
		
		if(heat.is_ok()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if(heat->dbl(i,0) > -.9e+31)
            	{
                	heat->dbl(i,0) *= 1000000.;
                	heat->dbl(i,1) *= 1000000.;
                	heat->dbl(i,2) *= 1000000.;
            	}
        	}
		}
		heat->change_xref(UNITS, "microW m^-2");
        heat->change_xref(FIELDNAM, "Heat Flux");
		heat->change_xref(FILL_VALUE, doubleFill );
        heat->change_xref(SI_CONVERSION, "1.0e-6>W m^-2");
        heat->change_xref(DISPLAYTYPE, "Time_Series");
        heat->change_xref(FRAME, "vector>gse_xyz");
        
		temperature->change_xref(DEPEND_0, tt);

		if(temperature.is_ok()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if(temperature->dbl(i) > -.9e+31) temperature->dbl(i) /= 1000000.;
        	}
		}
		temperature->change_xref(FILL_VALUE, doubleFill );
        temperature->change_xref(UNITS, "MK");
        temperature->change_xref(FIELDNAM, "Temperature");
        temperature->change_xref(FRAME, "scalar>na");
        temperature->change_xref(PARAMETER_TYPE, "Data");
        temperature->change_xref(SI_CONVERSION, "1e+6>K");
        temperature->change_xref(DISPLAYTYPE, "Time_Series");
    }

}

// ------------------

void PSD_Parameters::getMomentsIntegrated(int flags, DvObject_var &density, DvObject_var &velocity, DvObject_var &pressure, DvObject_var &heat, DvObject_var &temperature, QLabel *percLabl)
{
	vector <size_t> dim3;
	dim3.push_back(3);	// vector

    int seq_size=data->seqSize();
	
    density = new DvObject(0.0, seq_size);
    temperature = new DvObject(0.0, seq_size);

    velocity = new DvObject(0.0, dim3, seq_size);
    heat = new DvObject(0.0, dim3, seq_size);
	
	vector <size_t> dim33;
	dim33.push_back(3);	// tensor
	dim33.push_back(3);	// tensor

    pressure = new DvObject(0.0, dim33, seq_size);
    
    double tot_density, tot_numberflux[3], tot_pressure[3][3], tot_heat[3];
    double v_centre[3], v_centre_norm;
    double val, vol, x, y, z, w;
    
    for(int time=0; time<seq_size; time++)
    {
		int perc = (100 * time) / seq_size;

		if(percLabl) percLabl->setText(QString("%1%").arg(perc));
		qApp->processEvents();	
		
		if(this->stop) {
			seq_size = time;
			continue; // allows clean interrupt (keep what's done)
		}
		fill_cartesian(time);
        
        tot_density=0.;
        tot_numberflux[0]=tot_numberflux[1]=tot_numberflux[2]=0.; 
        
        // Computation of density and velocity
       

		buildTetra(time);
    		
		// Minimum velocity, to avoid integrating in the (undefined) centre of the space
		// Ignore tetra that have only VMinus bin edges at apexes
		
        double min_v=-1.;
        
        for(int e=0; e<=maxE; e++)
        {
            if(getVMinus(time, e)>0.)
            {
                if(min_v < 0.) min_v = _vel[e];
                else
                {
                    if(min_v > _vel[e]) min_v=_vel[e];
                }
            }
        }
        
        min_v=min_v*min_v*0.999; // slightly below data point for > test (avoids using == on doubles)


        if(!tetra.empty())
        {
            for(list<Element3D>::iterator iter=tetra.begin(); iter!=tetra.end(); iter++)
            {
				// At least one apex of tetrahedron must be above min_v (otherwise all on inner VMinus)
                if(	iter->dist1 > min_v ||
                	iter->dist2 > min_v ||
                	iter->dist3 > min_v ||
                	iter->dist4 > min_v  ) 
                {
				
				val=(iter->val[0]+iter->val[1]+iter->val[2]+iter->val[3]) * 0.25;
				vol=iter->volume();
				w = val*vol;
					
				tot_density += w;

				x=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+iter->points[3][0]);

				y=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+iter->points[3][1]);

				z=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+iter->points[3][2]);

				tot_numberflux[0] += x*w;
				tot_numberflux[1] += y*w;
				tot_numberflux[2] += z*w;
              
				}
				
            }

			tot_numberflux[0] *= 0.25;
			tot_numberflux[1] *= 0.25;
			tot_numberflux[2] *= 0.25;
			
            
            density->dbl(time) = tot_density;
            
            if(flags & MOM_VELOCITY)
            {
                velocity->dbl(time,0) = tot_numberflux[0]/(density->dbl(time));
                velocity->dbl(time,1) = tot_numberflux[1]/(density->dbl(time));
                velocity->dbl(time,2) = tot_numberflux[2]/(density->dbl(time));
            }
        }
        else
        {
            density->dbl(time) = -1;
            velocity->dbl(time,0) = -1e+31;
            velocity->dbl(time,1) = -1e+31;
            velocity->dbl(time,2) = -1e+31;
        }
        
        
        if(((flags & MOM_PRESSURE) || (flags & MOM_HEATFLUX)) && (flags & MOM_DENSITY) && (flags & MOM_VELOCITY) && velocity->dbl(time,0)!=-1.e+31)
        {
            tot_pressure[0][0]=tot_pressure[1][0]=tot_pressure[2][0]=0.;
            tot_pressure[0][1]=tot_pressure[1][1]=tot_pressure[2][1]=0.;
            tot_pressure[0][2]=tot_pressure[1][2]=tot_pressure[2][2]=0.;
            
            tot_heat[0]=tot_heat[1]=tot_heat[2]=0.;
            
            for(list<Element3D>::iterator iter=tetra.begin(); iter!=tetra.end(); iter++)
            {
                // At least one apex of tetrahedron must be above min_v (otherwise all on inner VMinus)
                if(	iter->dist1 > min_v ||
                	iter->dist2 > min_v ||
                	iter->dist3 > min_v ||
                	iter->dist4 > min_v  ) 
                {

					val=(iter->val[0]+iter->val[1]+iter->val[2]+iter->val[3]) * 0.25;
					vol=iter->volume();
					w = vol*val;
					
					x=(iter->points[0][0]+iter->points[1][0]+iter->points[2][0]+iter->points[3][0]) * 0.25;

					y=(iter->points[0][1]+iter->points[1][1]+iter->points[2][1]+iter->points[3][1]) * 0.25;

					z=(iter->points[0][2]+iter->points[1][2]+iter->points[2][2]+iter->points[3][2]) * 0.25;
                    
					v_centre[0]=x-velocity->dbl(time,0);
					v_centre[1]=y-velocity->dbl(time,1);
					v_centre[2]=z-velocity->dbl(time,2);
                    
					v_centre_norm=v_centre[0]*v_centre[0]+v_centre[1]*v_centre[1]+v_centre[2]*v_centre[2];
                    
					if(flags & MOM_PRESSURE)
					{
						for(int i=0; i<3; i++)
						{
							for(int j=0; j<3; j++)
							{
								tot_pressure[i][j] += v_centre[i]*v_centre[j]*w;
							}
						}
					}
                            
					if(flags & MOM_HEATFLUX)
					{
						for(int i=0; i<3; i++)
						{
							tot_heat[i] += v_centre_norm*v_centre[i]*w;
						}
					}
				} // end if on min tetra
                
            }
        } // end computation of pressure and heat
        
        if(flags & MOM_PRESSURE)
        {
            pressure->dbl(time,0,0) = tot_pressure[0][0]*mass;
            pressure->dbl(time,1,0) = tot_pressure[1][0]*mass;
            pressure->dbl(time,2,0) = tot_pressure[2][0]*mass;
            pressure->dbl(time,0,1) = tot_pressure[0][1]*mass;
            pressure->dbl(time,1,1) = tot_pressure[1][1]*mass;
            pressure->dbl(time,2,1) = tot_pressure[2][1]*mass;
            pressure->dbl(time,0,2) = tot_pressure[0][2]*mass;
            pressure->dbl(time,1,2) = tot_pressure[1][2]*mass;
            pressure->dbl(time,2,2) = tot_pressure[2][2]*mass;
            
            temperature->dbl(time) = (pressure->dbl(time,0,0)+pressure->dbl(time,1,1)+pressure->dbl(time,2,2))/(3.*density->dbl(time)*1.3806503e-23);
        }
        else
        {

			pressure->dbl(time,0,0)=-1e+31;
			pressure->dbl(time,1,0)=-1e+31;
			pressure->dbl(time,2,0)=-1e+31;
			pressure->dbl(time,0,1)=-1e+31;
			pressure->dbl(time,1,1)=-1e+31;
			pressure->dbl(time,2,1)=-1e+31;
			pressure->dbl(time,0,2)=-1e+31;
			pressure->dbl(time,1,2)=-1e+31;
			pressure->dbl(time,2,2)=-1e+31;
			temperature->dbl(time) = -1.e+31;
        }
        
        if(flags & MOM_HEATFLUX)
        {
            heat->dbl(time,0) = tot_heat[0]*mass*0.5;
            heat->dbl(time,1) = tot_heat[1]*mass*0.5;
            heat->dbl(time,2) = tot_heat[2]*mass*0.5;
        }
        else
        {
            heat->dbl(time,0) = -1.e+31;
            heat->dbl(time,1) = -1.e+31;
            heat->dbl(time,2) = -1.e+31;
        }
    }

    // Metadata
	DvObject_var unitFill = new DvObject(-1.);
	DvObject_var doubleFill = new DvObject(-1.e31);
	DvObject_var tt = data->getDep0();
    
    if(flags & MOM_DENSITY)
    {
		density->change_xref(DEPEND_0, tt);
		
		if(density.is_ok()){
			for(int i=0; i<seq_size; i++)
        	{
            	if(density->dbl(i) != -1.) density->dbl(i) /= 1000000.;
        	}
			density->change_xref(FILL_VALUE, unitFill );
        	density->change_xref("UNITS", "cm^-3");
        	density->change_xref("FIELDNAM", "Density - Tetra");
        	density->change_xref("FRAME", "scalar>na");
        	density->change_xref("PARAMETER_TYPE", "Data");
        	density->change_xref("SI_CONVERSION", "1.0e6>m^-3");
        	density->change_xref("DISPLAYTYPE", "Time_Series");
		}
    }
    
    if(flags & MOM_VELOCITY)
    {
		velocity->change_xref(DEPEND_0, tt);
		
		if(velocity.is_ok()){

        	for(int i=0; i<seq_size; i++)
			{
            	if(velocity->dbl(i,0) > -.9e+31)
            	{
                	velocity->dbl(i,0) /= 1000.;
                	velocity->dbl(i,1) /= 1000.;
                	velocity->dbl(i,2) /= 1000.;
            	}
        	}
        	velocity->change_xref("UNITS", "km s^-1");
        	velocity->change_xref("FIELDNAM", "Bulk Velocity");
			velocity->change_xref(FILL_VALUE, doubleFill );
        	velocity->change_xref("SI_CONVERSION", "1.0e3>m s^-1");
        	velocity->change_xref("DISPLAYTYPE", "Time_Series");
        	velocity->change_xref("FRAME", "vector>gse_xyz");
		}
    }
    
    if(flags & MOM_PRESSURE)
    {	
		pressure->change_xref(DEPEND_0, tt);

		if(pressure.is_ok()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if(pressure->dbl(i,0,0) > -.9e+31)
            	{
                	for(int j=0; j<3; j++)
                	{
                    	for(int k=0; k<3; k++)
                    	{
                       		pressure->dbl(i,j,k) *= 1000000000.;
                    	}
                	}
            	}
        	}
        
			DvObject_var rep = new DvObject(DvString(""), dim3);
			rep->str(0) = "x";
			rep->str(1) = "y";
			rep->str(2) = "z";
			pressure->change_xref(REPRESENTATION_1, rep);
			pressure->change_xref(REPRESENTATION_2, rep);
			pressure->change_xref(LABEL_1, rep);
			pressure->change_xref(LABEL_2, rep);
						
			pressure->change_xref(UNITS, "nPa");
			pressure->change_xref(FIELDNAM, "Pressure Tensor");
			pressure->change_xref(FILL_VALUE, doubleFill );
			pressure->change_xref(SI_CONVERSION, "1.eE-9>Pa");
			pressure->change_xref(DISPLAYTYPE, "Time_Series");
			pressure->change_xref(FRAME, "vector>gse_xyz");
			pressure->change_xref(TENSOR_ORDER, "2");
		}
    }
    
    if(flags & MOM_HEATFLUX || (flags & MOM_PRESSURE))
    {
		heat->change_xref(DEPEND_0, tt);
        
		if(heat.is_ok()){
        	for(int i=0; i<seq_size; i++)
        	{
           		if(heat->dbl(i,0) > -.9e+31)
            	{
                	heat->dbl(i,0) *= 1000000.;
                	heat->dbl(i,1) *= 1000000.;
                	heat->dbl(i,2) *= 1000000.;
            	}
        	}
        	heat->change_xref("UNITS", "microW m^-2");
        	heat->change_xref("FIELDNAM", "Heat Flux");
			heat->change_xref(FILL_VALUE, doubleFill );
        	heat->change_xref("SI_CONVERSION", "1.0e-6>W m^-2");
        	heat->change_xref("DISPLAYTYPE", "Time_Series");
        	heat->change_xref("FRAME", "vector>gse_xyz");
        }
		
		temperature->change_xref(DEPEND_0, tt);

		if(temperature.is_ok()){
        	for(int i=0; i<seq_size; i++)
        	{
            	if(temperature->dbl(i) > -.9e+31) temperature->dbl(i) /= 1000000.;
        	}
			temperature->change_xref(FILL_VALUE, doubleFill );
        	temperature->change_xref("UNITS", "MK");
        	temperature->change_xref("FIELDNAM", "Temperature");
        	temperature->change_xref("FRAME", "scalar>na");
        	temperature->change_xref("PARAMETER_TYPE", "Data");
        	temperature->change_xref("SI_CONVERSION", "1e+6>K");
        	temperature->change_xref("DISPLAYTYPE", "Time_Series");
		}
    }


}


