/* shockDistance.cc Jeremy Mitchell 28/08/2009
Uses qbowshock (sjs) routines 
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "qbowshock.h"

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include "Xrefs.h"

extern "C" QplugReturnStatus DvshockDistance(vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;

QplugReturnStatus DvshockDistance(vector <DvObject_var>& inputs, vector <DvObject_var>& outputs) {
  
	DvObject_var pos_in = inputs[0];
	DvObject_var sw_vel_in = inputs[1];
	DvObject_var model_in = inputs[2];
  
	if ( pos_in.is_nil() || !pos_in->is_dbl() ) {
    	QplugAppendTextDisplay("shockDistance Error: Can't get posn as doubles");
    	return QPLUG_FAILURE;
	}

	DvString frame = pos_in->getFrame();
	if( !frame.empty() && frame != "gse"){
		QplugAppendTextDisplay("shockDistance Error: position must be in GSE\n");
		return QPLUG_FAILURE;
	}
	
	pos_in->clearErr(); // ensure no inherited errors
	DvObject_var pos_Re = pos_in->changeUnitsTo("6.3712e6>m", "Re");  
  
	if (pos_Re->nError() > 0 ) {
    	QplugAppendTextDisplay("shockDistance Error: Can't convert units to Re");
    	QplugAppendTextDisplay( pos_Re->getError(0).c_str() );
    	return QPLUG_FAILURE;
	}
  
	if (!pos_Re->isThreeVector()) {
    	QplugAppendTextDisplay("shockDistance Error: Distance is not a vector");
    	return QPLUG_FAILURE;
	}
  
  
	if (sw_vel_in.is_nil() || sw_vel_in->arraySize() > 1) {
    	QplugAppendTextDisplay("shockDistance Error: Unable to unpack SW speed as scalar\n");
    	return QPLUG_FAILURE;
	}
  
	// convert sw_vel from m to km if necessary
  
	sw_vel_in = sw_vel_in->changeUnitsTo("1.0e3>m s^-1", "m/s");

	float vel_sw = (float)sw_vel_in->asDouble();
	if (vel_sw < 0) {
    	QplugAppendTextDisplay("shockDistance Error: SW vel < 0\n");
    	return QPLUG_FAILURE;
	}
  
	if (model_in.is_nil() || model_in->totalSize() > 1) {
    	QplugAppendTextDisplay("shockDistance Error: Unable to unpack model\n");
    	return QPLUG_FAILURE;
	}
	int model = (int)model_in->asInt();
	if ((model < 0) || (model > 13)) {
    	QplugAppendTextDisplay("shockDistance Error: Must have 0 <= model <= 13\n");
    	return QPLUG_FAILURE;
	}
  
  
	QBSMP_models model_arg = (QBSMP_models) model;
  
	vector <size_t> dims;
	dims.push_back(3);
	
	int size = pos_Re->seqSize();
	
	DvObject_var shock_out = new DvObject((double)0.0, dims, size);
	DvObject_var normal_out = new DvObject((double)0.0, dims, size);
      
    
    for (int i=0; i < size; i++) {
      
		float pos[3], dist, r_out[3], n_out[3];
		pos[0] = pos_Re->dbl(i,0); pos[1] = pos_Re->dbl(i,1); pos[2] = pos_Re->dbl(i,2);
      
		QBSDistanceAlongNormal(pos, vel_sw, model_arg, 1., &dist, r_out);
		QBSNormalAtSurfaceGSE (pos, vel_sw, model_arg, n_out);

      
		shock_out->dbl(i,0) = r_out[0] - (float)(pos_Re->dbl(i,0));
		shock_out->dbl(i,1) = r_out[1] - (float)(pos_Re->dbl(i,1));
		shock_out->dbl(i,2) = r_out[2] - (float)(pos_Re->dbl(i,2));
      
		normal_out->dbl(i,0) = n_out[0];
		normal_out->dbl(i,1) = n_out[1];
		normal_out->dbl(i,2) = n_out[2];   
      
    }
  
	//********************** Set Attributes *************************

	DvObject_var doubleFill = new DvObject(1.e30);
  
	shock_out->change_xref(FRAME, "vector>gse_xyz");
	normal_out->change_xref(FRAME, "vector>gse_xyz");
	shock_out->change_xref(SI_CONVERSION, "6.3712e6>m");
	normal_out->change_xref(SI_CONVERSION, "1>(unit vector)");
	shock_out->change_xref(UNITS, "Re");
	normal_out->change_xref(UNITS, "");
	shock_out->change_xref(LABLAXIS, "r_GSE");
	shock_out->change_xref(VALIDMIN, "-1.e30");
	shock_out->change_xref(VALIDMAX,  "1.e30");
	shock_out->change_xref(FILL_VALUE, doubleFill );
	shock_out->change_xref(SCALETYP,  "lin");
  
 	// Add timeline (or DEPEND_0) : Either timeline is safe if it exists
	DvObject_var qd0 = pos_in->getDep0();
	if ( qd0.is_ok() ){
    	shock_out->change_xref(DEPEND_0 , qd0 );
		normal_out->change_xref(DEPEND_0 , qd0 );
	}
  
	outputs.push_back( shock_out );
	outputs.push_back( normal_out );
  
	return QPLUG_SUCCESS;
  
}
