/****************************************************************************
							LICENSE

  sf_ChampToGeo - plug-in for QSAS software.
  Copyright (C) 2010 - 2012 Eugen Sorbalo

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  A copy of the GNU General Public License is available on the internet at:

  http://www.gnu.org/copyleft/gpl.html

  or you can write to:

  The Free Software Foundation, Inc.
  675 Mass Ave
  Cambridge, MA 02139, USA
****************************************************************************/

/**
 * \file sf_ChampToGeo.cc
 *
 * \brief Converts variable from Champ's CDF into a vector in GSE frame.
 *
 *
 */

#include <fstream>
#include <iostream>
//the next line is needed for visual studio to find M_PI and others.
#define _USE_MATH_DEFINES
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include "qrtn.h"


using namespace QSAS;
using namespace std;


// ---------------------------------------------------------

const short kScalarRank = 1; ///compare nr_coordinates with this to decided if it is a scalar
const short kVectorRank = 3; ///compare nr_coordinates with this to decided if it is a vector

// -------------------------------------------------

QplugReturnStatus  ConvertToDesiredFrame (DvObject_var &ts_in, int frame, DvObject_var &ts_out);


extern "C" QplugReturnStatus Dvsf_ChampToGeo(vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{
	//INPUT 0	SCALAR_TIME_SERIES geo_alt,
	//INPUT 1	SCALAR_TIME_SERIES geo_lat,
	//INPUT 2	SCALAR_TIME_SERIES geo_lon,
	//OUTPUT	VECTOR_TIME_SERIES geo_pos)

	QplugReturnStatus status = QPLUG_SUCCESS;
		
	double ref_radius = 6371.2;

	//get altitude

	DvObject_var geo_alt = inputs[0];
	
	if ( geo_alt.is_nil() || geo_alt->arraySize() != 1) {
    	QplugAppendTextDisplay("ChampToGeo Error: Could not understand altitude time series.\n");
    	return QPLUG_FAILURE;
	}
	
	DvString kRefRadiusAttribute = "REFRADIUS";
	if (geo_alt->xref_exists(kRefRadiusAttribute)) {
		DvObject_var refRadPtr = geo_alt->get_xref(kRefRadiusAttribute);
		ref_radius = refRadPtr->asDouble(0);
		if (ref_radius < 1.0) {
			QplugAppendTextDisplay("ChampToGeo warning: reference radius is either negative or almost zero.\n");
		}
	}
	else {
    	QplugAppendTextDisplay("ChampToGeo Warning: reference radius was not found in meta-data. Using r = 6371.2 km.\n");
	}
	size_t nr_measurements = geo_alt->seqSize();

	//get latitude
		
	DvObject_var geo_lat = inputs[1];
	if ( geo_lat.is_nil() || geo_lat->arraySize() != 1) {
    	QplugAppendTextDisplay("ChampToGeo Error: Could not understand latitude time series.\n");
    	return QPLUG_FAILURE;
	}
		
	if (nr_measurements != geo_lat->seqSize()) {
    	QplugAppendTextDisplay("ChampToGeo Error: Latitude  time tags disagree with those of altitude.\n");						
		return QPLUG_FAILURE;
	}

	//get longitude
	
	DvObject_var geo_lon = inputs[2];
	if ( geo_lon.is_nil() || geo_lon->arraySize() != 1) {
    	QplugAppendTextDisplay("ChampToGeo Error: Could not understand longitude time series.\n");
    	return QPLUG_FAILURE;
	}
	
	if (nr_measurements != geo_lon->seqSize()) {
    	QplugAppendTextDisplay("ChampToGeo Error: Longitude time tags disagree with those of altitude.\n");						
		return QPLUG_FAILURE;
	}

	//get NEC vector
	
	DvObject_var vec_nec_frame = inputs[3];
	if (vec_nec_frame.is_nil())
    	QplugAppendTextDisplay("ChampToGeo: No optional vector sequence in the NEC frame detected.\n");						

	//Get the reference frame of the output vector.
	
	int output_ref_frame = 2;  // GEOC frame
	DvObject_var fr_ptr = inputs[4];
	if ( fr_ptr.is_nil() || fr_ptr->totalSize() != 1) {
    	QplugAppendTextDisplay("calcBShock Warning: frame argument not an integer flag!\n");
	}
	else output_ref_frame = fr_ptr->asInt(0);
	

	// Compute the position vector  -----------------------------------------
	QplugAppendTextDisplay("calcBShock: Processing the data...\n");

	vector <size_t> dims;
	dims.push_back(3);
	DvObject_var champ_pos = new DvObject((double)0.0, dims, nr_measurements);
	for (size_t t = 0; t < nr_measurements; t++) {
		double radius = geo_alt->asDouble(t) + ref_radius;
		double rcos = radius * cos( geo_lat->asDouble(t) / 180. * M_PI);
		champ_pos->dbl(t,0) = rcos * cos(geo_lon->asDouble(t) / 180. * M_PI);
		champ_pos->dbl(t,1) = rcos * sin(geo_lon->asDouble(t) / 180. * M_PI);
		champ_pos->dbl(t,2) = radius * sin(geo_lat->asDouble(t) / 180. * M_PI);
	}

	// Convert the vector frome NEC to GEOC frame  --------------------------
	DvObject_var vec_geoc_frame;


	if( vec_nec_frame.is_ok() ) {
		vec_geoc_frame = new DvObject((double)0.0, dims, nr_measurements);
		double Bsph[3];  //frame defined as: [in colatitude direction; radially; eastward]
		double colat = -1, lon = -1, sin_col = -1, sin_lon = -1, cos_col = -1, cos_lon = -1;
		for (size_t t = 0; t < nr_measurements; ++t) {
			Bsph[0] = - vec_nec_frame->asDouble(t,(size_t)2);
			Bsph[1] = - vec_nec_frame->asDouble(t,(size_t)0);
			Bsph[2] =   vec_nec_frame->asDouble(t,(size_t)1);
			colat = M_PI/2. - geo_lat->asDouble(t) / 180. * M_PI;
			sin_col = sin(colat);
			cos_col = cos(colat);
			lon = geo_lon->asDouble(t) / 180. * M_PI;
			sin_lon = sin(lon);
			cos_lon = cos(lon);
			vec_geoc_frame->dbl(t,0) = Bsph[0] * sin_col * cos_lon  +  Bsph[1] * cos_col * cos_lon  -  Bsph[2] * sin_lon;
			vec_geoc_frame->dbl(t,1) = Bsph[0] * sin_col * sin_lon  +  Bsph[1] * cos_col * sin_lon  +  Bsph[2] * cos_lon;
			vec_geoc_frame->dbl(t,2) = Bsph[0] * cos_col            -  Bsph[1] * sin_col;
		}
	}

	
	// Preparing output champ pos ---------------------------------------------------
	if (champ_pos.is_ok()) {
		champ_pos->change_xref(FRAME, "vector>geoc_xyz");
		champ_pos->change_xref(SI_CONVERSION, "1.0E3>m");
		champ_pos->change_xref(UNITS, "km");
		champ_pos->change_xref(LABLAXIS, "Champ position");
		champ_pos->change_xref(FIELDNAM, "Champ position");
		DvObject_var tt = geo_alt->getDep0();
		champ_pos->change_xref(DEPEND_0, tt);
	}

	if (vec_geoc_frame.is_ok() && vec_nec_frame.is_ok()) {
		vec_geoc_frame->copy_xrefs_from(vec_nec_frame);
		vec_geoc_frame->change_xref(FRAME, "vector>geoc_xyz");
		DvObject_var tt = vec_nec_frame->getDep0();
		vec_geoc_frame->change_xref(DEPEND_0, tt);
	}

	// Convert from GEOC to the desired frame -------------------------------
	QplugReturnStatus sub_status = QPLUG_SUCCESS;
	DvObject_var champ_pos_new_frame;
	DvObject_var vec_new_frame;
	if (output_ref_frame == 2) {
		//no need for conversion
		sub_status = QPLUG_FAILURE;
	}
	else {
		sub_status = ConvertToDesiredFrame (champ_pos, output_ref_frame, champ_pos_new_frame);
		if ((sub_status != QPLUG_FAILURE) && vec_geoc_frame.is_ok())
			sub_status = ConvertToDesiredFrame (vec_geoc_frame, output_ref_frame, vec_new_frame);
		if (sub_status == QPLUG_FAILURE) {
			QplugAppendTextDisplay("ChampToGeo:Warning: Could not convert to the desired frame. Output will be in GEOC frame.\n");
		}
	}

	// Output.  -------------------------------------------------------------
	if (sub_status == QPLUG_SUCCESS) {
		outputs.push_back(champ_pos_new_frame);
		outputs.push_back(vec_new_frame);
	}
	else {
		outputs.push_back(champ_pos);
		outputs.push_back(vec_geoc_frame);
	}
	
	QplugAppendTextDisplay("ChampToGeo: Finished.\n");
	return status;
}


/**
 * Converts a vector object to the desired new frame.
 * \param[in] ts_in is a vector TS object, which has a well-defined frame
 * \param[in] frame is an integer. Acronyms are set to integer values by #define in qrtn.h
 *        0  GSE,  Geocentric Solar Ecliptic
 *        1  GSM,  Geocentric Solar Magnetic
 *        2  GEOC, Geographic (geocentric)
 *        3  SM,   Solar Magnetic
 *        4  GEI,  Geocentric Inertial
 *        5  MAGD,  Geomagnetic (dipole)
 *        6  HAE,  Heliospheric Aries Ecliptic
 *        7  HEE,  Heliospheric Earth Ecliptic
 *        8  HEEQ, Heliospheric Earth Equatorial
 *        9  GSEQ, Geocentric Solar Equatorial
 * \param[out] ts_out is the converted (rotated) TS object.
 */
QplugReturnStatus ConvertToDesiredFrame (DvObject_var &ts_in, int frame, DvObject_var &ts_out)
{
	QplugReturnStatus	status = QPLUG_SUCCESS;
	RtnVector vec;
	vec.n_recs = ts_in->seqSize();
	vec.in_frame = vec.QrGetFrameFromAttr(ts_in);
	vec.out_frame = frame;
	//       next item in list is int identifying target frame
	//       acronyms are set to integer values by #define in qrtn.h
	//       0  GSE,  Geocentric Solar Ecliptic
	//       1  GSM,  Geocentric Solar Magnetic
	//       2  GEOC, Geographic (geocentric)
	//       3  SM,   Solar Magnetic
	//       4  GEI,  Geocentric Inertial
	//       5  MAGD,  Geomagnetic (dipole)
	//       6  HAE,  Heliospheric Aries Ecliptic
	//       7  HEE,  Heliospheric Earth Ecliptic
	//       8  HEEQ, Heliospheric Earth Equatorial
	//       9  GSEQ, Geocentric Solar Equatorial


	vec.IsPositionVector = vec.QrIsPositionVector(ts_in);
	if(vec.IsPositionVector == 1) {
		vec.IsPositionVector = 0; /* issue warning & continue, delta r rotates */
		if(vec.QrOriginsDiffer(vec.in_frame, vec.out_frame)) {
			QplugAppendTextDisplay
			("WARNING: CoordRotation plug-in provides frame rotation only.\n");
			QplugAppendTextDisplay
			("Transformation is between geocentric and heliocentric frames.\n");
			QplugAppendTextDisplay
			("Position vector coordinate transformations will be WRONG.\n");
			QplugAppendTextDisplay
			("Note: Spacecraft separations will rotate OK.\n");
		}
	}

	ts_out = new DvObject(ts_in);
	vec.ptr = ts_out;

	vec.t = ts_in->getDep0();
	if( vec.t.is_nil() ) {
		QplugAppendTextDisplay ("Cannot find time tags, some frames are time dependent\n");
		return QPLUG_FAILURE;
	}

	// Do rotation
	long check = vec.QrDoRotation();

	ts_out = vec.ptr;
	

	return status;
}



