/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	CorrelFFT.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS CorrelFFT wrapper
*
* PURPOSE:      Wrapper for Fast Fourier Transform.
* 
* DESCRIPTION:  Plug In module checking input parameters and
*               constructing CorrelFFT arrays.
*
* ARGUMENTS:
*    IN:       list of plug in objects containing
*              see CorrelFFT.qtpl file
*   OUT:
*
* GLOBALS:
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
* FUNCTIONS_USED:	four1.c cdft.c
*
* ALGORITHM:  see Press et al. 'Numerical Recipes in C' (1992)
*
* HISTORY: created  20/03/2001 Markus Fraenz 
*
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"

extern "C" void rdft(int n, int isgn, double *a, int *ip, double *w);
extern "C" QplugReturnStatus DvCorrelFFT(vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;


QplugReturnStatus	DvCorrelFFT( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{

// define some C variables

int	        attribute_ok,list_length;
unsigned long	Ndata, Ndata2=2, Ndata4,i,np,nm,Ndatah,icorr=0,iacorr=0;
int             method, *ip_tak,j,j1;

int qar_status;

char str_out[150];
double  *data,*dataG,*dataP,*dataPF,*dataPG,deltat,T,*w_tak,dnm,dum;
double  sumf=0.0,sumg=0.0,dcorr=-1.0,dacorr=1.0;

// define QSAS data objects


DvObject_var out_tag_series;
DvObject_var out_correl_series;
DvObject_var out_freq_series;
DvObject_var out_coh_series;
DvObject_var out_phas_series;

//
// ...first, unpack the input data objects
//
QplugAppendTextDisplay("CorrelFFT 1.0 11 Apr 2001\n");

// ...first item is input Series

DvObject_var ff = inputs[0];

if(ff.is_nil() || ff->arraySize() != 1 || ff->not_dbl() ){
      QplugAppendTextDisplay ("correlFFT>> ERROR - F sequence type not handled\n");
      return QPLUG_FAILURE;  
}

Ndata = ff->seqSize();

DvObject_var gg = inputs[1];

if(gg.is_nil() || gg->arraySize() != 1 || gg->not_dbl()){
      QplugAppendTextDisplay ("correlFFT>> ERROR - G sequence type not handled\n");
      return QPLUG_FAILURE;  
}


// Test validity of inputs...

  if( !(ff->isJoined(*gg)) )   
  {
    QplugAppendTextDisplay ("correlFFT>> input sequences not joined properly\n");
     return QPLUG_FAILURE; 
  } 
  
//	..4th item is option 'Method'
method = inputs[2]->asInt();


//	..extract the time tags from the data series object

DvObject_var d0 = ff->getDep0();

if(d0.is_nil() ) d0 = gg->getDep0();

if(d0.is_ok()) deltat = d0->get_spacing();
else deltat=1.0;
  
if (Ndata <= 1) {
	QplugAppendTextDisplay ("CorrelFFT>> ERROR - Series must have at least 2 data points.\n");
	return QPLUG_FAILURE; }

//      (notice that NR routines start indices with 1!!) 

QplugAppendTextDisplay ("CorrelFFT>> using Ooura CDFT8G.\n");


// zero - padding :
while (Ndata2 < Ndata) Ndata2 *= 2;

Ndatah=Ndata2/2;

data    = new double[Ndata2]; 
dataG    = new double[Ndata2];
dataP    = new double[Ndata2];
dataPF    = new double[Ndata2];
dataPG    = new double[Ndata2];

// initialize Tak data arrays and transform input
ip_tak = new int[(long) sqrt((float) Ndatah) +3];
ip_tak[0]=0;
w_tak = new double[Ndatah];



//first find mean values of original series
for (i=0; i<Ndata; i++) {
 sumf +=  ff->dbl(i);
 sumg +=  gg->dbl(i);
}
sumf/=(double)Ndata;
sumg/=(double)Ndata;


for (i=0; i<Ndata; i++) data[i] =  ff->dbl(i) - sumf;
for (i=Ndata; i<Ndata2; i++) data[i] = 0.0; 
rdft((int) Ndata2, 1, data, ip_tak, w_tak);

for (i=0; i<Ndata; i++) dataG[i] =  gg->dbl(i) - sumg;
for (i=Ndata; i<Ndata2; i++) dataG[i] = 0.0; 
rdft((int) Ndata2, 1, dataG, ip_tak, w_tak);

// now multiply transforms:
for (i =0; i<Ndata2; i+=2) {
 dataP[i]=(data[i]*dataG[i]+data[i+1]*dataG[i+1]);
 dataP[i+1]=(data[i+1]*dataG[i]-data[i]*dataG[i+1]);
}
dataP[1]=data[1]*dataG[1];

//now calculate autocorr for each
for (i =0; i<Ndata2; i+=2) {
 dataPF[i]=(data[i]*data[i]+data[i+1]*data[i+1]);
 dataPF[i+1]=0.0;
}
dataPF[1]=data[1]*data[1];

for (i =0; i<Ndata2; i+=2) {
 dataPG[i]=(dataG[i]*dataG[i]+dataG[i+1]*dataG[i+1]);
 dataPG[i+1]=0.0;
}
dataPG[1]=dataG[1]*dataG[1];

//now calculate freq and coherence radius and phase
if (method > 0) {
 out_freq_series  = new DvObject((double)0.0, Ndatah);
 out_coh_series = new DvObject((double)0.0, Ndatah);
 out_phas_series = new DvObject((double)0.0, Ndatah);
 T  = (double)Ndata2 * deltat;
 for (i =0; i<Ndatah; i++) {
  j=2*i;
  out_freq_series->dbl(i) =  (double)i / T;
  dum = dataP[j]*dataP[j];
  dum /= dataPF[j]*dataPG[j];
  out_coh_series->dbl(i) = sqrt(dum);
  out_phas_series->dbl(i) = atan2(dataP[j+1],dataP[j]);
 }
}

if (method !=1) {
 out_tag_series  = new DvObject((double)0.0, Ndata2);
 out_correl_series = new DvObject((double)0.0, Ndata2);

 //and inverse transform
 rdft((int) Ndata2,-1, dataPF, ip_tak, w_tak);
 rdft((int) Ndata2,-1, dataPG, ip_tak, w_tak);
 rdft((int) Ndata2,-1, dataP, ip_tak, w_tak);

 // to make F_in leading sign(lag) is changed:
 dnm = (double)Ndatah/sqrt(dataPF[0]*dataPG[0]);

 for (i =0; i<Ndatah; i++) {
  np=Ndatah+i;
  out_tag_series->dbl(np) = -(double)i * deltat;
  out_tag_series->dbl(i) = (double)(Ndatah-i) * deltat; 
  out_correl_series->dbl(np) = dataP[i]*dnm/(double)(Ndatah-i);
  out_correl_series->dbl(i) = dataP[i+Ndatah]*dnm/(double)(i+1);
 }

 //determine extrema in interval -N/8,+N/8:
 for (i =3*Ndatah/4; i<5*Ndatah/4; i++) {
  if (out_correl_series->dbl(i)>dcorr) {
    dcorr=out_correl_series->dbl(i);
    icorr=i;
  }
  else if (out_correl_series->dbl(i)<dacorr) {
    dacorr=out_correl_series->dbl(i);
    iacorr=i;
  }
 }
}

//	...set some attributes

if (method !=1) {
  out_correl_series->change_xref("SI_conversion", "1>()");
  out_correl_series->change_xref("LABLAXIS", "CorrelFFT");
  out_correl_series->change_xref("FIELDNAM", "CorrelFFT");
  out_correl_series->change_xref("Frame", "scalar>na");

  out_tag_series->change_xref("SI_conversion", "1>s");
  out_tag_series->change_xref("LABLAXIS", "TimeLag");
  out_tag_series->change_xref("FIELDNAM", "TimeLag");
  out_tag_series->change_xref("UNITS", "s");
}
if (method >0) {
  out_freq_series->change_xref("SI_conversion", "1>s^-1 ");
  out_freq_series->change_xref("LABLAXIS", "frequency");
  out_freq_series->change_xref("FIELDNAM", "frequency");
  out_freq_series->change_xref("UNITS", "Hz");

  DvObject_var SIC = ff->get_xref("SI_conversion");
  out_coh_series->change_xref("SI_conversion", SIC);
  out_coh_series->change_xref("LABLAXIS", "Coherence");
  out_coh_series->change_xref("FIELDNAM", "Coherence");
  DvObject_var units =  ff->get_xref("UNITS");
  out_coh_series->change_xref("UNITS", units);

  out_phas_series->change_xref("SI_conversion", "1>rad");
  out_phas_series->change_xref("LABLAXIS", "Phase");
  out_phas_series->change_xref("FIELDNAM", "Phase");
  out_phas_series->change_xref("UNITS", "rad");
}

//	..some messages to the analysis interface window..

sprintf(str_out, "CorrelFFT>> %ld data points\n", Ndata);
QplugAppendTextDisplay (str_out);
sprintf(str_out, "CorrelFFT>> %ld zero padded data points\n", Ndata2);
QplugAppendTextDisplay (str_out);
if (method !=1) {
 sprintf(str_out, "CorrelFFT>> Max. Correlation %e, at lag %e\n", dcorr,(double) out_tag_series->dbl(icorr));
 QplugAppendTextDisplay (str_out);
 sprintf(str_out, "CorrelFFT>> Max. AntiCorrelation %e, at lag %e\n", dacorr,(double)out_tag_series->dbl(iacorr));
 QplugAppendTextDisplay (str_out);
}
QplugAppendTextDisplay ("CorrelFFT>> finished...\n");

//	..add them to the list

  outputs.push_back( out_tag_series);    
  outputs.push_back( out_correl_series);     
  outputs.push_back( out_freq_series);    
  outputs.push_back( out_coh_series);     
  outputs.push_back( out_phas_series);     


//	..clean up and exit

delete[] data; 
delete[] dataG; 
delete[] dataP;
delete[] dataPF;
delete[] dataPG;
delete ip_tak;
delete w_tak;


return QPLUG_SUCCESS;


}



