/***********************************************************************
 *           See Accompanying licence file QSAS_licence                 *
 ***********************************************************************/
/* PROJECT: Cluster UK CDHF
 *
 * COMPONENT:   CurlDivGrad_Calc
 * 
 * MODULE:	QSAS Manipulation of Vector Quantities
 *
 * LANGUAGE:	C++
 *
 * FUNCTION:
 *
 * PURPOSE:      Calculate the curl, gradient and divergence of a vector quantity
 * 
 * DESCRIPTION:
 *
 * ARGUMENTS:
 *    IN:
 *     VECTOR_TIME_SERIES    V_C1..V_C4           input vector quantity
 *     VECTOR_TIME_SERIES    Pos_C1..Pos_C4       position vectors for the four SC's
 *     TIME_INTERVAL         Time                 time to calculate curl or div for
 * 
 *    OUT:
 *     ANY_DATA_OBJECT       curlV, divV, gradV   results of the calculation;
 *                                                a scalar for div, a vector for curl
 *                                                quality ratio div/curl as Metadata in Curl
 *
 * GLOBALS:
 *
 * RETURN:
 *    TYPE:     QplugReturnStatus
 *    VALUES:   see qplug_if.h
 *
 *
 * ALGORITHM:   See chap 12 & 14 of Blue ISSI book.
 *
 * HISTORY:	SEH, 27 Aug 2003
 *                    26 Dec 2003 - QSAS 2 version
 *                     2 May 2005, flipped curl sign in function CurlVec()
 *               SEH, 21 Feb 2016 - QSAS 3 Version with DVOS
 *
 ***********************************************************************/


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"


extern  "C" QplugReturnStatus  DvCurlDivGrad_Calc  (vector <DvObject_var>&, vector <DvObject_var>&);

double sum_SCs(DvObject_var V1, DvObject_var V2, int compV1, int compV2);
DvObject_var CurlVec(DvObject_var V, DvObject_var R);
DvObject_var GradVec(DvObject_var V, DvObject_var R);
DvObject_var GradScalar(DvObject_var V, DvObject_var R);
double DivergenceVec(DvObject_var V, DvObject_var R);



using namespace QSAS;
QplugReturnStatus  DvCurlDivGrad_Calc  (vector <DvObject_var>& inputs, vector <DvObject_var>& outputs)
{
  if (inputs.size() < 10) {
    QplugAppendTextDisplay ("10 inputs (4 parameters, 4 positions, 1 normalization and 1 time object required\n");
    return QPLUG_FAILURE; 
  }

  // -- first, get time interval
  DvEvent start_end = inputs[9]->getTimeRange();

  // -- get parameters to caclulate gradients from --
  DvObject_var V1 = inputs[0];
  DvObject_var tt_out = V1->getTimeTags();
  DvObject_var V2 = inputs[1]; V2=V2->Join(tt_out);
  DvObject_var V3 = inputs[2]; V3=V3->Join(tt_out);
  DvObject_var V4 = inputs[3]; V4=V4->Join(tt_out);

 
  // -- get SC positions  --
  DvObject_var P1 = inputs[4]; P1 = P1->Join(tt_out);
  DvObject_var P2 = inputs[5]; P2 = P2->Join(tt_out); 
  DvObject_var P3 = inputs[6]; P3 = P3->Join(tt_out);
  DvObject_var P4 = inputs[7]; P4 = P4->Join(tt_out);

  // --- get scalefactor for Curl and check if this is a current (scale = mu_0 so that J = curl B / mu_0) ----
  DvObject_var sfac=inputs[8];
  cout << sfac->asDouble() << "\n";

  long nestsize=V1->seqSize();
  printf("Number of samples vec = %ld\n",nestsize);
  printf("Number of samples pos = %ld\n",P1->seqSize());

  vector <size_t> dims43;
  dims43.push_back(4);
  dims43.push_back(3);
  DvObject_var P = new DvObject(0.0, dims43, nestsize);
  DvObject_var V = new DvObject(0.0, dims43, nestsize);
 

  for (int t=0; t<nestsize; t++) {
    for (int i=0; i<3; i++) {
      V->dbl(t,0,i) = V1->dbl(t,i);
      V->dbl(t,1,i) = V2->dbl(t,i);    
      V->dbl(t,2,i) = V3->dbl(t,i);    
      V->dbl(t,3,i) = V4->dbl(t,i);
      P->dbl(t,0,i) = P1->dbl(t,i);
      P->dbl(t,1,i) = P2->dbl(t,i);
      P->dbl(t,2,i) = P3->dbl(t,i);
      P->dbl(t,3,i) = P4->dbl(t,i);
    }
  }

  double scalefac=1.0;
  bool iscurrent=false;


  if (!sfac.is_nil()) {
    scalefac = (double) sfac->asDouble(); 
    printf("> Curl scaled with %g\n",scalefac);
  }
  else {
    printf("> Curl unscaled\n");
  }

  // --- now, calculate curl, div and grad for each time step -----
  

  int c[]={0,1,2,0,1};
  int i, j, sc, c1=0, c2=1, c3=2, c4=3;
  double totvol=0.0, vol[]={0.0, 0.0, 0.0, 0.0};
 
  DvObject_var R = new DvObject(0.0, DIM33);               // Volumetric tensor - see Blue book sec 12.2.1 + 14.2.1
  DvObject_var K =  new DvObject(0.0, dims43);             // Reciprocalvector - see Blue book, sec 12.4
  DvObject_var Vs =  new DvObject(0.0, dims43);
  DvObject_var Curl =  new DvObject(0.0, DIM3);
  DvObject_var Grad =  new DvObject(0.0, DIM3);

  DvObject_var Curl_TS = new DvObject(0.0, DIM3, nestsize);
  DvObject_var Grad_TS = new DvObject(0.0, DIM3, nestsize);
  DvObject_var GradVec_TS = new DvObject(0.0, DIM33, nestsize);
  DvObject_var Div_TS = new DvObject(0.0, nestsize);
  //  DvObject_var Vol_TS = new DvObject(0.0, nestsize);
  DvObject_var Ratio_TS = new DvObject(0.0, nestsize);
  DvObject_var Bary_TS = new DvObject(0.0, DIM3, nestsize);

 
  for (int t=0; t<nestsize; t++) {                  // Reciprocal vectors according to Eq. (14.7) in Blue Book
    if (t < 6) {
      printf("V1 = [%g %g %g] \n",V->dbl(t,0,0),  V->dbl(t,0,1),  V->dbl(t,0,2));
      printf("P4 = [%g %g %g] \n",P->dbl(t,3,0),  P->dbl(t,3,1),  P->dbl(t,3,2));
    }

    for (i=0; i<3; i++) {
      K->dbl(0,c1,i)  =   (P->dbl(t,c3,c[i+1]) - P->dbl(t,c2,c[i+1])) *  (P->dbl(t,c4,c[i+2]) - P->dbl(t,c2,c[i+2])) - (P->dbl(t,c3,c[i+2]) - P->dbl(t,c2,c[i+2])) * (P->dbl(t,c4,c[i+1]) - P->dbl(t,c2,c[i+1]));
      K->dbl(0,c2,i)  = -((P->dbl(t,c4,c[i+1]) - P->dbl(t,c3,c[i+1])) *  (P->dbl(t,c3,c[i+2]) - P->dbl(t,c1,c[i+2])) - (P->dbl(t,c4,c[i+2]) - P->dbl(t,c3,c[i+2])) * (P->dbl(t,c3,c[i+1]) - P->dbl(t,c1,c[i+1])));
      K->dbl(0,c3,i)  =   (P->dbl(t,c4,c[i+1]) - P->dbl(t,c1,c[i+1])) *  (P->dbl(t,c4,c[i+2]) - P->dbl(t,c2,c[i+2])) - (P->dbl(t,c4,c[i+2]) - P->dbl(t,c1,c[i+2])) * (P->dbl(t,c4,c[i+1]) - P->dbl(t,c2,c[i+1]));
      K->dbl(0,c4,i)  =   (P->dbl(t,c2,c[i+1]) - P->dbl(t,c1,c[i+1])) *  (P->dbl(t,c3,c[i+2]) - P->dbl(t,c1,c[i+2])) - (P->dbl(t,c2,c[i+2]) - P->dbl(t,c1,c[i+2])) * (P->dbl(t,c3,c[i+1]) - P->dbl(t,c1,c[i+1]));
    }			 

    vol[c1] = -(P->dbl(t,c2,0)-P->dbl(t,c1,0)) * K->dbl(0,c1,0) -  (P->dbl(t,c2,1)-P->dbl(t,c1,1)) * K->dbl(0,c1,1) - (P->dbl(t,c2,2)-P->dbl(t,c1,2)) * K->dbl(0,c1,2);
    vol[c2] = -(P->dbl(t,c3,0)-P->dbl(t,c2,0)) * K->dbl(0,c2,0) -  (P->dbl(t,c3,1)-P->dbl(t,c2,1)) * K->dbl(0,c2,1) - (P->dbl(t,c3,2)-P->dbl(t,c2,2)) * K->dbl(0,c2,2);
    vol[c3] = -(P->dbl(t,c4,0)-P->dbl(t,c3,0)) * K->dbl(0,c3,0) -  (P->dbl(t,c4,1)-P->dbl(t,c3,1)) * K->dbl(0,c3,1) - (P->dbl(t,c4,2)-P->dbl(t,c3,2)) * K->dbl(0,c3,2);
    vol[c4] = +(P->dbl(t,c4,0)-P->dbl(t,c1,0)) * K->dbl(0,c4,0) +  (P->dbl(t,c4,1)-P->dbl(t,c1,1)) * K->dbl(0,c4,1) + (P->dbl(t,c4,2)-P->dbl(t,c1,2)) * K->dbl(0,c4,2);

    for (sc=0;sc<4; sc++) {
      for (i=0; i<3; i++) { 	
	Bary_TS->dbl(t,i) += 0.25 * V->dbl(t,sc,i);
	K->dbl(0,sc,i) /= vol[sc];
	Vs->dbl(0,sc,i) = V->dbl(t,sc,i);
	for(j=0; j<3; j++) {
	  R->dbl(0,i,j)  += P->dbl(t,sc,i) * P->dbl(t,sc,j); 													   
	}
	totvol += 0.25*fabs(vol[sc]);
      }
    } 

    Curl =  CurlVec(Vs,K);
    Div_TS->dbl(t) = DivergenceVec(Vs,K);
    Grad =  GradScalar(Vs,K);

    if (t < 6) printf("t = %d , CurlV = [%g %g %g] \n",t,Curl->dbl(t,0), Curl->dbl(t,1), Curl->dbl(t,2));

    for (i=0; i<3; i++) {
      Curl_TS->dbl(t,i) = Curl->dbl(i);
      Grad_TS->dbl(t,i) = Grad->dbl(i);
      //for(int j=0; j<3; j++) GradVec_TS->dbl(t,i,j) =  sum_SCs(Vs,K,i,j);
    }
    // --- for e.g., B-fields, the ratio |div B| / |Curl B| can be useful
    Ratio_TS->dbl(t) = abs(Div_TS->dbl(t)) / Curl->abs()->dbl(0);
  }

  Ratio_TS->change_xref(DEPEND_0, tt_out);
  Ratio_TS->change_xref(CATDESC,"the ratio |div(V)| / |curl(V)|, where V is the vector analyzed");

  

  // --- Meta data; frame and units --------

  DvString units;

  DvObject_var dummy = V1/P1;   // just used to get units
  Curl_TS->copy_xref_from(UNITS, dummy);

  double conv = dummy->convFactorToBaseSI();
  DvString si_conv=dummy->getBaseSI();
  if (!sfac.is_nil()) {
    dummy /= sfac;  // just to get units
    if (sfac->getXrefText("ObjectName").contains("Mu_0")){
      conv = dummy->convFactorToBaseSI();
      si_conv=dummy->getBaseSI();
      printf("Normalize with mu0 -> Amperes law");
      Curl_TS /= new DvObject(scalefac*1e12);   // 1e6  (km^2 -> m^2)
      units = "A/m^2";
      Curl_TS->change_xref(UNITS, units);
    }      
  }
//   cout << "Conversionfac " << conv << "\n";
//   cout << "SI_converesion " << si_conv << "\n";
//   cout << "Units :" << units << "\n";
 
  Curl_TS->copy_xref_from(COORDINATE_SYSTEM,V1);
  Curl_TS->copy_xref_from(REPRESENTATION_1, V1);
  Curl_TS->copy_xref_from(TENSOR_ORDER, V1);
  Curl_TS->change_xref(SI_CONVERSION, si_conv);
  Curl_TS->change_xref("QUALITYRATIO", Ratio_TS);
  Curl_TS->change_xref(DEPEND_0, tt_out);

  units = V1->getUnitsRatio(*P1);
  Grad_TS->copy_xref_from(COORDINATE_SYSTEM,V1);
  Grad_TS->copy_xref_from(TENSOR_ORDER, V1);
  Grad_TS->change_xref(SI_CONVERSION, si_conv);
  Grad_TS->change_xref(DEPEND_0, tt_out);
  Grad_TS->change_xref(UNITS, units);

  Div_TS->change_xref(SI_CONVERSION, si_conv);
  Div_TS->change_xref(DEPEND_0, tt_out);
  Div_TS->change_xref(UNITS,  units);

  outputs.push_back(Curl_TS);
  outputs.push_back(Grad_TS);
  outputs.push_back(Div_TS);
 

  return QPLUG_SUCCESS;
}
 



// -----------------------------------
//
// Summation over vectors
//
// -----------------------------------
double sum_SCs(DvObject_var V1, DvObject_var V2, int compV1, int compV2)
{
  double sum = 0.0;

  for (int i=0; i<4; i++) {
    sum += V1->dbl(0,i,compV1) * V2->dbl(0,i,compV2);
  }
  return (sum);
}



// ------------------------------------------
//
// Curl of a vector (cross prod)
//
// ------------------------------------------
DvObject_var CurlVec(DvObject_var V, DvObject_var R)
{
  DvObject_var Curl = new DvObject(0.0, DIM3);

  Curl->dbl(0) = (-sum_SCs(V,R,1,2) + sum_SCs(V,R,2,1));
  Curl->dbl(1) = (-sum_SCs(V,R,2,0) + sum_SCs(V,R,0,2));
  Curl->dbl(2) = (-sum_SCs(V,R,0,1) + sum_SCs(V,R,1,0));
  return(Curl);
}
 
// ------------------------------------------
//
// Divergence of a vector see Eq. 14.16 in Blue Book
// 
// Inputs
//   V = vector (e.g., B-field, E-field, flow)
//   R = reciprocal vector
// Output
//   returns a scalar (dot product)
//
// ------------------------------------------
double DivergenceVec(DvObject_var V, DvObject_var R)
{
  return(sum_SCs(V,R,0,0) + sum_SCs(V,R,1,1) + sum_SCs(V,R, 2,2));
}




// ------------------------------------------
//
// Gradient of a vector 
// Inputs
//   V = vector (e.g., B-field, E-field, flow)
//   R = reciprocal vector
// Output
//   returns a gradient tensor (matrix)
// ------------------------------------------
DvObject_var GradVec(DvObject_var V, DvObject_var R)
{
  DvObject_var Grad = new DvObject(0.0, DIM33);  

  for (int i=0; i<3; i++) {
    for(int j=0; j<3; j++) {
      Grad->dbl(0,i,j) = sum_SCs(V,R,i,j);
    }
  }
  return(Grad);
}


// ------------------------------------------
//
// Gradient of a scalar 
// Inputs
//   V = vector (e.g., B-field, E-field, flow)
//   R = reciprocal vector
// Output
//   returns a gradient (vector)
// ------------------------------------------
DvObject_var GradScalar(DvObject_var V, DvObject_var R)
{
  vector <size_t> dims41;
  dims41.push_back(4);
  dims41.push_back(1);
 
DvObject_var Grad = new DvObject(0.0, DIM3); 
DvObject_var Magn = new DvObject(0.0, dims41); 
DvObject_var Vec  = new DvObject(0.0, DIM3); 

  for (int i=0; i<4; i++) {
    for (int j=0; j<3; j++) Vec->dbl(j) = V->dbl(0,i,j);
    Magn->dbl(i) = sqrt(Vec->dbl(0)*Vec->dbl(0) + Vec->dbl(1)*Vec->dbl(1) + Vec->dbl(2)*Vec->dbl(2));
  }

  for (int i=0; i<3; i++) Grad->dbl(i) = sum_SCs(Magn,R,0,i);
  return(Grad);
}

