/*****************************************************************
* PROJECT: Cluster UK CDHF
* COMPONENT: Plug In library
* MODULE: Dwt.cc
* LANGUAGE: C++
* FUNCTION: QSAS Dwt Plug In
* PURPOSE: Computes the Discrete Wavelet Transform of a scalar series.
* 	
* ARGUMENTS:  
*    IN: scalar time series.
*	 time interval
*	 transform DWT or MODWT
*	 wavelet filter and its width
*	 level J
*	 plotting device
*    OUT: 
*	 QSAS scalar sequence containing wavelet coefficients for
*	 j=1,...,J and the scaling coefficients at level J.
*
* RETURN: QplugReturnStatus
*
* History: version 1.0 Created Dec. 2001. 
* 	   version 1.1 June 2002.
* A. Serroukh, 
******************************************************************/

#include<cstdio>
#include<cmath>
#include<algorithm> 
#include<string.h>
#include<cstdlib>
#include<fstream>

#include "qplug_if.h"
#include "qdutil.h"
#include "Xrefs.h"

// #include "cpgplot.h"
// #include "PlotWindow.h"
#include "Dwt_set_plot.h"

extern "C" QplugReturnStatus DvDwt_plugin(vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;

/*****************************************************************/
QplugReturnStatus DvDwt_plugin( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs ){

   QplugReturnStatus ret_status;  

   DvObject_var  xx = inputs[0];
   DvObject_var  pad_xx_subseq; // xx with len power of 2
   DvObject_var  Int_in = inputs[1]; 
   DvObject_var  dwt_str = inputs[2];
   DvObject_var  filt_name = inputs[3];
   DvObject_var  flen = inputs[4];
   DvObject_var  J = inputs[5]; // level
   char buf_out[100];

   //  scalar seq.
   if(xx.is_nil() || xx->not_dbl() || xx->arraySize() > 1){
     QplugAppendTextDisplay("DWT Error. Input data type not handled\n");
     return QPLUG_FAILURE;
   }

   //read the time interval
   if(Int_in.is_nil()  ){
     QplugAppendTextDisplay("DWT Error. Input time interval not found\n");
     return QPLUG_FAILURE;
   }
   DvEvent timeInt = Int_in->getTimeRange();
   
   xx = xx->subSequence(timeInt);
   
   //get timetags from xx
   DvObject_var out_tags = xx->getTimeTags();     
   if( out_tags.is_nil() ||  xx->seqSize() < 2){
     QplugAppendTextDisplay("DWT Error. Object is not a time series of at least 2 records\n");
     return QPLUG_FAILURE;
   } 
	
   
   double spacing;

   if( !xx->isRegular(spacing) ){
      QplugAppendTextDisplay("Dwt Error: Irregular time series. \n");
      return QPLUG_FAILURE;
   }   
   
   //read what to run DWT or MODWT
   Yes_or_No DWT_YES_NO = NO;
   Yes_or_No MODWT_YES_NO = YES;
   Yes_or_No PAD_DWT_YES_NO = NO;
   
   if(dwt_str.is_nil() || dwt_str->not_str()){
      QplugAppendTextDisplay("DWT Error: Transform undefined\n");
      return QPLUG_FAILURE;
   } 
   if(dwt_str->str() == "MODWT" || dwt_str->str().empty())  
   		DWT_YES_NO = NO;
   else if(dwt_str->str() ==  "DWT") {
   		DWT_YES_NO = YES;
		MODWT_YES_NO = NO;
   }
   else {     
     	QplugAppendTextDisplay("DWT Error. Wavelet Transform undefined\n");
     	return QPLUG_FAILURE;
   } 
   
   //get the filter name
   if(filt_name.is_nil() || filt_name->not_str()){
     QplugAppendTextDisplay("DWT Error. Filter name not handled\n");
     return QPLUG_FAILURE;
   } 

   //get the filter length
   if(flen.is_nil() || flen->asInt() < 2){
     QplugAppendTextDisplay("DWT Error. Filter not handled\n");
     return QPLUG_FAILURE;
   }
  
   //check whether the filter length is even
   int nf = flen->asInt() / 2;
   if(nf * 2 != flen->asInt()){
      QplugAppendTextDisplay("DWT Error. Filter length is not even.\n");
      return QPLUG_FAILURE; 
   }
   
   int Flength = flen->asInt();
   //read the filter from file /QSAS_HOME/examples/FilterBank/Bank.flt
   DvObject_var gscal = new DvObject(0.0, Flength);
   ret_status = ReadFilter(filt_name->asStr().c_str(), gscal);
   if(ret_status==QPLUG_FAILURE) return QPLUG_FAILURE;

   //rescale twice if MODWT
   if(MODWT_YES_NO == YES) *gscal /= sqrt(2.0);
     
   // level 
   if(J.is_nil() ){
     QplugAppendTextDisplay("Dwt Error: Level not handled\n");
     return QPLUG_FAILURE;
   }
   
   //test if the sample size can be processed
   if( xx->seqSize() <= Flength ){
      QplugAppendTextDisplay("DWT Error. Unvalid sample size.\n");
      return QPLUG_FAILURE; 
   } 
     

   long Nsubsize = out_tags->seqSize();    

   //display sample size on the plug-in window
   sprintf(buf_out,"processed sample size:%ld\n",Nsubsize);
   QplugAppendTextDisplay (buf_out);
   
   if(DWT_YES_NO ==  YES) {
		int jpower = 2;
		// if series length not a power of 2, pad with zeros without changing the size of out_tags.    
		if(isPowerOf2(Nsubsize) == NO){
			PAD_DWT_YES_NO = YES;
			while( Nsubsize > jpower) jpower *= 2;
			pad_xx_subseq = new DvObject(0.0, jpower);
			// copy over values, leaves zero pads at end
			for(int i=0; i<Nsubsize; i++) pad_xx_subseq->dbl(i) = xx->dbl(i);
		}
   }
   
   //work out the maximum level 
   
   double Jmax = (double)Nsubsize - 1.0; 
   Jmax /= flen->asDouble() - 1.0;
   Jmax = log(Jmax + 1.0)/log(2.0);

   //run the full dwt if Jlevel =0 
   int Jlevel = J->asInt();
   if(Jlevel == 0) Jlevel = (int)Jmax;   

   //check whether the required level transform is <= Jmax
   if( Jlevel <0 || Jlevel > (int)Jmax){
     QplugAppendTextDisplay("Dwt Error: DWT level is out of range.\n");
     return QPLUG_FAILURE;
   }
   
   //compute the wavelet filter
   DvObject_var hwav = new DvObject(0.0, Flength);
   
   int lsign = -1;
   for(int l=0; l< Flength; l++){
		lsign = -lsign;
		hwav->dbl(l) = lsign*gscal->dbl(Flength -1 - l); //g[Flen-1-l];
   }       
   
   //define the output data 
   DvObject_var wav_coef;
      if(DWT_YES_NO == YES){
         ret_status = Run_Dwt(pad_xx_subseq, Jlevel, hwav, gscal, wav_coef);
      }
      else {
         ret_status = Run_MoDwt(xx, Jlevel, hwav, gscal, wav_coef); 
      }
     			
   //now apply the dot product to get |W_{j,t}|^2
   wav_coef *= wav_coef;	

   if(wav_coef.is_nil()){
     QplugAppendTextDisplay("Dwt Error: mult operation error\n");
     return QPLUG_FAILURE;
   }
   
   //copy cross references  
   wav_coef->copy_xrefs_from(xx);
   wav_coef->delete_xref(DEPEND_0); 
   
   //setting and defining attributes

   double time_step = out_tags->get_spacing();   
   
   wav_coef->change_xref("TimeInterval", timeInt.getISOstring().c_str()); 
   wav_coef->change_xref("TimeStep", time_step);      


   
   //set the transform type
   DvString transf_name;
   if(DWT_YES_NO == YES) transf_name = "DWT_"; 
   else transf_name = "MODWT_"; 
      
   DvString transf_type = transf_name;   
   if(PAD_DWT_YES_NO == YES) transf_type += "Padded with zeros"; 
   wav_coef->change_xref("Discret_Wav_Transform", transf_type); 
        
   DvString oname = xx->getXrefText("ObjectName");
   if(DWT_YES_NO == YES) oname += "_DwtCoef";
   else oname += "_MODwtCoef";
 
   wav_coef->change_xref("DwtObjectName", oname);
   
   //set the scaling filter name
   DvString scal_filter = filt_name->asStr(); 
   scal_filter += "(";
   scal_filter += Flength;
   scal_filter += ")";
   wav_coef->change_xref("Scaling_Filter", scal_filter);
   
   //set the DWT run level
   DvString dwt_level = "DWT run level: ";
   dwt_level += Jlevel;
   wav_coef->change_xref("DWT_Level", dwt_level);
   
   outputs.push_back(wav_coef);
      
   /*---------------------------------------------------------*/
   /* prepare the plotting of X_t and wav. coef.             */

   PGSpectro_ret  ret_value;
   PGSpectro pgspectro, *pgspectrop;

   pgspectrop = &pgspectro;
      
   DvString tmp_str;
      
   //FIELDNAM 
   tmp_str = transf_name; 
   tmp_str += xx->getXrefText(FIELDNAM);   
   wav_coef->change_xref(FIELDNAM, tmp_str);
   
   //LABLAXIS
   tmp_str = transf_name; 
   tmp_str += xx->getXrefText(LABLAXIS);
   wav_coef->change_xref(LABLAXIS, tmp_str);
   
   //SI_conversion   
   tmp_str = xx->getSICPower(2.0);
   wav_coef->change_xref(SI_CONVERSION, tmp_str);
   
   //UNIT y_label and wedge_label
   tmp_str = xx->getUnitsPower(2.0);
   wav_coef->change_xref(UNITS, tmp_str);


	ret_status = SPectro_set(wav_coef, out_tags, Jlevel, pgspectrop,
	gscal , DWT_YES_NO, "Time ", "Scale(Second)", tmp_str.c_str() );
	

   //ALBAN
   std::cout << pgspectrop->nx << " " << pgspectrop->ny << std::endl;
   
   vector<size_t> dims;
   dims.push_back(pgspectrop->ny);
   DvObject_var Spectro_return = new DvObject(0.0, dims, pgspectrop->nx);
   for(int i=0; i<pgspectrop->nx; ++i)
   {
	   for(int j=0; j<pgspectrop->ny; ++j)
	   {
		   Spectro_return->dbl(i,j) = pgspectrop->z[i*pgspectrop->ny+j];
	   }
   }
  
   
   DvObject_var spectro_depend_1 = new DvObject(0.0, dims);
   DvObject_var spectro_d1_delta_plus=new DvObject(0.0, dims);
   DvObject_var spectro_d1_delta_minus=new DvObject(0.0, dims);
   for(int i=0; i<pgspectrop->ny; ++i)
   {
	   spectro_depend_1->dbl(i) = (pgspectrop->ybinlow[i]+pgspectrop->ybinhigh[i])/2.;
	   spectro_d1_delta_plus->dbl(i) = (pgspectrop->ybinhigh[i]-pgspectrop->ybinlow[i])/2.;
	   spectro_d1_delta_minus->dbl(i) = spectro_d1_delta_plus->dbl(i);
   }

   spectro_depend_1->change_xref(UNITS, "s");
   spectro_depend_1->change_xref(LABLAXIS, "Wavelet scale");
   spectro_depend_1->change_xref(SI_CONVERSION, "1>s");
   spectro_depend_1->change_xref(SCALETYP, "Log");
   
   spectro_d1_delta_plus->change_xref(UNITS, "s");
   spectro_d1_delta_plus->change_xref(SI_CONVERSION, "1>s");
   
   spectro_d1_delta_minus->change_xref(UNITS, "s");
   spectro_d1_delta_minus->change_xref(SI_CONVERSION, "1>s");
   
   DvString wunits = wav_coef->getXrefText(UNITS);
   Spectro_return->change_xref(UNITS, wunits);
   DvString wsic = wav_coef->getXrefText(SI_CONVERSION);
   Spectro_return->change_xref(SI_CONVERSION, wsic);
   
   Spectro_return->change_xref(DEPEND_0, out_tags);
   
   spectro_depend_1->change_xref(DELTA_PLUS, spectro_d1_delta_plus);
   spectro_depend_1->change_xref(DELTA_MINUS, spectro_d1_delta_minus);
   Spectro_return->change_xref(DEPEND_1, spectro_depend_1);
   
   DvObject_var sFill = new DvObject(-1.e+31);
   Spectro_return->change_xref(FILL_VALUE, sFill);
   Spectro_return->change_xref(SCALETYP,"Log");
   
   DvObject_var smin = new DvObject((double)pgspectrop->zscalemin);
   DvObject_var smax = new DvObject((double)pgspectrop->zscalemax);
   Spectro_return->change_xref(SCALEMIN, smin);
   Spectro_return->change_xref(SCALEMAX, smax);
   
   outputs.push_back(Spectro_return);
	//FIN ALBAN
  
   clean_all(pgspectrop);
  
   return QPLUG_SUCCESS;
  }
  
