#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include <list>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"

using namespace QSAS;
using namespace std;

extern "C"
{
    QplugReturnStatus DvSort( vector <DvObject_var>&, vector <DvObject_var>& );
    QplugReturnStatus DvUnion( vector <DvObject_var>&, vector <DvObject_var>& );
    QplugReturnStatus DvIntersection( vector <DvObject_var>&, vector <DvObject_var>& );
    QplugReturnStatus DvExclusion( vector <DvObject_var>&, vector <DvObject_var>& );
}



struct sortCriterionIncStart : public std::binary_function< DvEvent&, DvEvent&, bool > 
{
    inline bool operator()(const DvEvent& i1, const DvEvent& i2)
    {
        return (i1.start() < i2.start() || (i1.start()==i2.start() && i1.end() < i2.end()));
    }
};

struct sortCriterionDecStart : public std::binary_function< DvEvent&, DvEvent&, bool > 
{
    inline bool operator()(const DvEvent& i1, const DvEvent& i2)
    {
        return (i2.start() < i1.start() || (i2.start()==i1.start() && i2.end() < i1.end()));
    }
};


struct sortCriterionIncEnd : public std::binary_function< DvEvent&, DvEvent&, bool > 
{
    inline bool operator()(const DvEvent& i1, const DvEvent& i2)
    {
        return (i1.end() < i2.end() || (i1.end()==i2.end() && i1.start() < i2.start()));
    }
};

struct sortCriterionDecEnd : public std::binary_function< DvEvent&, DvEvent&, bool > 
{
    inline bool operator()(const DvEvent& i1, const DvEvent& i2)
    {
        return (i2.end() < i1.end() || (i2.end()==i1.end() && i2.start() < i1.start()));
    }
};

struct sortCriterionIncCentre : public std::binary_function< DvEvent&, DvEvent&, bool > 
{
    inline bool operator()(const DvEvent& i1, const DvEvent& i2)
    {
        DvTime centre1 = i1.centre();
        DvTime centre2 = i2.centre();
        return (centre1 > centre2 || (centre1==centre2 && i1.start() < i2.start()));
    }
};

struct sortCriterionDecCentre : public std::binary_function< DvEvent&, DvEvent&, bool > 
{
    inline bool operator()(const DvEvent& i1, const DvEvent& i2)
    {
        DvTime centre1 = i1.centre();
        DvTime centre2 = i2.centre();
        return (centre2 < centre1 || (centre2==centre1 && i2.start() < i1.start()));
    }
};

QplugReturnStatus DvSort( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{
	
    DvObject_var table = inputs[0];
    
    if(table.is_nil() || table->not_event())
    {
        QplugAppendTextDisplay("Failed: Input is not an event table.");
        return QPLUG_FAILURE;
    }
    
    DvObject_var sortType = inputs[1];
    
    if(sortType.is_nil() )
    {
        QplugAppendTextDisplay("Failed: sort criterion.");
        return QPLUG_FAILURE;
    }
	
	vector<DvEvent> seq;
	seq.resize(table->seqSize(), DvEvent());
	for(size_t i=0; i<table->seqSize(); i++) seq[i] = table->event(i); // copy
	
    switch(sortType->asInt())
    {
        case 1: std::sort(seq.begin(), seq.end(), sortCriterionIncCentre());
                break;
       
        case 2: std::sort(seq.begin(), seq.end(), sortCriterionIncEnd());
                break;
                
        case 3: std::sort(seq.begin(), seq.end(), sortCriterionDecStart());
                break;
       
        case 4: std::sort(seq.begin(), seq.end(), sortCriterionDecCentre());
                break;
                
        case 5: std::sort(seq.begin(), seq.end(), sortCriterionDecEnd());
                break;
       
        default: std::sort(seq.begin(), seq.end(), sortCriterionIncStart());
                break;
    }
    
	DvObject_var res = new DvObject(DvEvent(), seq.size());
	for(size_t i=0; i<seq.size(); i++) res->event(i) = seq[i];
	
    outputs.push_back(res);
	
    return QPLUG_SUCCESS;
}

typedef struct
{
    std::vector<DvEvent> elements;
    std::vector<bool>    isInSet1;
    std::vector<bool>    isInSet2;
} set_elements;

void getSetElements(DvObject_var& l1, DvObject_var& l2, set_elements* res)
{
	// creates a vector of short events that are entirely inside or entirely outside 
	// of the events in the input event lists
	
    res->elements.clear();
    res->isInSet1.clear();
    res->isInSet2.clear();
    
    std::list<DvTime> points;
    
	// assemble list of all start and end times
    for(size_t i=0; i<l1->seqSize(); i++)
    {
        points.push_back(l1->event(i).start());
        points.push_back(l1->event(i).end());
    }
    
    for(size_t i=0; i<l2->seqSize(); i++)
    {
        points.push_back(l2->event(i).start());
        points.push_back(l2->event(i).end());
    }
    
	// ensure they are in order.
	
    points.sort();
    
	// break time range into contiguous events between each time (not original events)
	
    DvTime t1, t2;
    t1=points.front();
    points.pop_front();
    while(!points.empty())
    {
        t2=points.front();
        points.pop_front();
        
        res->elements.push_back(DvEvent(t1, t2));
        t1=t2;
    }
    
    int nb_elements=res->elements.size();
    
    res->isInSet1.resize(nb_elements, false);
    res->isInSet2.resize(nb_elements, false);

    // find if each event in elements is inside an event of either list (or in neither) 
	// it must be entirely inside or outside as contiguous events are based on event edges
	// l1 and l2 may not be sorted, but elements is sorted
	
    for(size_t i=0; i<l1->seqSize(); i++)
    {
        t1=l1->event(i).start();
        t2=l1->event(i).end();
        bool end=false;
        for(int j=0; j<nb_elements && !end; j++)
        {
			DvTime c = res->elements[j].centre();
            if(c > t1 && c < t2)
            {
                res->isInSet1[j]=true;
            }
            if(c > t2) end=true;
        }
    }
    
    for(size_t i=0; i<l2->seqSize(); i++)
    {
        t1=l2->event(i).start();
        t2=l2->event(i).end();
        bool end=false;
        for(int j=0; j<nb_elements && !end; j++)
        {
 			DvTime c = res->elements[j].centre();
			if(c > t1 && c < t2)
            {
                res->isInSet2[j]=true;
            }
            if(c > t2) end=true;
        }
    }
}

QplugReturnStatus DvUnion( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{
    DvObject_var times1 = inputs[0];
    DvObject_var times2 = inputs[1];
    
    
    if(times1.is_nil() || times1->not_event())
    {
        QplugAppendTextDisplay("Failed: First object is not an event table.");
        return QPLUG_FAILURE;
    }
    
    if(times2.is_nil() || times2->not_event())
    {
        QplugAppendTextDisplay("Failed: Second object is not an event table.");
        return QPLUG_FAILURE;
    }
    
	// create vector of sub events each entirely inside or outside input event lists
    set_elements elts;
    getSetElements(times1, times2, &elts);
    
    // assemble list of sub intervals into new list of contiguous events
	vector<DvEvent> newList;
	
    DvTime t1, t2;
    bool isInInterval=false;
    for(int i=0; i<elts.elements.size(); i++)
    {
        if(elts.isInSet1[i] || elts.isInSet2[i]) // If in union
        {
            if(!isInInterval) // We have just entered the union interval
            {
                t1=elts.elements[i].start();
                t2=elts.elements[i].end();
                isInInterval=true;
            }
            else
            {
                t2=elts.elements[i].end();
            }
        }
        else
        {
            if(isInInterval) // We have just left the union interval
            {
                newList.push_back(DvEvent(t1, t2));
                isInInterval=false;
            }
        }
    }
	
	DvObject_var res=new DvObject(DvEvent(), newList.size());
	for(size_t i=0; i<newList.size(); i++) res->event(i) = newList[i];

    outputs.push_back(res);
	
    return QPLUG_SUCCESS;
}

QplugReturnStatus DvIntersection( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{
    DvObject_var times1 = inputs[0];
    DvObject_var times2 = inputs[1];
    
    if(times1.is_nil() || times1->not_event())
    {
        QplugAppendTextDisplay("Failed: First object is not an event table.");
        return QPLUG_FAILURE;
    }
    
    if(times2.is_nil() || times2->not_event())
    {
        QplugAppendTextDisplay("Failed: Second object is not an event table.");
        return QPLUG_FAILURE;
    }
    
	// create vector of sub events each entirely inside or outside input event lists
    set_elements elts;
    getSetElements(times1, times2, &elts);
    
    // assemble list of sub intervals into new list of contiguous intersect events 
	vector<DvEvent> newList;
    
    DvTime t1, t2;
    bool isInInterval=false;
    for(int i=0; i<elts.elements.size(); i++)
    {
        if(elts.isInSet1[i] && elts.isInSet2[i]) // If in union
        {
            if(!isInInterval) // We have just entered the union interval
            {
                t1=elts.elements[i].start();
                t2=elts.elements[i].end();
                isInInterval=true;
            }
            else
            {
                t2=elts.elements[i].end();
            }
        }
        else
        {
            if(isInInterval) // We have just left the union interval
            {
                newList.push_back(DvEvent(t1, t2));
                isInInterval=false;
            }
        }
    }
	
	DvObject_var res=new DvObject(DvEvent(), newList.size());
	for(size_t i=0; i<newList.size(); i++) res->event(i) = newList[i];

    outputs.push_back(res);

    return QPLUG_SUCCESS;
}

QplugReturnStatus DvExclusion( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{
    DvObject_var times1 = inputs[0];
    DvObject_var times2 = inputs[1];
        
    if(times1.is_nil() || times1->not_event())
    {
        QplugAppendTextDisplay("Failed: First object is not an event table.");
        return QPLUG_FAILURE;
    }
    
    if(times2.is_nil() || times2->not_event())
    {
        QplugAppendTextDisplay("Failed: Second object is not an event table.");
        return QPLUG_FAILURE;
    }

	// create vector of sub events each entirely inside or outside input event lists
    
    set_elements elts;
    getSetElements(times1, times2, &elts);
    
    // assemble list of sub intervals into new list of contiguous exclusion events 
	vector<DvEvent> newList;
    
    DvTime t1, t2;

    bool isInInterval=false;
    for(int i=0; i<elts.elements.size(); i++)
    {
        if(elts.isInSet1[i] && !elts.isInSet2[i]) // If in union
        {
            if(!isInInterval) // We have just entered the union interval
            {
                t1=elts.elements[i].start();
                t2=elts.elements[i].end();
                isInInterval=true;
            }
            else
            {
                t2=elts.elements[i].end();
            }
        }
        else
        {
            if(isInInterval) // We have just left the union interval
            {
                newList.push_back(DvEvent(t1, t2));
                isInInterval=false;
            }
        }
    }
	
	DvObject_var res=new DvObject(DvEvent(), newList.size());
	for(size_t i=0; i<newList.size(); i++) res->event(i) = newList[i];

    outputs.push_back(res);

    return QPLUG_SUCCESS;
}
