/* featherPlot.cc Jeremy Mitchell 10/08/2009 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include "Xrefs.h"

extern "C" QplugReturnStatus DvfeatherPlot(vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;

QplugReturnStatus DvfeatherPlot( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs ) {
  // Unpack input data
  
  DvObject_var pos1_in   = inputs[0];
  DvObject_var deltax_in = inputs[1];
  DvObject_var scale_in  = inputs[2];
  DvObject_var step_in   = inputs[3];
  DvObject_var boxCar_in  = inputs[4];
  DvObject_var head_in    = inputs[5];
  DvObject_var hScale_in  = inputs[6];
  
  //********************** Test Inputs ***********************************
  
  if (pos1_in.is_nil() || pos1_in->not_dbl() ||  !pos1_in->isThreeVector() ) {
    QplugAppendTextDisplay("featherPlot error: Position 1 is not a vector\n");
    return QPLUG_FAILURE;
  }
    
  if (deltax_in.is_nil() || !deltax_in->isThreeVector()) {
	QplugAppendTextDisplay("featherPlot error: delta x is not a vector\n");
	return QPLUG_FAILURE;
  }
  
  if (scale_in.is_nil() || scale_in->not_dbl()) {
    QplugAppendTextDisplay("featherPlot error: Scale is not a Scalar\n");
    return QPLUG_FAILURE;
  }
  
  if (step_in.is_nil() ) {
    QplugAppendTextDisplay("featherPlot error: Step is not provided\n");
    return QPLUG_FAILURE;
  }
  
  int step = step_in->asInt();
  if (step <= 0) {
    QplugAppendTextDisplay("featherPlot error: step must be > 0\n");
    return QPLUG_FAILURE;
  }
  
  if (boxCar_in.is_nil()) {
    QplugAppendTextDisplay("featherPlot error: Boxcar missing\n");
    return QPLUG_FAILURE;
  }
  int boxCar = boxCar_in->asInt();
  if (boxCar < 0) {
    QplugAppendTextDisplay("featherPlot error: boxCar must be >= 0\n");
    return QPLUG_FAILURE;
  }
  
  if (head_in.is_nil()) {
    QplugAppendTextDisplay("featherPlot error: Draw Head is missing\n");
    return QPLUG_FAILURE;
  }
  int head = head_in->asInt();
  int arrowLength = 7;
  if ((head < 0) || (head > 3)) {
    QplugAppendTextDisplay("featherPlot error: Draw Head must be between 0 and 3\n");
    return QPLUG_FAILURE;
  }
  if (head == 0)
    arrowLength = 3;
  
  if (hScale_in.is_nil()) {
    QplugAppendTextDisplay("featherPlot error: Width missing\n");
    return QPLUG_FAILURE;
  }
  float hScale = (float)hScale_in->asDouble();
  if ((hScale < 0)) {
    QplugAppendTextDisplay("featherPlot WARNING: Width is less than 0\n\tArrow may point wrong way!");
  }
  
  //***************** step stuff ***************************
  
  // DETERMINE INFORMATION ABOUT THE TIME STEPS
  
  bool timeStep = false;
  
  // Determine if step is in space or time:
  //   If step UNITS = seconds AND pos1 has time_tags, then TIME step
  //   First, check units of Step:
  //   NB!!! IF timeStep = true, but arrows have no time_tags, then only consider the first arrow
  if (step_in->xref_exists(UNITS)) {
    DvString stepUnits = step_in->getXrefText(UNITS);
    
    // if step not in seconds, then space step
    if ( (stepUnits == "s") ||  (stepUnits == "S") || (stepUnits == "sec") ||
      (stepUnits == "Sec") || (stepUnits == "SEC") ) {
      
      // check pos1 for time_tags
      timeStep = true;
      if ( !pos1_in->hasTimeTags() ) {
			QplugAppendTextDisplay("featherPlot Warning: Pos1 has no TIME_TAGS => Spatial Step!\n");
			timeStep = false;
      }
      
    }
  }
  
  //***************** Find Drawn Arrow positions ***************************
  
  int nout,nStep;		// sized of final array and number of drawn arrows, respectively
  DvObject_var outArr;	// output array, with arrows
  DvObject_var arrowArr;	// array with the vectors showing arrows that will be drawn
  DvObject_var ttout;		// time_tags for output array (if applicable)
  int initNstep = pos1_in->seqSize();	// size of initial pos1 array
  int *arrowPlace;	// array containing position of drawn arrows are in outArr
  

  if ( !timeStep ) {
    // SPATIAL TIMESTEP STUFF
    //-----------------------
    
    QplugAppendTextDisplay("featherPlot: Spatial steps chosen");
    
    float* path = new float[initNstep];
    path[0] = 0.;
    for (int i=1; i < initNstep; i++) {
      float px = (float)pos1_in->dbl(i,0) - (float)pos1_in->dbl(i-1,0);
      float py = (float)pos1_in->dbl(i,1) - (float)pos1_in->dbl(i-1,1);
      float pz = (float)pos1_in->dbl(i,2) - (float)pos1_in->dbl(i-1,2);
      path[i] = path[i-1] + sqrt(px*px + py*py + pz*pz);
    }
    
    nStep = (int)( floorf(path[initNstep-1]/step) );
    nout = initNstep + arrowLength * nStep;
    
    outArr = new DvObject(0.0, DIM3, nout);
    arrowArr = new DvObject(0.0, DIM3, nStep);
    
    arrowPlace = new int[nStep];
    
    int k=0;
    for (int i=0; i<nStep; i++) {
      float pos = step * i;
      
      while (path[k+1] < pos) {
    	outArr->dbl(k + i*arrowLength, 0) = pos1_in->dbl(k,0);
    	outArr->dbl(k + i*arrowLength, 1) = pos1_in->dbl(k,1);
    	outArr->dbl(k + i*arrowLength, 2) = pos1_in->dbl(k,2);
    	k++;
		if(k >= initNstep) {
        	QplugAppendTextDisplay("pos error\n");
        	return QPLUG_FAILURE;
		}
      }
      
      arrowPlace[i] = k + i*arrowLength;
      float wt1 = (pos - path[k])/(path[k+1] - path[k]);
	  float wt2 = (pos - path[k+1])/(path[k] - path[k+1]);
      outArr->dbl(k + i*arrowLength,0) = wt1*(float)(pos1_in->dbl(k,0)) + wt2*(float)(pos1_in->dbl(k+1,0));
      outArr->dbl(k + i*arrowLength,1) = wt1*(float)(pos1_in->dbl(k,1)) + wt2*(float)(pos1_in->dbl(k+1,1));
      outArr->dbl(k + i*arrowLength,2) = wt1*(float)(pos1_in->dbl(k,2)) + wt2*(float)(pos1_in->dbl(k+1,2));
     
      if ( deltax_in->seqSize() == 1 ) {
    	// all arrows the same
		arrowArr->dbl(i,0) = deltax_in->dbl(0,0);
    	arrowArr->dbl(i,1) = deltax_in->dbl(0,1);
    	arrowArr->dbl(i,2) = deltax_in->dbl(0,2);
      } 
	  else if ( boxCar == 0) {
        // Linear interpolation of arrows
		arrowArr->dbl(i,0) = wt1*(float)(deltax_in->dbl(k,0)) + wt2*(float)(deltax_in->dbl(k+1,0));
        arrowArr->dbl(i,1) = wt1*(float)(deltax_in->dbl(k,1)) + wt2*(float)(deltax_in->dbl(k+1,1));
        arrowArr->dbl(i,2) = wt1*(float)(deltax_in->dbl(k,2)) + wt2*(float)(deltax_in->dbl(k+1,2));
      } 
	  else {
        // Boxcar average of arrows
		int l=k, kmin=k, kmax=k;
		while ((pos - path[l] <= boxCar/2.) && (l >= 0)) l--;
		kmin = l+1; l = k;
		while ((path[l] - pos <= boxCar/2.) && (l <= initNstep)) l++;
		kmax = l-1;
	
		arrowArr->dbl(i,0) = 0.; 
		arrowArr->dbl(i,1) = 0.; 
		arrowArr->dbl(i,2) = 0.;
		for (l=kmin; l <= kmax; l++) {
	  		arrowArr->dbl(i,0) += deltax_in->dbl(l,0) / (kmax - kmin + 1);
	  		arrowArr->dbl(i,1) += deltax_in->dbl(l,1) / (kmax - kmin + 1);
	  		arrowArr->dbl(i,2) += deltax_in->dbl(l,2) / (kmax - kmin + 1);
		}
      }
    }

    for (int m=k + nStep*arrowLength; m < nout; m++) {
      outArr->dbl(m,0) = pos1_in->dbl(m - nStep*arrowLength, 0);
      outArr->dbl(m,1) = pos1_in->dbl(m - nStep*arrowLength, 1);
      outArr->dbl(m,2) = pos1_in->dbl(m - nStep*arrowLength, 2);
    }
    delete [] path;

  } 
  else {
    // timestep is true:
    
    QplugAppendTextDisplay("feathPlot: Time steps chosen");
    DvObject_var ttin = pos1_in->getTimeTags();
    DvTime tt0 = ttin->time(0);
    DvTime ttf = ttin->time(initNstep-1);
    
    nStep = (int)( floorf(ttf - tt0) / step) ;
    nout = initNstep + arrowLength * nStep;
    
    outArr = new DvObject(0.0, DIM3, nout);
    
    arrowPlace = new int[nStep];
    
    DvObject_var ttjoin = new DvObject(DvTime(), nStep);
    ttout = new DvObject(DvTime(), nout);
    for (int i=0; i<nStep; i++)
      ttjoin->time(i) = tt0 + i * step;
    
	DvObject_var outTmp;

      if (boxCar == 0) {
		pos1_in->change_xref(DV_GAP_OPTION, DV_ZERO_FILL);
		pos1_in->change_xref(DV_GAP_WIDTH, step+10.);
		outTmp = pos1_in->linearJoin(*ttjoin) ;
    } 
	else {
		pos1_in->change_xref(DV_GAP_OPTION, DV_ZERO_FILL);
		pos1_in->change_xref(DV_BOX_WIDTH, (double) boxCar);
		outTmp = pos1_in->boxcarJoin(*ttjoin) ;
    }
    
    
    if (deltax_in->seqSize() > 1) {
	    if (boxCar == 0) {
			deltax_in->change_xref(DV_GAP_OPTION, DV_ZERO_FILL);
			deltax_in->change_xref(DV_GAP_WIDTH, step+10.);
			arrowArr = deltax_in->linearJoin(*ttjoin );
	    } 
		else {
			deltax_in->change_xref(DV_GAP_OPTION, DV_ZERO_FILL);
			deltax_in->change_xref(DV_BOX_WIDTH, (double) boxCar);
			arrowArr = deltax_in->boxcarJoin(*ttjoin);
	    }
	
    } 
	else {
		arrowArr = new DvObject(0.0, DIM3, nStep);
		for (int i=0; i<nStep; i++) {
			arrowArr->dbl(i,0) = deltax_in->dbl(0,0);
			arrowArr->dbl(i,1) = deltax_in->dbl(0,1);
			arrowArr->dbl(i,2) = deltax_in->dbl(0,2);
	  }
    }
   
    int k=0,i=0;
    while( i < initNstep) {
      if ( (ttin->time(i) - tt0 <= step*k) || (k == nStep) ) {
		outArr->dbl(i+k*arrowLength,0) = pos1_in->dbl(i,0);
        outArr->dbl(i+k*arrowLength,1) = pos1_in->dbl(i,1);
        outArr->dbl(i+k*arrowLength,2) = pos1_in->dbl(i,2);
		ttout->time(i+k*arrowLength) = ttin->time(i);
		i++;
      } 
	  else if (k < nStep) {
		outArr->dbl(i+k*arrowLength,0) = outTmp->dbl(k,0);
        outArr->dbl(i+k*arrowLength,1) = outTmp->dbl(k,1);
        outArr->dbl(i+k*arrowLength,2) = outTmp->dbl(k,2);
		arrowPlace[k] = i+k*arrowLength;
		for (int m=0; m < arrowLength; m++)
	  		ttout->time(i+k*arrowLength+m) = ttjoin->time(k);
			//i += arrowLength;
			k++;
      	}
      }
  }
      QplugAppendTextDisplay("do arrow\n");

  //******************** calculate *************************
  
  for (int k = 0; k < nStep; k++) {
    
    int api = arrowPlace[k];
    
    float delx = (float)(arrowArr->dbl(k,0)) * (float)scale_in->dbl();
    float dely = (float)(arrowArr->dbl(k,1)) * (float)scale_in->dbl();
    float delz = (float)(arrowArr->dbl(k,2)) * (float)scale_in->dbl();
    outArr->dbl(api+1,0) = (float)(outArr->dbl(api,0)) + delx;
    outArr->dbl(api+1,1) = (float)(outArr->dbl(api,1)) + dely;
    outArr->dbl(api+1,2) = (float)(outArr->dbl(api,2)) + delz;
        
    if (head != 0) {
      float phi, xarr, yarr, zarr, width;
      
      switch (head) {
    	case 1:
    	  width = hScale*sqrt(delx*delx + dely*dely);
    	  phi = atan2(dely, delx);
    	  xarr = -sin(phi)*width;
    	  yarr = cos(phi)*width;
  	  	  zarr = 0;
  	  	  break;
    	case 2:
    	  width = hScale*sqrt(delx*delx + delz*delz);
    	  phi = atan2(delz, delx);
    	  xarr = -sin(phi)*width;
    	  yarr = 0;
  	  	  zarr = cos(phi)*width;
  	  	  break;
    	case 3:
    	  width = hScale*sqrt(dely*dely + delz*delz);
    	  phi = atan2(dely, delz);
    	  xarr = 0;
    	  yarr = cos(phi)*width;
		  zarr = -sin(phi)*width;
  	  	  break;
      }
      
      float onemw = 1. - hScale;
      outArr->dbl(api+2,0) = (float)(outArr->dbl(api,0)) + onemw*delx + xarr;
      outArr->dbl(api+2,1) = (float)(outArr->dbl(api,1)) + onemw*dely + yarr;
      outArr->dbl(api+2,2) = (float)(outArr->dbl(api,2)) + onemw*delz + zarr;
      
      outArr->dbl(api+3,0) = outArr->dbl(api+1,0);
      outArr->dbl(api+3,1) = outArr->dbl(api+1,1);
      outArr->dbl(api+3,2) = outArr->dbl(api+1,2);
      
      outArr->dbl(api+4,0) = (float)(outArr->dbl(api,0)) + onemw*delx - xarr;
      outArr->dbl(api+4,1) = (float)(outArr->dbl(api,1)) + onemw*dely - yarr;
      outArr->dbl(api+4,2) = (float)(outArr->dbl(api,2)) + onemw*delz - zarr;
      
      outArr->dbl(api+5,0) = outArr->dbl(api+1,0);
      outArr->dbl(api+5,1) = outArr->dbl(api+1,1);
      outArr->dbl(api+5,2) = outArr->dbl(api+1,2);
      
      outArr->dbl(api+6,0) = outArr->dbl(api+0,0);
      outArr->dbl(api+6,1) = outArr->dbl(api+0,1);
      outArr->dbl(api+6,2) = outArr->dbl(api+0,2);
    } 
	else {
      outArr->dbl(api+2,0) = outArr->dbl(api+0,0);
      outArr->dbl(api+2,1) = outArr->dbl(api+0,1);
      outArr->dbl(api+2,2) = outArr->dbl(api+0,2);
    }
  }  
  DvObject_var dep0 = new DvObject(0.0, nout);
  int d0=0;
  if (head == 3)
    d0 = 1;
  for (int i=0; i<nout; i++) dep0->dbl(i) = outArr->dbl(i,d0);
  
  // Add attributes:
  
  outArr->copy_xrefs_from(pos1_in);
  dep0->copy_xrefs_from(pos1_in);
  
  if (dep0->xref_exists("TIME_TAGS")) {
    dep0->delete_xref("TIME_TAGS");
    outArr->delete_xref("TIME_TAGS");
  }
  if (dep0->xref_exists(DEPEND_0)) {
      dep0->delete_xref(DEPEND_0);
      outArr->delete_xref(DEPEND_0);
  }
  if (timeStep) {
    outArr->change_xref("DEPEND_0", ttout);
    dep0->change_xref("DEPEND_0", ttout);
  }
  
  if (dep0->xref_exists(DEPEND_1)) {
    dep0->delete_xref(DEPEND_1);
    outArr->delete_xref(DEPEND_1);
    QplugAppendTextDisplay("Deleting outArr DEPEND_1\n");
  }
  
  dep0->change_xref(FRAME, "scalar>na");
  outArr->change_xref("DEPEND_1", dep0);
  
    
  //***********************************************************
  
  // Assign data to output:
  
  delete [] arrowPlace;
    
  outputs.push_back(outArr);

  QplugAppendTextDisplay ("featherPlot>> finished...\n");
  
  return QPLUG_SUCCESS;
}
