/*-*- c++ -*-
// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Generic Residue Analysis of Conservation Laws
// 
//
// Inputs :
// 	bfield		- vector time series with magnetic field
//	velocity	- vector time series with plasma velocity
//	density	        - scalar time series with plasma density
//      ptensor         - 6 scalar time series with the pressure tensor entries
//
// Outputs
//
// SEH, 5 Sep, 2002
//      31 May 2006, added MVA dB/dt
// AJA  24 Aug 2006, added defesive programming (input, unit checks etc.)
// SEH  08 Aug 2008, removed QT dependencies, replaced Qdos::invert with invert function
//
// ------------------------------------------------------------------*/
 
#include "GenericMinVar.h"
#include "GenericResidue.h"

DvObject_var str2Arr(DvString s, size_t len);

extern "C" QplugReturnStatus  DvGenericResidue(vector <DvObject_var>&, vector <DvObject_var>&);


QplugReturnStatus  DvGenericResidue( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{ 
  long    nestsize, nseq=0;
  int i,j,k,constrained=0, min_nest=0, nb=0;
  double deltat, avg_width;
  DvString SC_id = "";
  char txtbuf[255] = "";

  bool inSup[10];
  bool doType[10]; 
  const char* inSyms[10]={"B","dBdt","V","N","E","Q","P","",""};
  const char* doSyms[10]={"MVAB","MVAdBdt","HT","MFR","MMR","MLMR","MER","MTER","MVAV","COM"};

  for (int i=0; i<10; i++) inSup[i] = doType[i] = false;

  DvObject_var  B;                                       // B = Magnetic field series
  DvObject_var  dB;                                      // dB = time derivative of magnetic field series
  DvObject_var  V;                                       // V = Velocity series
  DvObject_var  N;                                       // N = Plasma density
  DvObject_var  E;                                       // E = E-field vector series
  DvObject_var  H;                                       // H = Heat flux vector series
  DvObject_var  P;                                       // P = Pressure tensor
  DvObject_var  PS;                                      // PS = .. or pressure as scalar
  DvObject_var  Q;

  DvObject_var ttJoined;                                    // tt of joined V, N, B are from the pressure tensor

  DvObject_var  Erpar;                           // Residual electric field para
  DvObject_var  Erperp;                          // Residual electric field perp
  DvObject_var  Weights;



 // --- start by checking inputs 0 : Bfield -----------------

 B = inputs[0];
 B = B->changeUnitsTo("1.0e-9>T", "nT");   
 if(B.is_ok()) {
   inSup[0] = true;   
   nb = B->seqSize();                             // get length of B 
   ttJoined = B->getTimeTags();                  // dafault : tt of joined are from B;
   if (ttJoined.is_nil()) {
     QplugAppendTextDisplay("Bad or missing time tags in input data\n");
     return QPLUG_FAILURE;
   }
   if ( B->not_dbl()) {
     QplugAppendTextDisplay("input B field not double\n");
     return QPLUG_FAILURE;
   }
   deltat = ttJoined->get_spacing();
   avg_width = 1.5 * deltat;
   printf("Sampling time = %f.\n",(double) deltat); 
   nseq+=1;
 }


 // ---- check for dB/dt, take from B if input == 1 ------------

 bool dBdt_supplied=false;
 dB = inputs[1];        // check if dB/dt supplied as TS...
 if (dB.is_ok() &&  dB->seqSize() > 1) {
   nseq +=1;
   inSup[1] = true;
 }
 else if(B.is_ok()) {
   // if not..  check if dB/dt supplied as integer
   dBdt_supplied = dB.is_ok();
   bool dBdt_replace = (dB.is_ok() && dB->asInt() == 1);  // if '1' => calc dB/dt from B
   if (!dBdt_supplied && deltat > 0.0 && dBdt_replace) {   
     QplugAppendTextDisplay("Calculating dB / dt from B field input\n");                      
     dB =  new DvObject(1.0, DIM3, nb);
     for (int i=1; i<nb-1; i++) {
       for (int j=0; j<3; j++) dB->dbl(i,j) = ( B->dbl(i+1,j) - B->dbl(i-1,j) ) / (2.0*deltat);
     }
     for (int j=0; j<3; j++) {
       dB->dbl(0,j) = dB->dbl(1,j);
       dB->dbl(nb-1,j) = dB->dbl(nb-2,j);
     }
     dB->change_xref(DEPEND_0, ttJoined);
     DvString SICxref = "1.0e-9>T s^-1";
     dB->change_xref("SI_conversion", SICxref);
     nseq +=1;
     inSup[1] = true;        
   }       
 }
 
 printf("nseq = %ld\n", nseq);


 // ----- check V ------------

 V = inputs[2];
 V = V->changeUnitsTo("1.0e3>m s^-1", "km/s");   // get V
 if (V.is_ok() && V->seqSize() > 1 && V->is_dbl() && V->seqSize() > 1 ) {
   inSup[2] = true;   
   ttJoined = V->getTimeTags();                  // if V supplied, tt of joined are from V;
   if (ttJoined.is_nil()) {
     QplugAppendTextDisplay("Bad or missing time tags in input data\n");
     return QPLUG_FAILURE;
   }
   deltat = ttJoined->get_spacing();
   avg_width = 1.5 * deltat;
   nseq +=1;
 }
 else {
   QplugAppendTextDisplay("WARNING: cannot confirm V units  or no V given");
   QplugAppendTextDisplay("Check SI_CONVERSION and FRAME attributes");		
 }


 // ----- check N ------------

 N = inputs[3];
 N = N->changeUnitsTo("1.0e6>m^-3", "cm-3");   // get N
 if ( N.is_ok() && N->arraySize() == 1 && N->is_dbl() && N->seqSize() > 1 ) {
   inSup[3] = true;   
   nseq +=1;
 } 
 else {
   QplugAppendTextDisplay("WARNING: cannot confirm that N units are 1.0e6m^-3 or no N given");		
   QplugAppendTextDisplay("Check SI_CONVERSION and FRAME attributes");		
 }


 // ----- check E ------------

 E = inputs[4];
 E = E->changeUnitsTo("1.0e3>V m^-1", "kV/m");   // get E
 if (E.is_ok() && E->is_dbl() && E->seqSize() > 1 ) {
   nseq +=1;
   inSup[4] = true;   
 } 
 else {
	E = inputs[4];
	E = E->changeUnitsTo("1.0e-6>m s^-1 T", "nT km/s ");   // try units from VxB
	if(E.is_nil() || E->seqSize() == 1) {
		QplugAppendTextDisplay("WARNING: cannot confirn E units are either 1.0e3>V m^-1  or 1.0e-6>m s^-1 T \n");
		QplugAppendTextDisplay("Check SI_CONVERSION and FRAME attributes");		 
	} 
	else {
		nseq +=1;
		inSup[4] = true;   
	}
 }


 // ----- check Q - heat flux vector ------------

 Q = inputs[5];
 if (Q.is_ok() && Q->arraySize() == 3 && Q->is_dbl() && Q->seqSize() > 1 ) {
   nseq +=1;
   inSup[5] = true;   
 } 
 else {
   QplugAppendTextDisplay("WARNING: no Q given");		
 }


 // ----- check P pressure ------------

 P = inputs[6];   // get P scalar
 if (P.is_ok() && P->is_dbl() && P->seqSize() > 1 ) {
   nseq +=1;
   inSup[6] = true;   
 } 
 else {
   QplugAppendTextDisplay("WARNING: no P given");		
 }

 // ----- Matrix of what to do - depends on supplied inputs -- 

 doType[0] = inSup[0];                                                               // MVAB
 doType[1] = inSup[1];                                                               // MVAdBdt
 doType[2] = inSup[0] && inSup[2];                                                   // HT needs B and (V or E)
 doType[3] = inSup[0] && inSup[2];                                                   // MFR needs B and V
 doType[4] = inSup[2] && inSup[3];                                                   // MMR needs N and V
 doType[5] = inSup[0] && inSup[2] && inSup[3] && inSup[6];                           // MLMR : N,V,N,P
 doType[6] = inSup[0] && inSup[2] && inSup[3] && inSup[6];                           // MER : N,V,N,P
 doType[7] = inSup[0] && inSup[2] && inSup[3] && inSup[4]  && inSup[5] && inSup[6];  // MTER : N,V,N,E,Q,P
 doType[8] = inSup[2];                                                               // MVAV requires V;

 for (int i=0; i<10; i++) {
   sprintf(txtbuf,"insup[%d] : %d\n",i,(int)(inSup[i] == 1)); 
   QplugAppendTextDisplay(txtbuf);
 }

 sprintf(txtbuf,"Joining %ld sequences\n\nDoing:",nseq);
   QplugAppendTextDisplay(txtbuf);

 for (int i=0; i<10; i++) {
   if(doType[i]) {
     sprintf(txtbuf,"%s\n",doSyms[i]); 
     QplugAppendTextDisplay(txtbuf);
   }
 }

 // ---- now, join the data -----
 
DvJoinList ToJoinList;
DvJoinList JoinedList;

 if (inSup[0]) {
   ToJoinList.append("B", B);
   ttJoined = B->getTimeTags(); 
   SC_id =  B->getXrefText("SC_id");		//get SC ID
   B->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
   B->change_xref(DV_GAP_WIDTH, avg_width);
   B->change_xref(DV_GAP_OPTION, DV_LINEAR); 
 }
 if (inSup[1]) {
   ToJoinList.append("dB", dB);
   if (!inSup[0]) {
   	ttJoined = dB->getTimeTags();
   	SC_id =  dB->getXrefText("SC_id");		//get SC ID 
   }
   dB->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
   dB->change_xref(DV_GAP_WIDTH, avg_width);
   dB->change_xref(DV_GAP_OPTION, DV_LINEAR); 
 }
 if (inSup[2]) {
   ToJoinList.append("V", V);
   ttJoined = V->getTimeTags(); 
   SC_id =  V->getXrefText("SC_id");		//get SC ID 
   V->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
   V->change_xref(DV_GAP_WIDTH, avg_width);
   V->change_xref(DV_GAP_OPTION, DV_LINEAR); 
 }
 if (inSup[3]) {
   ToJoinList.append("N", N);
   N->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
   N->change_xref(DV_GAP_WIDTH, avg_width);
   N->change_xref(DV_GAP_OPTION, DV_LINEAR); 
 }
 if (inSup[4]) {
   ToJoinList.append("E", E);
   E->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
   E->change_xref(DV_GAP_WIDTH, avg_width);
   E->change_xref(DV_GAP_OPTION, DV_LINEAR); 
 }
 if (inSup[5]) {
   ToJoinList.append("H", Q);
   Q->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
   Q->change_xref(DV_GAP_WIDTH, avg_width);
   Q->change_xref(DV_GAP_OPTION, DV_LINEAR); 
 }
 if (inSup[6]) {
   ToJoinList.append("P", P);
   P->change_xref(DV_JOIN_METHOD, DV_LINEAR);	
   P->change_xref(DV_GAP_WIDTH, avg_width);
   P->change_xref(DV_GAP_OPTION, DV_LINEAR); 
 }


// get interval of interest

 DvEvent start_end = inputs[7]->getTimeRange();
 if(start_end == DvEvent()){
     QplugAppendTextDisplay("Error : Time interval not recognised ");
     return (QPLUG_FAILURE);
 }
 
 ttJoined = ttJoined->subSequence(start_end);
 if(ttJoined.is_nil() || ttJoined->seqSize() <= 0 ){
     QplugAppendTextDisplay("Error : No time tags in range ");
     return (QPLUG_FAILURE);
 }

  // Multijoin will give output on ttJoined tags (no need to subsequence later)

  ToJoinList.MultiJoin(ttJoined, JoinedList);



 if (inSup[0]) {
   printf("joined B.\n");
   DvNode *Bnode = JoinedList.find("B");
   DvObject_var masked_B;
     
   if(Bnode) masked_B = Bnode->obj();
   
   if(masked_B.is_nil() || masked_B->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No B field in range ");
     return (QPLUG_FAILURE);
   }

   B = masked_B->changeUnitsTo("1.0e-9>T", "nT"); // ensure units are in those assumed
   if(B.is_nil() || B->seqSize() < 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm B units are 1.0e-9>T\n");
     B = masked_B;
   }    
   nestsize = B->seqSize();
 }

 if (inSup[1]) {
   printf("joined ddtB.\n");
   DvNode *dBnode = JoinedList.find("dB");
   DvObject_var masked_dB;

   if(dBnode) masked_dB = dBnode->obj();
   if(masked_dB.is_nil() || masked_dB->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No dB/dt data in range ");
     return (QPLUG_FAILURE);
   }
   
   dB = masked_dB->changeUnitsTo("1.0e-9>T s^-1", "nT/s"); // ensure units are in those assumed
   if(dB.is_nil() || dB->seqSize() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm dB units are 1.0e-9>T s^-1\n");
     dB = masked_dB;
   } 
   nestsize = dB->seqSize();
 }

 if (inSup[2]) {
   printf("joined V.\n");
   DvNode *Vnode = JoinedList.find("V");
   DvObject_var masked_V;
     
   if(Vnode) masked_V = Vnode->obj();
   if(masked_V.is_nil() || masked_V->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No V data in range ");
     return (QPLUG_FAILURE);
   } 
   
   V = masked_V->changeUnitsTo("1.0e3>m s^-1", "km/s"); // ensure units are in those assumed
   if(V.is_nil() || V->seqSize() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm N units are 1.0e6>m^-3\n");
     V = masked_V;
   }     
   nestsize = V->seqSize();
 }

 if (inSup[3]) {
   printf("joined N.\n");
   DvNode *Nnode = JoinedList.find("N");
   DvObject_var masked_N;
     
   if(Nnode) masked_N = Nnode->obj();
   if(masked_N.is_nil() || masked_N->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No N data in range ");
     return (QPLUG_FAILURE);
   } 
   
   N = masked_N->changeUnitsTo("1.0e6>m^-3", "cm-3"); // ensure units are in those assumed
   if(N.is_nil() || N->seqSize() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm N units are 1.0e6>m^-3\n");
     N = masked_N;
   }  
 }

 if (inSup[4]) {
   DvNode *Enode = JoinedList.find("E");
   DvObject_var masked_E;
     
   if(Enode) masked_E = Enode->obj();
   if(masked_E.is_nil() || masked_E->seqSize() <= 0){
     QplugAppendTextDisplay("Error : No E data in range ");
     return (QPLUG_FAILURE);
   }
   
   E = masked_E->changeUnitsTo("1.0e-3>V m^-1", "V/km"); // ensure units are in those assumed   
   if(E.is_nil() || E->seqSize() == 1) {
       QplugAppendTextDisplay("WARNING: cannot confirm E units are either V km^-1  or km s^-1 nT \n");
       E = masked_E;
   }
 }

 if (inSup[5]) {
     DvNode *Hnode = JoinedList.find("H");
     DvObject_var masked_H;
     
     if(Hnode) masked_H = Hnode->obj();
     if(masked_H.is_nil() || masked_H->seqSize() <= 0){
         QplugAppendTextDisplay("Error : No Heat Flux in range ");
         return (QPLUG_FAILURE);
     }
   
   H = masked_H->changeUnitsTo("1.0>J m^-2 s^-1", "J/m^2/s"); // ensure units are in those assumed 
   if(H.is_nil() || H->seqSize() == 1) {
     QplugAppendTextDisplay("WARNING: Cannot confirm H units are 1.0>J m^-2 s^-1\n");
     H = masked_H;
   }  
 }

    if (inSup[6]) {
        DvNode *Pnode = JoinedList.find("P");
        DvObject_var masked_P;
        
        if(Pnode) masked_P = Pnode->obj();
	if (masked_P.is_nil() || masked_P->seqSize() <= 0) {                                                     // ..if neither tensor or scalar -> failure
       QplugAppendTextDisplay("Error : no P in range ");
       return (QPLUG_FAILURE);
	}
	
	if (masked_P->arraySize() == 1 ) {                                                        //  scalar pressure
     
		PS = masked_P->changeUnitsTo("1.0e12>J m^-3", " "); // ensure units are in those assumed 
		if(PS.is_nil() || PS->seqSize() == 1) {
			QplugAppendTextDisplay("WARNING: Cannot confirm Pressure units are 1.0e12>J m^-3\n");
			PS = masked_P;
		}  
		P = new DvObject(0.0, DIM33, nestsize);                         // ..construct empty symmetric P matrix 
		printf("P is scalar - filling diagonal....\n");
		for (k=0; k<nestsize; k++){
			for (i=0; i<3; i++) 
				for (j=0; j<3; j++) 
					P->dbl(k,i,i) = PS->dbl(k);                                       // .. and fill the trace of P with scalar p
		}
	} 
	else {
     
		P = masked_P->changeUnitsTo("1.0e12>J m^-3", " "); // ensure units are in those assumed 
		if(P.is_nil() || P->seqSize() == 1) {
			QplugAppendTextDisplay("WARNING: Cannot confirm P units are 1.0e12>J m^-3\n");
			P = masked_P;
     }    
     printf("P is tensor....\n");
   } 
 }


 // ---- Get minimum nest size (if 0 - single nest size) ------------------

 DvObject_var minnest_do = inputs[8];
 if( minnest_do.is_nil() ) {
   min_nest = 0;
 } else  {
   min_nest  = minnest_do->asInt();
   if (min_nest != 0 && min_nest <= 3) min_nest = 3;
 }
 if (min_nest >= nestsize) min_nest=0;  // do single nest if insufficeient samples.
 

 // ---- Get type MVA (constrained or normal) ------------------

 DvObject_var in_8 = inputs[9];
 double alpha=1.0;   // walenslope
 if( in_8.is_nil()){
 	QplugAppendTextDisplay("Error : No constraint ");
	return (QPLUG_FAILURE);
 }
 
 if(in_8->is_str() ) {
	DvObject_var c_vector = str2Arr(in_8->str(), 3);
	constrained = c_vector->asInt(0);
 } 
 else {
	constrained = in_8->asInt();
 }


 // --- certain constrains not possible without specific inputs :
 if ( (!inSup[0] && (constrained == 1 || constrained >=3))  ||  (!inSup[2] && constrained == 2) ) {
   constrained = 0;
   QplugAppendTextDisplay("WARNING : Insufficient inputs for selected constrain\n");
   
 }
 printf("Constrain = %d  \n",constrained);

 // ---- Get weights - 4 numbers with Q-ratios ------


 double sum_weights=0.0;
 DvObject_var W_string = inputs[10];
 if (W_string.is_ok()) {
    Weights = str2Arr( W_string->str(), 8);
 }
 for (i=0; i<8; i++) sum_weights += Weights->dbl(i);

 
 // ---- check for sensible amount of data -----

 if (nestsize > 1000) {
  QplugAppendTextDisplay("\n == More than 1000 datapoints. Try reducing time interval ==\n");
  return(QPLUG_FAILURE);
 }


 // ---- Get Q_matrices for the various conservation laws ------------------

 if  (min_nest <= 0) {
   printf("Single nest analysis..\n");
   if (!SingleAnalysis(V,B,dB,P,E,N,H,ttJoined,Weights,constrained,alpha,SC_id,doType,nestsize, outputs)) printf("Single nest analysis OK\n");
 }
 else {
   printf("Nested analysis - min nest size = %d\n",min_nest);
   if (!NestedAnalysis(V,B,dB,P,E,N,H,ttJoined,Weights,constrained,alpha,min_nest,SC_id,doType,nestsize)) printf("Nested analysis OK\n");
 }

 // QString s2 = "\nTime elapsed : " + QString::number(t.elapsed()/1000.0) +"\n\n";
 // printf("T elapsed = %f\n",t.elapsed()/1000.0);
 //QplugAppendTextDisplay(s2);

 return (QPLUG_SUCCESS);
}



// -------------------------------------------------------------------
/*!
  \brief Parses the a text and converts text to a Vector
  \return  QSAS matrix with the parsed values of the string
  \param s - a string on the form '[2.00 -23.0  9.0]'
*/
//
// SEH, 08 Aug 2008, Non-QT version
//
// -------------------------------------------------------------------
DvObject_var str2Arr(DvString s, size_t len)
{
	vector <size_t> dims;
	dims.push_back(len);
	
    DvObject_var v = new DvObject(0.0, dims);
    s.strip('[');
    s.strip(']');
    s=s.trimmed(); // remove surrounding white space
	for(size_t i=0; i<len-1; i++){
    	v->dbl(i) = s.before(' ').toDouble();
		s=s.after(' ');
		s=s.trimmed();
	}
    v->dbl(len-1) = s.toDouble();

    return(v);
}
