// ------------------------------------------------------------------
//
// ISSI QSAS PlugIn : Minimum Faraday Residue calculation
// 
// This plugin takes a magnetic field vector and a plasma velocity
// vector as inputs, calculates the E-field (from V x B) and 
// thereafter calculates the MFR velocity and MFR normal
//
// See Khrabov & Sonnnerup (KS), GRL, p2374, 1998 
//
// Inputs :
// 	bfield		- vector time series with magnetic field
//	velocity	- vector time series with plasma velocity
//
// Outputs
//
// SEH, 29 May, 2002
//
// ------------------------------------------------------------------

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"
#include "ConstrainedMinVarMFR.h"


extern "C" QplugReturnStatus  DvMinFar(vector <DvObject_var>&, vector <DvObject_var>&);

using namespace QSAS;

QplugReturnStatus  DvMinFar( vector <DvObject_var>& inputs, vector <DvObject_var>& outputs )
{ 
 long    nestsize;
 char    txtbuf[200];
 float   dB2=0.0, dE2=0.0,u_n;
 double  deltat;
 int     nseq=2,constrained=0;
 int     i,j,k;

 // ---- QSAS types (see also Khrabov & Sonnerup, GRL, p2374, 1998) ---------

 DvObject_var  B;                                       // B = Magnetic field series
 DvObject_var  V;                                       // V = Velocity series
 DvObject_var  E;                                       // E = Electric field series
 DvObject_var  B_ave;                                   // <B> = averaged B
 DvObject_var  E_ave;                                   // <E> = averaged E
 DvObject_var  dB;                                      // <dB> 
 DvObject_var  dE;                                      // <dE> 
 DvObject_var  lB;                                      // temporary <dB> used in (3) below
 DvObject_var  lE;                                      // temporary <dE> used in (3) below 
 DvObject_var  P;                                       // P = Poynting flux
 DvObject_var  Q;                                       // Q = (see KS pap)
 DvObject_var  n_MFR;                                   // the MFR normal;
 DvObject_var  mvatype_do;                              // Constrained (1) or unconstrained (0) MVA 


 //  --- first we join B measurements to V measurements -----------------
  

 B = inputs[0];
 if(B.is_nil() || B->not_dbl() ){
   QplugAppendTextDisplay("Bad or missing B field input data\n");
   return QPLUG_FAILURE;
 }

 V = inputs[1];
 if(V.is_nil() || V->not_dbl() ){
   QplugAppendTextDisplay("Bad or missing Velocity input data\n");
   return QPLUG_FAILURE;
 }
 

 //  ---- get Time interval and subset ----------------

 DvEvent start_end = inputs[2]->getTimeRange();

 B = B->subSequence(start_end);
 V = V->subSequence(start_end);

 V->isRegular(deltat);

 DvObject_var gapIf = new DvObject(1.5 * deltat);

 if (VERBOSE_MV)
 {
	printf("\nMinFar > delta t = %f  ->  gap_fill if t > %f\n", deltat, gapIf->dbl());
	fflush(stdout);
  }

 B->change_xref(GAP_TOLERANCE, gapIf);
 B->change_xref(GAP_METHOD, DV_LINEAR);
 
 B = B->Join(V);

 if (VERBOSE_MV)
 {
	printf("MinFar > Gap if = %f\n", gapIf->dbl());
	fflush(stdout);
 }
   
 if(B.is_nil())
 {
	printf("Joined B is nil");
	fflush(stdout);
 }
 
 // added by AJA to ensure B field units are correct (nT)
 B = B->changeUnitsTo("1.0e-9>T", "nT");

 if( B->is_nil() ){
     QplugAppendTextDisplay("ERROR: cannot convert to nT");
     return QPLUG_FAILURE;
 }

 nestsize = B->seqSize();
 
 if (VERBOSE_MV)
 {
	printf("MinFar > Nestsize B = %ld\n", nestsize);
	fflush(stdout);
 }
 
 // added by AJA to ensure velocity units are correct (km s^-1)
 V = V->changeUnitsTo("1.0e3>m s^-1", "km/s");
 
 if( V->is_nil() ){
     QplugAppendTextDisplay("ERROR: cannot convert to km/s");
     return QPLUG_FAILURE;
 }

 nestsize = V->seqSize();
 if (VERBOSE_MV)
 {
	printf("MinFar > Nestsize V = %ld\n",nestsize);
	fflush(stdout);
 }

 mvatype_do = inputs[3];
 if( mvatype_do.is_nil() ) {
   QplugAppendTextDisplay("ERROR: cannot understand constrained choice");  
   return QPLUG_FAILURE;
 }
 constrained = mvatype_do->asInt();
 if( constrained > 1 || constrained < 0 ){
   QplugAppendTextDisplay("ERROR: constrained choice must be 0 or 1");  
   return QPLUG_FAILURE;
 }


 // ----- Start calculation as described in the K & S  paper ------


 // ---- (1)  Calculate E = -V x B = B x V  and <E> and <B> ---------------- 

 E = new DvObject(0.0, DIM3, nestsize);   
 E_ave = new DvObject(0.0, DIM3);
 B_ave = new DvObject(0.0, DIM3);

 float cc = 1.0e-03;   // conversion due to B in [nT] and V in [km/s]  

 for (k=0; k<nestsize; k++) {
   E->dbl(k,0) = cc*( B->dbl(k,1) * V->dbl(k,2)  -  B->dbl(k,2) * V->dbl(k,1) );
   E->dbl(k,1) = cc*( B->dbl(k,2) * V->dbl(k,0)  -  B->dbl(k,0) * V->dbl(k,2) );
   E->dbl(k,2) = cc*( B->dbl(k,0) * V->dbl(k,1)  -  B->dbl(k,1) * V->dbl(k,0) );
   for (j=0; j<3; j++) {
      B_ave->dbl(j) += B->dbl(k,j);
      E_ave->dbl(j) += E->dbl(k,j);
   }
 }   

 *B_ave /= (double)nestsize;
 *E_ave /= (double)nestsize;
 

 if (VERBOSE_MV) {
   printf("MinFar > B_ave = [%9.3f %9.4f %9.4f]\n",(float)B_ave->dbl(0),(float)B_ave->dbl(1),(float)B_ave->dbl(2));
   printf("MinFar > E_ave = [%9.3f %9.4f %9.4f]\n",(float)E_ave->dbl(0),(float)E_ave->dbl(1),(float)E_ave->dbl(2));
   fflush(stdout);
 }


 // ---- (2) Calculate electric field variance matrix M^E -------- 

 DvObject_var ME = new DvObject(0.0, DIM33);

 for (k=0; k<nestsize; k++) {
   for (i=0; i<3; i++) {
     for (j=0; j<3; j++) {  
       ME->dbl(0,i,j) += (E->dbl(k,i) - E_ave->dbl(i)) *  (E->dbl(k,j)- E_ave->dbl(j));
     }
   }
 }
 
 *ME /= (double)nestsize;
 
 if (VERBOSE_MV) {
   printf("MinFar > ME =\n");
   for (i=0; i<3; i++) {
     for (j=0; j<3; j++) printf("%10.5f ",(float)ME->dbl(0,i,j) );
     printf("\n");
   }
   printf("\n");
   fflush(stdout);
 }
 

 // ---- (3) Calculate <|dB|^2> and <|dE|^2>  -------- 

 lE = new DvObject(0.0, DIM3);
 lB = new DvObject(0.0, DIM3);
 dB2 = 0.0;
 dE2 = 0.0;

 for (k=0; k<nestsize; k++) {
   for (i=0; i<3; i++) {
     lB->dbl(i) = B->dbl(k,i) - B_ave->dbl(i);
     lE->dbl(i) = E->dbl(k,i) - E_ave->dbl(i);
   }
   DvObject_var lB2 = lB->dot(lB);
   DvObject_var lE2 = lE->dot(lE);
   dB2 += lB2->dbl();
   dE2 += lE2->dbl();
 }

 dB2 /= (float) nestsize;
 dE2 /= (float) nestsize;

 if (VERBOSE_MV)
 {
	printf("MinFar > dB2 = %10.4f, dE2 = %10.4f\n",dB2,dE2);
	fflush(stdout);
 }


 // ---- (4) Calculate Poynting vector P = <dB x dE> --------------

 P  = new DvObject(0.0, DIM3);
 dB = new DvObject(0.0, DIM3);
 dE = new DvObject(0.0, DIM3);

 if (VERBOSE_MV)
 {
	printf("MinFar > P, dB and dE initialized\n");
	fflush(stdout);
 }

 for (k=0; k<nestsize; k++) {
   for (i=0; i<3; i++) {
     dB->dbl(i) = B->dbl(k,i) - B_ave->dbl(i);
     dE->dbl(i) = E->dbl(k,i) - E_ave->dbl(i);
   }
   P->dbl(0) += ( (dE->dbl(1) * dB->dbl(2)) - (dE->dbl(2) * dB->dbl(1)) );
   P->dbl(1) += ( (dE->dbl(2) * dB->dbl(0)) - (dE->dbl(0) * dB->dbl(2)) );
   P->dbl(2) += ( (dE->dbl(0) * dB->dbl(1)) - (dE->dbl(1) * dB->dbl(0)) );
 }

 *P /= (float) nestsize;

 if (VERBOSE_MV) 
 {
	printf("MinFar > P = [%10.4f %10.4f %10.4f]\n",(float)P->dbl(0),(float)P->dbl(1),(float)P->dbl(2));
	fflush(stdout);
 }


 // ---- (5) Calculate Q matrix ------------------------------------

 Q  = new DvObject(0.0, DIM33);

 for(i=0; i<3; i++)
   for(j=0; j<3; j++)
     Q->dbl(0,i,j) = -1.0 * ME->dbl(0,i,j) - P->dbl(i)*P->dbl(j)/dB2;

 for(i=0; i<3; i++) Q->dbl(0,i,i) += dE2;                  // add the <|dE|^2>d_ij term
 
 if (VERBOSE_MV) {
   printf("MinFar > Q =\n");
   for (i=0; i<3; i++) {
     for (j=0; j<3; j++) printf("%10.5f ",(float)Q->dbl(0,i,j));
     printf("\n");
   }
   printf("\n");
   fflush(stdout);
 }


 // ---- (6) Find eigenvals + eigenvecs of Q --------------------- 

  ConstrainedMinVarMFR minvar(constrained, B, Q);   // dvos minvariance call

  DvObject_var Rmatx = minvar.eigenvectors();
  DvObject_var Eigenvalues = minvar.eigenvalues(); 

 if (VERBOSE_MV) {
   printf("MinFar > Eigenvalues, eigenvectors are :\n");
   for (i=0; i<3; i++){
     printf("%10.5g  (%10.5g, %10.5g, %10.5g )\n", 
	     Eigenvalues->dbl(i),
	     Rmatx->dbl(0,i,0),
	     Rmatx->dbl(0,i,1), 
	     Rmatx->dbl(0,i,2) );
   }   
   fflush(stdout);
 }



 // ----- (7) Calculate V_MFR along normal --------

 u_n = 0.0;
 for (i=0; i<3; i++){
   u_n += (float)P->dbl(i) * (float)Rmatx->dbl(0,0,i) / dB2;
 }
  
 u_n *= 1000.0;                  // Since P [nT * mV/m], and we want u_n in [km/s]

 if (VERBOSE_MV)
 {
	printf("\nMinFar > u_n = %10.4f\n",u_n);
	fflush(stdout);
 }

 // ----- Print output values ---------------------


 n_MFR = new DvObject(0.0, DIM3);

 sprintf(txtbuf,"MinFar > Nestsize = : %ld \n",nestsize);
 QplugAppendTextDisplay ( txtbuf);

 sprintf(txtbuf,"MinFar > Eigenvalues, eigenvectors are :\n");
 QplugAppendTextDisplay ( txtbuf);
 for (i=0; i<3; i++){
     sprintf(txtbuf,"          %10.5g  [%10.5g, %10.5g, %10.5g ]\n", 
	   Eigenvalues->dbl(i),
	   Rmatx->dbl(0,i,0),
	   Rmatx->dbl(0,i,1), 
	   Rmatx->dbl(0,i,2) );
     QplugAppendTextDisplay ( txtbuf);
     n_MFR->dbl(i) = Rmatx->dbl(0,0,i);
 }   

 sprintf(txtbuf,"MinFar > U_n = %10.4f km/s\n",u_n);
 QplugAppendTextDisplay( txtbuf);



 // ----- Output normal and Efield ------

 DvString frame_in = B->getFrameAttr();
 n_MFR->change_xref(FRAME, frame_in);
 E->change_xref(FRAME, frame_in);
 E->change_xref(UNITS,"mV/m");
 E->change_xref(SI_CONVERSION,"0.001>V m^-1");
 E->change_xref(LABLAXIS,"E");

 outputs.push_back(E);
 outputs.push_back(n_MFR);

 return QPLUG_SUCCESS;

}





