
/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	MMS_CDF_tensors.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Plugin to check/fix metadata and make 
*                  tensors from components for MMS FPI cdf files
*
* PURPOSE:      As in FUNCTION
* 
* DESCRIPTION:  Check/add SC_ID, SI_CONVERSION, LABLAXIS, merge components
*                 into tensors. Also centre timetags.
*
* ARGUMENTS:  (ALL inputs are individually optional; as are outputs!)
*    IN:      
*             xx, xy, xz, yy, yz, zz indep components of tensor
* 
*   OUT:
*             t_out  3x3 tensor
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
*
* HISTORY:   SJS 7 May 2016; 
*/

#include "qplug_if.h"
#include "DvObj.h"

//#define DEBUG_FPI_TENS    // for testing

#ifdef DEBUG_FPI_TENS
void print_tens(DvObject_var tens_ptr, size_t nrec);
#endif

extern "C" QplugReturnStatus DvMMS_FPI_CDF_tensors(vector<DvObject_var>&, vector<DvObject_var>&);

QplugReturnStatus CentreObjTimetags (DvObject_var dobj_ptr);

QplugReturnStatus DvMMS_FPI_CDF_tensors(vector<DvObject_var>& inputs, vector<DvObject_var>& outputs)
{
	// Outputs
	DvObject_var t_out_ptr;  
		

#ifdef DEBUG_FPI_TENS
	printf("starting\n");
#endif
	
 	if(inputs.size() < 6) {
    	QplugAppendTextDisplay ("insufficient inputs\n");
    	return QPLUG_FAILURE; 
 	} 
 
	//	Unpack and process the input data objects

	// ***************************
	// SCALAR COMPONENTS of TENSOR
	// ***************************
	DvObject_var xx_ptr = inputs[0];
	DvObject_var xy_ptr = inputs[1];
	DvObject_var xz_ptr = inputs[2];
	DvObject_var yy_ptr = inputs[3];
	DvObject_var yz_ptr = inputs[4];
	DvObject_var zz_ptr = inputs[5];
	
	DvString xx_name = xx_ptr->getXrefText("ObjectName");
	DvString xy_name = xy_ptr->getXrefText("ObjectName");
	DvString xz_name = xz_ptr->getXrefText("ObjectName");
	DvString yy_name = yy_ptr->getXrefText("ObjectName");
	DvString yz_name = yz_ptr->getXrefText("ObjectName");
	DvString zz_name = zz_ptr->getXrefText("ObjectName");
	
	DvString xx_sc   = xx_name.str()[3];
	DvString xy_sc   = xy_name.str()[3];
	DvString xz_sc   = xz_name.str()[3];
	DvString yy_sc   = yy_name.str()[3];
	DvString yz_sc   = yz_name.str()[3];
	DvString zz_sc   = zz_name.str()[3];

	DvString xx_coor;
	if (xx_name.contains("gse")) {xx_coor = "gse";
	} else {xx_coor="dbcs";}
	DvString xy_coor;
	if (xy_name.contains("gse")) {xy_coor = "gse";
	} else {xy_coor="dbcs";}
	DvString xz_coor;
	if (xz_name.contains("gse")) {xz_coor = "gse";
	} else {xz_coor="dbcs";}
	DvString yy_coor;
	if (yy_name.contains("gse")) {yy_coor = "gse";
	} else {yy_coor="dbcs";}
	DvString yz_coor;
	if (yz_name.contains("gse")) {yz_coor = "gse";
	} else {yz_coor="dbcs";}
	DvString zz_coor;
	if (zz_name.contains("gse")) {zz_coor = "gse";
	} else {zz_coor="dbcs";}

	
// initial checks for inputs exist and same seqSize and des or dis and same tensor
	if(  !xx_ptr->is_nil() 
	  && !xy_ptr->is_nil() 
	  && !xz_ptr->is_nil()
	  && !yy_ptr->is_nil()
	  && !yz_ptr->is_nil()
	  && !zz_ptr->is_nil()	  
	  && (
	     (xx_name.contains("des") && xy_name.contains("des") && xz_name.contains("des")
		 && yy_name.contains("des") && yz_name.contains("des") && zz_name.contains("des"))
	   || (xx_name.contains("dis") && xy_name.contains("dis") && xz_name.contains("dis")
		 && yy_name.contains("dis") && yz_name.contains("dis") && zz_name.contains("dis"))
	  )
	  && xx_name.contains("xx")
	  && xy_name.contains("xy")
	  && xz_name.contains("xz")
	  && yy_name.contains("yy")
	  && yz_name.contains("yz")
	  && zz_name.contains("zz")
	  && (
	     (xx_name.contains("pres") && xy_name.contains("pres") && xz_name.contains("pres")
		 && yy_name.contains("pres") && yz_name.contains("pres") && zz_name.contains("pres"))
	   || (xx_name.contains("temp") && xy_name.contains("temp") && xz_name.contains("temp")
		 && yy_name.contains("temp") && yz_name.contains("temp") && zz_name.contains("temp"))
	  )
	  && xx_ptr->seqSize() == xy_ptr->seqSize() 
	  && xx_ptr->seqSize() == xz_ptr->seqSize() 
	  && xx_ptr->seqSize() == yy_ptr->seqSize() 
	  && xx_ptr->seqSize() == yz_ptr->seqSize() 
	  && xx_ptr->seqSize() == zz_ptr->seqSize() 
	  && (xx_sc=="1" || xx_sc=="2" || xx_sc=="3" || xx_sc=="4")
	  && xx_sc == xy_sc 
	  && xx_sc == xz_sc 
	  && xx_sc == yy_sc 
	  && xx_sc == yz_sc 
	  && xx_sc == zz_sc 
	  && xx_coor == xy_coor
	  && xx_coor == xz_coor
	  && xx_coor == yy_coor
	  && xx_coor == yz_coor
	  && xx_coor == zz_coor
	  
	  ){   // ALL CHECKS PASSED

		
		QplugAppendTextDisplay ( "Inputs passed checks; tensor being processed");

		// after approach in makevector		
		DvObject_var xx_unit = new DvObject(0.0, DIM33, 1);
			xx_unit->dbl(0,0,0) = 1.0;
		DvObject_var yy_unit = new DvObject(0.0, DIM33, 1);
			yy_unit->dbl(0,1,1) = 1.0;
		DvObject_var zz_unit = new DvObject(0.0, DIM33, 1);
			zz_unit->dbl(0,2,2) = 1.0;
		
		DvObject_var xyyx_unit = new DvObject(0.0, DIM33, 1);
			xyyx_unit->dbl(0,0,1) = 1.0; 
			xyyx_unit->dbl(0,1,0) = 1.0;
		DvObject_var xzzx_unit = new DvObject(0.0, DIM33, 1);
			xzzx_unit->dbl(0,0,2) = 1.0; 
			xzzx_unit->dbl(0,2,0) = 1.0;
		DvObject_var yzzy_unit = new DvObject(0.0, DIM33, 1);
			yzzy_unit->dbl(0,1,2) = 1.0; 
			yzzy_unit->dbl(0,2,1) = 1.0;
		
		// convert to tensors
		
		DvObject_var XX_ptr = xx_ptr * xx_unit;
		DvObject_var YY_ptr = yy_ptr * yy_unit;
		DvObject_var ZZ_ptr = zz_ptr * zz_unit;
		
		DvObject_var XY_ptr = xy_ptr * xyyx_unit;   // both off diag elements in xy_ptr
		DvObject_var XZ_ptr = xz_ptr * xzzx_unit;
		DvObject_var YZ_ptr = yz_ptr * yzzy_unit;
		
		// try setting SI_CONVERSION for all to ensure += operator works
		// will change later for final output object
		XX_ptr->change_xref("SI_CONVERSION","1>m");  
		YY_ptr->change_xref("SI_CONVERSION","1>m"); 
		ZZ_ptr->change_xref("SI_CONVERSION","1>m"); 
		XY_ptr->change_xref("SI_CONVERSION","1>m"); 
		XZ_ptr->change_xref("SI_CONVERSION","1>m"); 
		YZ_ptr->change_xref("SI_CONVERSION","1>m"); 
		
		// add together
		t_out_ptr  = new DvObject(XX_ptr); 
		t_out_ptr += YY_ptr;
		t_out_ptr += ZZ_ptr;
		t_out_ptr += XY_ptr;
		t_out_ptr += XZ_ptr;
		t_out_ptr += YZ_ptr;

		// set up output object
		DvString t_out_name = GetSlotText(0);
		if (t_out_name.blank() || t_out_name.empty()){
			t_out_name = xx_name.replace("xx","_tensor");
			SetSlotText(0, t_out_name.c_str());
		}
		// add object name here for centre timetag parsing if needed
		// qplugin interface will rename using sanitised output slot text.
		t_out_ptr->change_xref("ObjectName",xx_name);

		// CENTRE TIMETAGS
		if(CentreObjTimetags(t_out_ptr)==QPLUG_FAILURE){
			QplugAppendTextDisplay("Failed to centre timetags for "+t_out_name);
			return QPLUG_FAILURE;
		}

		DvString quantity_str;  // for later use in label
		// SI_CONVERSION
		if (xx_name.contains("temp")){
			quantity_str = "temp";
			t_out_ptr->change_xref("SI_CONVERSION","11604.50520>K");
		} else if (xx_name.contains("pres")) {
			quantity_str = "press";
			t_out_ptr->change_xref("SI_CONVERSION","1.0e-1>Pa");
		} else {  // this can't be reached as initial checks allow temp, pres only
			QplugAppendTextDisplay("Unknown units & SI_conversion for "+t_out_name);
			t_out_ptr->change_xref("SI_CONVERSION","1>(unknown");
			quantity_str = "unk";
		}

		// SC_ID
		t_out_ptr->change_xref("SC_ID",xx_sc);
		
		// FRAME
		DvString frame = "tensor>gse_xyz";    // force gse even for dbcs
		t_out_ptr->change_xref("FRAME",frame);
		if (xx_coor != "gse"){
			t_out_ptr->change_xref("DBCS2GSE","Input in DBCS relabelled as near GSE and given Frame=gse_xyz for qsas convenience");
			QplugAppendTextDisplay("Caution: dbcs frame relabelled as gse for internal QSAS convenience.");
		}
		t_out_ptr->change_xref("TENSOR_ORDER","2");
		t_out_ptr->change_xref("COORDINATE_SYSTEM","GSE");
		DvString repstr = "x";
		DvObject_var rep = new DvObject(repstr,DIM3,1);
		rep->str(1) = "y"; rep->str(2) = "z";
		t_out_ptr->change_xref("REPRESENTATION_1",rep);
		t_out_ptr->change_xref("LABEL_1",rep);
		t_out_ptr->change_xref("REPRESENTATION_2",rep);
		t_out_ptr->change_xref("LABEL_2",rep);		
		if(t_out_ptr->xref_exists("REPRESENTATION")){
			t_out_ptr->delete_xref("REPRESENTATION");
		}
		t_out_ptr->change_xref("TENSOR_ORDER","2");

		// LABLAXIS & FIELDNAM
		DvString fieldnam_str = "mms";
		fieldnam_str +=xx_sc;
		if (xx_name.contains("dis")){
			quantity_str += "_i";
			t_out_ptr->change_xref("LABLAXIS", quantity_str);
			fieldnam_str +="_dis_";
			fieldnam_str +=quantity_str;
			t_out_ptr->change_xref("FIELDNAM", fieldnam_str);
		} else {
			quantity_str += "_e";
			t_out_ptr->change_xref("LABLAXIS", quantity_str);
			fieldnam_str +="_des_";
			fieldnam_str +=quantity_str;
			t_out_ptr->change_xref("FIELDNAM", fieldnam_str);
		}

		// push back to outputs
		outputs.push_back(t_out_ptr);

	 } else {
	   QplugAppendTextDisplay("Inputs checks failed. Check order, spacecraft, coord system, joined etc.");
	   return QPLUG_FAILURE;
	 }

	 return QPLUG_SUCCESS;
		
}

#ifdef DEBUG_FPI_TENS

void print_tens(DvObject_var tens_ptr, size_t nrec){
cout << "" << endl;
cout << tens_ptr->dbl(nrec,0,0) << " " << tens_ptr->dbl(nrec,0,1) << " " << tens_ptr->dbl(nrec,0,2)<< endl;
cout << tens_ptr->dbl(nrec,1,0) << " " << tens_ptr->dbl(nrec,1,1) << " " << tens_ptr->dbl(nrec,1,2)<< endl;
cout << tens_ptr->dbl(nrec,2,0) << " " << tens_ptr->dbl(nrec,2,1) << " " << tens_ptr->dbl(nrec,2,2)<< endl;
}

#endif
	
