/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/*************************************************************************
* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	ClusterConfig.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Cluster Config (Qtetra) Plugin wrapper
*
* PURPOSE:      Show/Manipulate Cluster 4 s/c positions
* 
* DESCRIPTION:  .
*
* ARGUMENTS:
*    IN:      QdCDOList  list of plug in objects containing
*
*             sc1, sc2, sc3, sc4      Constant vector or Vector TS of 4s/c
*                                     Typically sc_drn deltas from AUX
*             time_c				  time of interest
*
*   OUT:
*             Nothing; separate gui launched
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
* FUNCTIONS USED: built on Qtetra, and bits copied/modified from QBNCoords
*
* HISTORY: March 2002 Steve Schwartz 
*
***********************************************************************/

// #include "Qtetra_gui.h"
#include "qplug_if.h"
#include "qrtn.h"
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qdutil.h"
#include "DvObj.h"
#include "QCCWindow.h"

#define MAXSTRLEN 500 


using namespace QSAS;  

// local prototypes

QplugReturnStatus QCFPosition(RtnVector *vecp,
					DvObject_var &pos_in,
					DvObject_var &time_c);

void RtnVector_diag_print(RtnVector *vecp);


/**********************************************************************/
extern "C" QplugReturnStatus DvQClusterConfig( vector<DvObject_var>& inputs, vector<DvObject_var>& outputs ){
	/**********************************************************************/

    PlotWindow *win;
    
    if(plotWin_id == NULL) {
        plotWin_id = new PlotID("Cluster Configuration");
        win = plotWin_id->newPlotWindow(600, 600, 72, QPrinter::A4, QPrinter::Portrait, true);
    }
    else{
        win = plotWin_id->currentWindow;
    }
    
    if(plotWidgets == NULL) plotWidgets = new PlugPlotWidgets(win);
    win->addWidget(plotWidgets, 0, 0, Qt::AlignCenter, 9);
    
	DvObject_var cin[4];       // 0,1,2,3
	DvObject_var   time_c;     // 4

	RtnVector  c[4];

	QplugReturnStatus ret_status;

// ...first 4 items are input position vectors/vector TS 
//... last one is time 

	int isc;
	for (isc = 0; isc < 4; isc++){
		cin[isc] = inputs[isc];
		if (cin[isc].is_nil()){
			QplugAppendTextDisplay("QClusterConfig>> Error getting position object\n");
			return QPLUG_FAILURE;
		}
		DvString attr = cin[isc]->getXrefText(SC_ID);
		int scid = attr.toInt() - 1;
		if ( scid < 0 || scid != isc){
			QplugAppendTextDisplay ("WARNING, unknown or WRONG Spacecraft number\n");
			QuMsgBar("WARNING: Check Plugin messages", true);
		}
	}

	time_c = inputs[4];
	if( time_c.is_nil() ) {
		QplugAppendTextDisplay ("QClusterConfig>> Time object invalid.\n");
		return QPLUG_FAILURE;
	}

	
	for (isc = 0; isc < 4; isc++)
	{

  // first check if this is a position vector
		bool warn=false;
		if (!cin[isc]->isThreeVector() ) warn = true;
		if( !cin[isc]->sameBaseSI( "1>m" ) ) warn = true;
		if(warn){
			QplugAppendTextDisplay( "CAUTION, spacecraft positions do not appear to be vector distances\n");
		}
  
		ret_status = QCFPosition(&(c[isc]), cin[isc], time_c);
		if (ret_status != QPLUG_SUCCESS){
			return QPLUG_FAILURE;
		}
	}
//  c[i] now has a sequence with 1 element, the xyz position
// of the ith spacecraft with one timetag. 
	
// 	QString s;
// 	for (isc = 0; isc < 4; isc++){
// 		qtet_widget->sc_details[isc]->posx->setText(s.setNum(c[isc].ptr[0][0]));
// 		qtet_widget->sc_details[isc]->posy->setText(s.setNum(c[isc].ptr[0][1]));
// 		qtet_widget->sc_details[isc]->posz->setText(s.setNum(c[isc].ptr[0][2]));
// 	}
	QString qtitlepre = "Cluster Config ";
	QString qtitle  = qtitlepre.append(c[0].t->time().getISOstring().c_str());
	
	DvString frameptr = c[--isc].ptr->getFrame();

	DvString attr;

// we now have frame sub-str 
	if (c[isc].ptr->xref_exists(UNITS)) attr = c[isc].ptr->getXrefText(UNITS);
	else if (c[isc].ptr->xref_exists(LABLAXIS)) attr = c[isc].ptr->getXrefText(LABLAXIS);
	else attr = "()";

	char units_str[MAXSTRLEN];
	sprintf(units_str,"%s (%s)",attr.c_str(), frameptr.c_str());
	
	QString units(units_str);
	plotWidgets->setParameters(c[0], c[1], c[2], c[3], qtitle, units);

    

	return QPLUG_SUCCESS;

}

/**********************************************************************/
QplugReturnStatus QCFPosition(RtnVector *vecp,
			      DvObject_var &pos_in,
				  DvObject_var &time_c){
/**********************************************************************/

    QplugReturnStatus ret_status;

    // sort out position. Input can either be a time series or
    // a constant vector, so this a bit complicated. Separate routines
    // return in vecp a sequence of 1 cartesian vector whose timetag
    // is the time interval time_c, and whose frame is
    // that of pos_in. 

    // prepare vecp 
    if( pos_in->seqSize() > 1) vecp->ptr = pos_in->Join(time_c);
    else vecp->ptr = pos_in;
    
    vecp->t = time_c;
    vecp->n_recs = 1;
    vecp->in_frame = vecp->QrGetFrameFromAttr(pos_in);

    return QPLUG_SUCCESS;
}


/**************************************************************************/
void RtnVector_diag_print(RtnVector *vecp){
    printf("IsPositionVector = %d, in_frame = %d, out_frame = %d, n_recs = %d\n",
        vecp->IsPositionVector, vecp->in_frame, vecp->out_frame, vecp->n_recs);
    printf("First Vector: (%9.3g, %9.3g, %9.3g) at %s\n",
        vecp->ptr->dbl(0), vecp->ptr->dbl(1), vecp->ptr->dbl(2),
			    vecp->t->time().getISOstring().c_str());
    fflush(stdout);
    return;
}



