
/***********************************************************************
*           See Accompanying licence file QSAS_licence                 *
***********************************************************************/
/* PROJECT: Cluster UK CDHF
*
* COMPONENT:    Plug In library
* 
* MODULE:	DvSetAttribute.cc
*
* LANGUAGE:	C++
*
* FUNCTION:     QSAS Plugin to add an attribute, or change the string in an 
*				existing attribute
*
* PURPOSE:      As in FUNCTION
* 
* DESCRIPTION:  Within the Calculator, often users need to add or change a
*				text attribute, e.g., to fix UNITS, LABLAXIS, or specify
*				the FRAME.
*
* ARGUMENTS:
*    IN:      
*             inobj			Input data object (any type)
*             att_name		Name of the attribute to be changed/added
*			  att_ptr	 	Object or number or (un-quoted) string to be used as new attribute (value)
* 
*   OUT:
*             outobj		output object name. If left blank, changes are
*							made to the original WL object. Otherwise, the
*							original is untouched and changes are made to
*							a copy which is placed on the WL with this name.
*
* GLOBALS:    none
*
* RETURN:
*    TYPE:    QplugReturnStatus
*    VALUES:
*
*
*
* HISTORY:   SJS 21 Sep 2015
*/

#include "qplug_if.h"
#include "DvObj.h"

//#define DEBUG_ATTRIBUTES    // for testing

extern "C" QplugReturnStatus DvSetAttribute(vector<DvObject_var>&, vector<DvObject_var>&);

QplugReturnStatus DvSetAttribute(vector<DvObject_var>& inputs, vector<DvObject_var>& outputs)
{
	DvObject_var inobj_ptr, attName_ptr, att_ptr, outobj_ptr ;
		
	DvString out_name;				// name of output object
	DvString attName;

 	if(inputs.size() < 2) {
    	QplugAppendTextDisplay ("insufficient inputs\n");
    	return QPLUG_FAILURE; 
 	} 
 
	//	Unpack and check the input data objects
	inobj_ptr = inputs[0];
	if(inobj_ptr->is_nil()){
		QplugAppendTextDisplay ("Input object is empty");
		return QPLUG_FAILURE;
	}

	attName_ptr   = inputs[1];
	att_ptr = inputs[2];
	if (attName_ptr->is_nil() || att_ptr->is_nil()){
		QplugAppendTextDisplay ("Missing attribute name and/or string value inputs\n");
		return QPLUG_FAILURE;
	}
	
	attName=attName_ptr->asStr().trimmed();
	if(att_ptr->is_str()) att_ptr->str() = att_ptr->str().trimmed();

	out_name = GetSlotText(0);	

	// If no output obj specified work in inobj:
	if (out_name == "None"){
		QplugAppendTextDisplay ("No output object given, so will modify original object\n");		
		inobj_ptr->change_xref(attName, att_ptr);
		if(att_ptr->is_str()) QplugAppendTextDisplay ("Attribute "+attName+" changed to "+att_ptr->asStr()+"\n");
		else QplugAppendTextDisplay ("Attribute "+attName+" changed \n");
		
		return QPLUG_SUCCESS;	
	}
	
	// otherwise copy input object and work on that
	outobj_ptr = new DvObject(inobj_ptr);
	outobj_ptr->change_xref(attName, att_ptr);
	
	// that's it - just push back to WL and exit
	outputs.push_back(outobj_ptr);	
	if(att_ptr->is_str()) QplugAppendTextDisplay("Copy created with attribute "+attName+" = "+att_ptr->str()+"\n");
	else QplugAppendTextDisplay("Copy created with new attribute "+attName+"\n");
	
	return QPLUG_SUCCESS;
}


