/*************************************************************************
* PROJECT:      Cluster UK CDHF
* COMPONENT:    Plug In library
* MODULE:	XY2Polar.cc
* LANGUAGE:	C++
* FUNCTION:     QSAS XY2Polar Plug In
* PURPOSE:      Calculate radius and phase of two scalar series.  
* ARGUMENTS:
*    IN:    Scalar series X.  
*           Scalar series Y.
*   OUT:    Radius = SQRT(X^2+Y^2) Scalar series
*           Phase  = ATAN2(Y,X) Scalar series
* RETURN:   QplugReturnStatus
* FUNCTIONS_USED: 
* HISTORY: created  15/3/2001 Markus Fraenz 
*          corrected atan2(xx,yy) to atan2(yy,xx) AJA 18/08/2005
***********************************************************************/

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>

#include "qplug_if.h"
#include "qdutil.h"
#include "DvObj.h"

using namespace QSAS;


extern "C" QplugReturnStatus  DvXYtoPolar( vector<DvObject_var>& inputs, vector<DvObject_var>& outputs )
{
int  qar_status;
unsigned long	Ndata;
char str_out[150];
double twopi= 8.0 * atan(1.);

// Display initial message in QpluGui window 
QplugAppendTextDisplay("XY2Polar Plugin 1.1  Apr 2016\n");

// ...first, unpack the input data objects
DvObject_var X_in = inputs[0];  
DvObject_var Y_in = inputs[1];

//check whether these are scalar series, if not return.
if(X_in.is_nil() || X_in->not_dbl()){
   QplugAppendTextDisplay ("XY2Polar>> ERROR - X sequence type not handled\n");
   return QPLUG_FAILURE;  
}

if(Y_in.is_nil() || Y_in->not_dbl() ){
   QplugAppendTextDisplay ("XY2Polar>> ERROR - Y sequence type not handled\n");
   return QPLUG_FAILURE;  
}

Ndata = X_in->seqSize();


// check whether both have same units
if(!X_in->sameUnits(*Y_in)) {
  QplugAppendTextDisplay (
"Arithmetic>> Inputs not in same units (double click object and correct SI_conversion)\n");
  return QPLUG_FAILURE; 
}
  
//now tell user that plugin is actually executing
QplugAppendTextDisplay ("XY2Polar>> Calculating.\n");

//create output data objects
DvObject_var out_rad_series;
DvObject_var out_phas_series;
DvObject_var out_phas_series_0_360;

//calculate radius and phase

  DvObject_var x2 = X_in*X_in;
  DvObject_var y2 = Y_in*Y_in;
  DvObject_var r2 = x2 + y2;
  out_rad_series = r2->sqrt();
  out_phas_series = Y_in->atan2(X_in);

//set minimal attributes by copying from input objects
  sprintf(str_out,"r(%s)", X_in->getXrefText("FIELDNAM").c_str());
  out_rad_series->change_xref(LABLAXIS, str_out);
  out_rad_series->change_xref(FIELDNAM, str_out);
  DvString units = X_in->getXrefText(UNITS);
  out_rad_series->change_xref(UNITS, units);
  out_rad_series->delete_xref(DEPEND_1);
  
  out_phas_series->change_xref(SI_CONVERSION, "1>rad");
  sprintf(str_out,"Phase(%s)", X_in->getXrefText("FIELDNAM").c_str());
  out_phas_series->change_xref(LABLAXIS, str_out);
  out_phas_series->change_xref(FIELDNAM, str_out);
  out_phas_series->change_xref(UNITS, "rad");
  out_phas_series->delete_xref(DEPEND_1);

  out_phas_series_0_360 = new DvObject(out_phas_series);
  for(size_t i=0; i<out_phas_series_0_360->totalSize(); i++){
    if(out_phas_series_0_360->dbl(i) < 0) out_phas_series_0_360->dbl(i) += twopi;
  }
  
  //add output objects to  QSAS working list
  outputs.push_back( out_rad_series);    
  outputs.push_back( out_phas_series);     
  outputs.push_back( out_phas_series_0_360);     


//tell user that Plug In finished
sprintf(str_out, "XY2Polar>> %ld data points\n", Ndata);
QplugAppendTextDisplay (str_out);
QplugAppendTextDisplay ("XY2Polar>> finished...\n");


return QPLUG_SUCCESS;
}



