#include "PlotWindow.h"
#include <QPrinterInfo>
#include <QPrintDialog>
#include <QGroupBox>

#include <cmath>

#include <QMessageBox>

#if QT_VERSION > 0x050000
#include <QPageSize>
#endif

#include "qwui.h"
extern qwuiWindow* qsasWindow;  

	 
PlotWindow::PlotWindow(int plot_id, int width, int height, double saveWidth, double saveHeight, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation, bool withGUI, QWidget* parent)
	:QMainWindow(parent)
{
    extraLength = 0;
    extraBottom = 0;
    extraSide = 0;
    az = 0;
    elev = 0;
    
    percent = 100;
	QMenu * plotMenu = menuBar()->addMenu("Plot");

    char *qsas_home_env = getenv("QSAS_HOME"); 
    QString qsas_home=NULL;
    if( qsas_home_env != NULL) {
        qsas_home = QString(qsas_home_env);
        qsas_home = qsas_home.remove(QString("\""));
    } else qsas_home="";
    
	plotMenu->addAction(QIcon(qsas_home+"/icons/png.png"), "Save Plot As PNG", this, SLOT(savePlotPNG()));
    plotMenu->addAction(QIcon(qsas_home+"/icons/jpg.png"), "Save Plot As JPG", this, SLOT(savePlotJPG()));
	plotMenu->addAction(QIcon(qsas_home+"/icons/svg.png"), "Save Plot As SVG", this, SLOT(savePlotSVG()));
	plotMenu->addAction(QIcon(qsas_home+"/icons/pdf.png"), "Save Plot As PDF", this, SLOT(savePlotPDF()));
//	plotMenu->addAction("Save Plot As EPS", this, SLOT(savePlotPS()));

    plotMenu->addSeparator();
	plotMenu->addAction(QIcon(qsas_home+"/icons/fileprint.png"), "Print", this, SLOT(printPlot()), Qt::CTRL+Qt::Key_P);
    plotMenu->addSeparator();
	toolBar = plotMenu->addAction("Show Tool Bar", this, SLOT(toggleToolbar()), Qt::CTRL+Qt::Key_T);

    
    compressionQuality = menuBar()->addMenu("JPG Quality (Medium)");
    compressionQuality->addAction("Low", this, SLOT(setQL()));
    compressionQuality->addAction("Medium Low", this, SLOT(setQML()));
    compressionQuality->addAction("Medium", this, SLOT(setQM()));
    compressionQuality->addAction("Medium High", this, SLOT(setQMH()));
    compressionQuality->addAction("High", this, SLOT(setQH()));
    compressionQuality->setToolTip("Quality of jpg files. Higher quality gives larger files.");
    
	printWeightMenu = menuBar()->addMenu("Print Weight (M)");
    printWeightMenu->addAction("Pale", this, SLOT(setWeightP()));
    printWeightMenu->addAction("Light", this, SLOT(setWeightL()));
	printWeightMenu->addAction("Medium", this, SLOT(setWeightM()));
    printWeightMenu->addAction("Bold", this, SLOT(setWeightB()));
    printWeightMenu->addAction("Heavy", this, SLOT(setWeightH()));
    printWeightMenu->addAction("V Heavy", this, SLOT(setWeightVH()));
	
	setWeightM();
	
	this->saveWidth=saveWidth;
	this->saveHeight=saveHeight;
	this->pageSize=pageSize;
	this->orientation=orientation;
	this->resolution=resolution;
	
    QGroupBox * mainWidget=new QGroupBox(this);
    mainLayout=new QGridLayout(mainWidget);
    mainWidget->setObjectName("PlotMainWindow");
    mainLayout->setObjectName("PlotMainLayout");
    
    setCentralWidget(mainWidget);
    
    if(orientation == QPrinter::Landscape)   plot=new QtPLWidget(width, height, saveWidth, saveHeight, resolution, this);
    else    plot=new QtPLWidget(height, width, saveWidth, saveHeight, resolution, this);
    mainLayout->addWidget(plot, 0, 0);
    mainLayout->setRowStretch(0, 100);
    mainLayout->setColumnStretch(0, 100);

    mainLayout->setContentsMargins(0,0,0,0);

    if(withGUI){
        QGroupBox *GUIgroupBox = new QGroupBox;
    
        layout=new QGridLayout;
        layout->setObjectName("GUILayout");
        GUIgroupBox->setLayout(layout);
    
        mainLayout->addWidget(GUIgroupBox, 2, 0, Qt::AlignCenter);
        mainLayout->setRowStretch(1, 1);
        layout->setContentsMargins(0,0,0,0);
    }
	connect(plot, SIGNAL(resized(double)), this, SLOT(changeDisplayedRatio(double)));
        
    setSize(width, height);
    setId(plot_id);
    plmkstrm(&strm);
    plsetqtdev((void*)plot);
    plsdev ("qt");
    plwid(1);
    plinit();
	
    connect(qsasWindow,  SIGNAL( directPrintRequest(const char *)), this, SLOT(savePlotDirect(const char *)));
    
	connect(plot, SIGNAL(clicked(double, double, int, int)), this, SIGNAL(clicked(double, double, int, int))); // redirection of widget "clicked" events
	connect(plot, SIGNAL(doubleClicked(double, double, int, int)), this, SIGNAL(doubleClicked(double, double, int, int))); // redirection of widget "doubleClicked" events
	
	intervalDialog=new IntervalDialog(this);
	intervalDialog->hide();

	intervalDialog->setVisible(false);
	
	connect(intervalDialog, SIGNAL(rescaleScalar(double, double)), this, SIGNAL(rescaleScalar(double, double)));
	connect(intervalDialog, SIGNAL(rescaleTime(DvEvent)), this, SIGNAL(rescaleTime(DvEvent)));
	
	qApp->processEvents(); // ensure initialisation is finished
    
    ownerGUI=NULL;

    createActions();
    createToolBars(); 
}

void PlotWindow::toggleToolbar(){ 
    if(_file_tool_bar->isVisible()){
        _file_tool_bar->setVisible(false);
        toolBar->setText("Show Tool Bar");
    }
    else{
        _file_tool_bar->setVisible(true);
        toolBar->setText("Hide Tool Bar");
    }
    
}

// ---------------------------------------------------
void PlotWindow::createToolBars()
{
  _file_tool_bar = addToolBar("Plot");
  _file_tool_bar->setIconSize(QSize(24, 24));
  _file_tool_bar->addAction(printAct);
  _file_tool_bar->addAction(printDirectAct);

  _file_tool_bar->insertSeparator(pdfAct);  
  _file_tool_bar->addAction(pngAct);	
  _file_tool_bar->addAction(svgAct);	
  _file_tool_bar->addAction(jpgAct);	
  _file_tool_bar->addAction(pdfAct);
  _file_tool_bar->setVisible(false);
  
//  if(QSAS_profiles->current()->ShowToolbar) toggleToolbar();

}

// ---------------------------------------------------SEH Toolbar
void PlotWindow::createActions()
{
  char *qsas_home_env = getenv("QSAS_HOME"); 
  QString qsas_home=NULL;
  if( qsas_home_env != NULL) {
    qsas_home = QString(qsas_home_env);
    qsas_home = qsas_home.remove(QString("\""));
  } else qsas_home="";

  pdfAct = new QAction(QIcon(qsas_home+"/icons/pdf.png"), tr("Save as PDF"), this);
  connect(pdfAct, SIGNAL(triggered()), this, SLOT(savePlotPDF()));

  pngAct = new QAction(QIcon(qsas_home+"/icons/png.png"), tr("Save as PNG"), this);
  connect(pngAct, SIGNAL(triggered()), this, SLOT(savePlotPNG()));

  svgAct = new QAction(QIcon(qsas_home+"/icons/svg.png"), tr("Save as SVG"), this);
  connect(svgAct, SIGNAL(triggered()), this, SLOT(savePlotSVG()));

  jpgAct = new QAction(QIcon(qsas_home+"/icons/jpg.png"), tr("Save as JPEG"), this);
  connect(jpgAct, SIGNAL(triggered()), this, SLOT(savePlotJPG()));

  printAct = new QAction(QIcon(qsas_home+"/icons/fileprint.png"), tr("Print... "), this);
  connect(printAct, SIGNAL(triggered()), this, SLOT(printPlot()));

  printDirectAct = new QAction(QIcon(qsas_home+"/icons/fileprintdirect.png"), tr("Send to default printer "), this);
  connect(printDirectAct, SIGNAL(triggered()), this, SLOT(printPlotDirect()));
}

//-------------------------------------------

void PlotWindow::addWidget(QWidget *w, int r, int c, Qt::Alignment align, int numRows){
    layout->addWidget(w, r, c, align);
    // extend window height to accomodate widgets
    int rowPosn =30*(r+numRows);
    if( extraLength <  rowPosn) extraLength = rowPosn;
    
    setSize(saveWidth, saveHeight, true);
}

QSlider * PlotWindow::addSlider(Qt::Orientation hv, int from, int to, int step){
    
    QSlider *newSlider = new QSlider(hv, this);
    newSlider->setRange(from, to);
    newSlider->setSingleStep(step);

    if(hv == Qt::Vertical) {
        mainLayout->addWidget(newSlider, 0, 1, Qt::AlignLeft);
        extraSide = 20;
        connect(newSlider, SIGNAL(valueChanged(int)), this, SLOT(setElev(int)));
    }
    else {
        mainLayout->addWidget(newSlider, 1, 0, Qt::AlignTop);
        extraBottom = 20;
        connect(newSlider, SIGNAL(valueChanged(int)), this, SLOT(setAz(int)));
    }

    setSize(saveWidth, saveHeight, true);
    return newSlider;
}


void PlotWindow::setPrintRes(int res)
{
    plot->setPrintRes(res);
    resolution = res;
}

void PlotWindow::setWeightP()
{
    weight = .75;
    printWeightMenu->setTitle("Print Weight (Pale)");
}

void PlotWindow::setWeightL()
{
    weight = 1.5;
    printWeightMenu->setTitle("Print Weight (Light)");
}

void PlotWindow::setWeightM()
{
    weight = 2.5;
	printWeightMenu->setTitle("Print Weight (Medium)");
}

void PlotWindow::setWeightB()
{
    weight = 3.5;
    printWeightMenu->setTitle("Print Weight (Bold)");
}

void PlotWindow::setWeightH()
{
    weight = 5.0;
    printWeightMenu->setTitle("Print Weight (Heavy)");
}

void PlotWindow::setWeightVH()
{
    weight = 10.0;
    printWeightMenu->setTitle("Print Weight (Very Heavy)");
}

void PlotWindow::setPLFocus()
{
    plsstrm(strm);    
}

void PlotWindow::setTitle(const QString& _title)
{
    this->title=_title;
    setWindowTitle(QString("%1 (%2%)").arg(title).arg(percent));
}

void PlotWindow::setAz(int value)
{
    az = value;
    setWindowTitle(QString("%1 (%2%)            (Az %3 deg, Elev %4 deg)").arg(title).arg(percent).arg(az).arg(elev));
}
void PlotWindow::setElev(int value)
{
    elev = value;
    setWindowTitle(QString("%1 (%2%)            (Az %3 deg, Elev %4 deg)").arg(title).arg(percent).arg(az).arg(elev));
}

void PlotWindow::changeDisplayedRatio(double r)
{
    percent = (int) (r*100);
	setWindowTitle(QString("%1 (%2%)").arg(title).arg((int)(r*100.)));
}

                        
void PlotWindow::setSmooth(bool b)
{
    if(plot) plot->setSmooth(b);
}

void PlotWindow::setId(int plid){this->plot_id=plid;}
                        
int PlotWindow::getId(){return plot_id;}

void PlotWindow::setSize(double width, double height, bool forceResize)
{
    bool mustResize=(width!=saveWidth || height!=saveHeight || orientation != saveOrientation || forceResize);
    saveWidth=width;
    saveHeight=height;
    saveOrientation=orientation;
    PLINT old_strm;
    plgstrm(&old_strm);
    plsstrm(strm);
    plsstrm(old_strm);
    
    if(mustResize)
    {
        int useWidth = width;
        int useHeight = height;


        if(orientation == QPrinter::Portrait) {
            if(useWidth>1000){
                useWidth = 1000;
                useHeight = useWidth * height / width;
                percent =  (int) (useWidth / width * 100);
            }
            
            if(useHeight>700){
                useHeight = 700;
                useWidth = useHeight * width / height;
                percent =  (int) (useHeight / height * 100);
            }

            resize(useWidth + extraSide, useHeight + extraLength + extraBottom);
            plot->setNewSaveSize(width, height);
        }
        else  {
            if(useWidth>700){
                useWidth = 700;
                useHeight = useWidth * height / width;
                percent =  (int) (useWidth / width * 100);
            }
            
            if(useHeight>1000){
                useHeight = 1000;
                useWidth = useHeight * width / height;
                percent =  (int) (useHeight / height * 100);
            }
            resize(useHeight + extraSide, useWidth + extraLength + extraBottom);
            plot->setNewSaveSize(height, width);
        }
    }
}


void PlotWindow::savePlotPNG(QString saveName)
{
	if(saveName.isEmpty() ) saveName=QFileDialog::getSaveFileName(this, "Save plot as...", saveDirectory, "PNG file (*.png)");
	
	if(saveName.isEmpty()) return;

	if(saveName.contains("/") ) saveDirectory = saveName.section('/', 0, -2);

	if(!saveName.endsWith(".png", Qt::CaseInsensitive))
	{
		saveName.append(".png");
	}

    double scaleWt =  weight * resolution / 300.;

    if(orientation == QPrinter::Landscape)  plot->savePixmap(saveName, "PNG", (int)saveHeight, (int)saveWidth, 10, scaleWt);
    else  plot->savePixmap(saveName, "PNG", (int)saveWidth, (int)saveHeight, 10, scaleWt);
}

void PlotWindow::savePlotJPG(QString saveName)
{
    if(saveName.isEmpty() ) saveName=QFileDialog::getSaveFileName(this, "Save plot as...", saveDirectory, "JPG file (*.jpg *.jpeg)");
    
    if(saveName.isEmpty()) return;
    
    if(saveName.contains("/") ) saveDirectory = saveName.section('/', 0, -2);
    
    if(!saveName.endsWith(".jpg", Qt::CaseInsensitive) && !saveName.endsWith(".jpeg", Qt::CaseInsensitive))
    {
        saveName.append(".jpg");
    }
    

    double scaleWt;

     scaleWt =  weight * resolution / 300.;

    bool ok;
    if(orientation == QPrinter::Landscape)  ok = plot->savePixmap(saveName, "JPEG", (int)saveHeight, (int)saveWidth, qual, scaleWt);
    else   ok = plot->savePixmap(saveName, "JPEG", (int)saveWidth, (int)saveHeight, qual, scaleWt);

    if(!ok){
        QMessageBox msgBox;
        msgBox.setText("problem writing JPEG format");
        msgBox.exec();	
    }
    
}


/*
void PlotWindow::savePlotPS()
{
	QString saveName=QFileDialog::getSaveFileName(this, "Save plot as...",  saveDirectory, "Encapsulated Postscript file (*.eps)");

	if(saveName.isEmpty()) return;

	if(saveName.contains("/") ) saveDirectory = saveName.section('/', 0, -2);

	if(!saveName.endsWith(".ps", Qt::CaseInsensitive) && !saveName.endsWith(".eps", Qt::CaseInsensitive))
	{
		saveName.append(".eps");
	}
	
	QPrinter p;
	p.setResolution(resolution);
	p.setOutputFileName(saveName);
	p.setOrientation(orientation);
	p.setColorMode(QPrinter::Color);
	
	p.setOutputFormat(QPrinter::PostScriptFormat);

	
	p.setPageSize(pageSize);

	fixAspectRatio(saveWidth, saveHeight);

	if(orientation == QPrinter::Landscape) p.setPaperSize(QSizeF(saveHeight, saveWidth), QPrinter::DevicePixel);
	else p.setPaperSize(QSizeF(saveWidth, saveHeight), QPrinter::DevicePixel);

#endif

	double plotFactor = (saveWidth + saveHeight) / 2000.;
	plot->print(p, weight*plotFactor); 

}
*/

void PlotWindow::fixAspectRatio(double &saveWidth, double &saveHeight)
{
	// ensure aspect ratio is not too large for pdf,ps and svg drivers
	
	double aspect = saveWidth/saveHeight;
	
	// reset height if ratio too great
	if(aspect > 2.5 ) {
		saveHeight = saveWidth * 0.4;
		return;
	}
	else if( aspect < 0.4) {
		saveHeight = saveWidth * 2.5;
	}
	
}

void PlotWindow::savePlotSVG(QString saveName)
{
			
	if(saveName.isEmpty()) saveName=QFileDialog::getSaveFileName(this, "Save plot as...", saveDirectory, "SVG file (*.svg)");

	if(saveName.isEmpty()) return;

	if(saveName.contains("/") ) saveDirectory = saveName.section('/', 0, -2);

	if(!saveName.endsWith(".svg", Qt::CaseInsensitive))
	{
		saveName.append(".svg");
	}
	
	fixAspectRatio(saveWidth, saveHeight);

    if(orientation == QPrinter::Landscape)  plot->saveSVG(saveName, (int)saveHeight, (int)saveWidth, resolution, weight);
    else   plot->saveSVG(saveName, (int)saveWidth, (int)saveHeight, resolution, weight);

	
	this->hide();
	this->show();		// force repaint on Mac


	

}

void PlotWindow::savePlotPDF(QString saveName)
{
    if(saveName.isEmpty())  saveName=QFileDialog::getSaveFileName(this, "Save plot as...", saveDirectory, "PDF document (*.pdf)");
    
    if(saveName.isEmpty()) return;
    
    if(saveName.contains("/") ) saveDirectory = saveName.section('/', 0, -2);
    
    if(!saveName.endsWith(".pdf", Qt::CaseInsensitive))
    {
        saveName.append(".pdf");
    }
    
    QPrinter p;
    p.setResolution(resolution);
    p.setOutputFileName(saveName);
    p.setOrientation(orientation);
    p.setColorMode(QPrinter::Color);
    
    
    
    fixAspectRatio(saveWidth, saveHeight);
    double plotFactor;

#ifdef WINDOWS
    plotFactor = 0.002 / 72. * resolution;
    p.setOutputFormat(QPrinter::PdfFormat);
    p.setPaperSize(QSizeF(saveWidth, saveHeight), QPrinter::DevicePixel);
#else
    plotFactor = 1.2; // with PdfFormat use .002    
    p.setOutputFormat(QPrinter::NativeFormat); // seems to work better (thinner plot lines)
    p.setPaperSize(QSizeF(saveWidth/72*resolution, saveHeight/72*resolution), QPrinter::DevicePixel);
#endif
    plot->print(p, weight*plotFactor);
}

void PlotWindow::printPlot()
{
    QPrinter printer;
    QPrintDialog printDialog(&printer, this);

    printer.setResolution(resolution);
	printer.setOrientation(orientation);
	printer.setDocName("QSAS_plot");
	printer.setColorMode(QPrinter::Color);
    printer.setOutputFormat(QPrinter::NativeFormat);
    
    
	// vector graphics so rescalable.

	fixAspectRatio(saveWidth, saveHeight);
    double plotWeight=weight;
    
#ifdef WINDOWS
    plotWeight = weight*15;
#else
    plotWeight = weight*1.5;
#endif

    
#if !defined(WINDOWS) && !defined(Q_OS_MACX)
	if(QSAS_profiles->current()->UseKPrinter)
	{
        
        if(orientation == QPrinter::Landscape) printer.setPaperSize(QSizeF(saveHeight/resolution, saveWidth/resolution), QPrinter::Inch);
        else printer.setPaperSize(QSizeF(saveWidth/resolution, saveHeight/resolution), QPrinter::Inch);
		printer.setPrintProgram(QString("kprinter"));
		printer.setPrinterName("foo"); // Requires a name. Ideally a printer name, but anything is fine in practice.
		plot->print(printer, plotWeight);
	}
	else
	{
        QPrintDialog printDialog(&printer, this);
		if (printDialog.exec() == QDialog::Accepted)
		{
        	printer.setPrintProgram(QString("lpr"));
			plot->print(printer, plotWeight);
		}
	}
#else

    if (printDialog.exec() == QDialog::Accepted)
    {
        plot->print(printer, plotWeight);
    }
#endif

	
}



void PlotWindow::savePlotDirect(const char * saveName)
{
    DvString dvname = saveName;
    int pageID = this->getId();
    if(pageID != 1) {
        DvString extn = dvname.after('.');
        dvname = dvname.before('.');
        dvname += pageID;
        dvname += '.';
        dvname += extn;
    }
    
    QString name = dvname.c_str();
    
    if(name.isEmpty()) return;
    
    if(name.endsWith(".pdf", Qt::CaseInsensitive)) {
        savePlotPDF(name);
    }
    else if(name.endsWith(".png", Qt::CaseInsensitive)) {
        savePlotPNG(name);
    } 
    else if(name.endsWith(".jpg", Qt::CaseInsensitive)) {
        savePlotJPG(name);
    }
    else if(name.endsWith(".svg", Qt::CaseInsensitive)) {
        savePlotSVG(name);
    }

}

// ----------------------------


PlotID::PlotID(const char* _title, QList<PlotWindow*> *_plotWindows){
    
    title = _title;
    if(_plotWindows == NULL) {
        plotWindows = new QList<PlotWindow*>;
        delListOnClose = true;
    }
    else{
        plotWindows = _plotWindows;
        delListOnClose = false;
    }
    
    currentWindow = NULL;
    const QString newWin("Use New Window");
    this->insertItem(plotWindows->size(), newWin);

    connect(this, SIGNAL(currentIndexChanged(int)), this, SLOT(setWindow(int)));
}

DvString PlotID::getName(){
    
    DvString name = title;
    name += " ";
    if(currentWindow) name += currentWindow->getId();
    return name;
}

void PlotID::addWindow(PlotWindow *w){
    
    plotWindows->push_back(w);
    
    this->clear();
    
    const QString newWin("Use New Window");
    for(size_t i=0; i<plotWindows->size(); ++i)
    {
        DvString winID =  title;
        winID += " ";
        winID += (*plotWindows)[i]->getId();
        this->insertItem(i, winID.c_str());
        
    }
    this->insertItem(plotWindows->size(), newWin);

    currentWindow = w;
}

PlotWindow*  PlotID::focusWindow(double width, double height, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation, bool withGUI, bool hidden){
 
    PlotWindow *win = this->currentWindow;

    if(win)
    {
        if(width>0. && height > 0.)
        {
            win->resolution=resolution;
            win->pageSize=pageSize;
            win->orientation=orientation;
            win->setSize(width, height);
        }
        if(!hidden) win->show();
        win->setPLFocus();
        win->activateWindow();
        if(!hidden) win->raise();
        return win;
    }
    
    // If reached this point, the window does not exist.
    // Force creation of Window with this id
    win = this->newPlotWindow(width, height, resolution, pageSize, orientation, withGUI, hidden);
    
    return win;

}

    
void PlotID::setWindow(int index){

    if(index < plotWindows->size() && index >=0 && plotWindows->size() > 0 )
        currentWindow = (*plotWindows)[index];
    else
        currentWindow = NULL; // force new Window
    
    return;
}


PlotWindow* PlotID::newPlotWindow(double width, double height, int resolution, QPrinter::PageSize pageSize, QPrinter::Orientation orientation, bool withGUI, bool hidden)
{
    // find free id, allows for deletions from list
    int plot_id=1;
    while( this->IDused(plot_id) ) plot_id++;

    PlotWindow* res=NULL;
    res=new PlotWindow(plot_id, width, height, height, width, resolution, pageSize, orientation, withGUI, NULL);
    
    this->addWindow(res);
 
    res->ownerGUI = this;

    this->setCurrentIndex(this->size()-1);
    
    res->setTitle(this->getName().c_str());
    if(!hidden) res->show();

    qApp->processEvents();
	
    return res;
}


void PlotWindow::printPlotDirect()
{
  QString qsas_printer = QSAS_profiles->current()->defPrinter;

  printf("Print direct to %s\n",(const char*) qsas_printer.toLatin1().data());

  QPrinter printer;
  printer.setResolution(resolution);
  printer.setPageSize(pageSize);
  printer.setOrientation(orientation);
  printer.setColorMode(QPrinter::Color);
  printer.setPrinterName(qsas_printer); 
  plot->print(printer,weight);
}



