// qpui_event.cc
//
// Plot Interface Implementation - Event Object
//
// Janet Barnes
// July 2002
//
// Tony Allen
// April 2008

#include "qpui.h"
#include "qwui.h"

extern int qsas_restore_release;

EventItem::EventItem(const QString& name, PanelItem* parent, const QString& WLcontext, int position):
		PlotItem(name, parent, WLcontext, position)
{	
	I_Type = QP_EVENT;

	eventTab = new QpEventTab(tabWidget, "Event Tab", this);
	eventTab->hide();

	// set default values for Event object

	eventTab->_char_size_fld->setValue(QSAS_profiles->current()->BaseFontSize+QPUI_LABELS_CHAR_SIZE_ADD);
// 	eventTab->_event_label_font_fld->setCurrentItem(QPUI_NORMAL);
	_event_label_colour = Qt::black;
	
	tabWidget->adjustSize();
	setIcon(QP_EVENT);
}

EventItem::EventItem(EventItem* item, PanelItem* parent, int position):
		PlotItem(item, parent, position)
{
	I_Type = QP_EVENT;

	_event_label_colour = item->_event_label_colour;
	
	eventTab = new QpEventTab(item->eventTab, this);
	eventTab->hide();
	
	tabWidget->adjustSize();
	setIcon(QP_EVENT);
}

EventItem::~EventItem()
{
	delete eventTab;
	foreach(PlotItem* i, linkedPlots)
	{
		i->linkedPlots.removeAll(this);
	}
}


// -------------------------------------------

void EventItem::setDataYObj(QString txt)
{
	if(txt.isEmpty()) return;

	eventTab->_y_data_fld->setText(txt);
	eventTab->_y_data_fld->reValidate(true);
}

// -------------------------------------------

void EventItem::setDataXObj(QString txt)
{
	if(txt.isEmpty()) return;

	eventTab->_x_data_fld->setText(txt);
	eventTab->_x_data_fld->reValidate(true);
}

// -------------------------------------------

void EventItem::showTabs()
{
	// ensure all tabs are hidden
	int Ntabs = tabWidget->count();
	for(int i = Ntabs-1; i >= 0; i--) 
	{
		if(tabWidget->widget(i))
		{
			tabWidget->widget(i)->hide();
			tabWidget->removeTab(i);
		}
	}

	tabWidget->addTab(eventTab, "Event Tab");
	
	// value return is always shown
	tabWidget->addTab(page()->ValueReturnTab, "Value Return");
	tabWidget->setCurrentIndex(tabWidget->indexOf(eventTab));
	
}

// -----------------------------------------------

QpuiReturnStatus EventItem::plot()
{
	bool OK;
	bool backwards=false;
	double y_posn;
	double x_posn;
	PLFLT x_data;
	PLFLT y_data;

	DvObject_var eventTxtObj = eventTab->_event_label_fld->getObject();
	if(eventTxtObj.is_nil() || !eventTxtObj->is_str()) return QPUI_SUCCESS;

	plwid(2);
	
	// get y position
	DvObject_var ydo = eventTab->_y_data_fld->getObject();
	
	if(ydo->is_dbl()){
		y_posn = ydo->dbl(0);
	}
	else {
		std::cout << " No y posn " << std::endl;
		plmtex("lv",-10.0,0.5,0.0,"Plot ERROR: Y Data not a Scalar Object");
		std::cout << " No y posn plotted" << std::endl;

		return QPUI_FAILURE;
	}
    
	if(panel()->yAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) y_posn = SafeLog10(y_posn);
		
	y_data = (PLFLT) y_posn;
	
	
	// get x or t position
	
	DvObject_var xdo = eventTab->_x_data_fld->getObject();
	
	if(frameType() == QP_TS)
	{
		// Time axis
		DvTime time_t;
		if(xdo->is_time()){
			time_t = xdo->time(0);
		}
		else {	
			DvObject_var tt = xdo->getTimeTags();

			if(tt.is_nil())
			{
				plmtex("lv",-10.0,0.5,0.0,"Event Label: Invalid T Data");
				return QPUI_FAILURE;
			}
			else
			{
				time_t = tt->time(0);
			}
		}

		DvTime t_plot_start(frame()->tAxisTab->_axis_time_fld->getStart().getEpoch2000());
		x_data=time_t.getDiffSecs(t_plot_start);

	}
	else
	{
		// X axis
		if( xdo->is_dbl())
		{
			x_posn = xdo->dbl(0);
		}
		else
		{	
			plmtex("lv",-10.0,0.5,0.0,"Event Label: Invalid X Data");
			return QPUI_FAILURE;
		}
					
		if(frame()->xAxisTab->_axis_linlog_fld->currentIndex() == QP_LOG) x_posn = SafeLog10(x_posn);
		x_data = (float) x_posn;
	}
	
	// Not clipped => Extend panel range and viewport to whole page, write label, and reduce panel and viewport back

	PLFLT save_min_x, save_max_x, save_min_y, save_max_y; // viewport coordinates in page
	plgvpd(&save_min_x, &save_max_x, &save_min_y, &save_max_y);
		
	PLFLT save_min_x_data, save_max_x_data, save_min_y_data, save_max_y_data; // data ranges in original viewport
	plgvpw(&save_min_x_data, &save_max_x_data, &save_min_y_data, &save_max_y_data);

	double val0x, val1x, val0y, val1y; // data range when viewport is whole page
		
	double ratiox=(save_max_x_data-save_min_x_data)/(save_max_x-save_min_x); // range of time for the whole page width
	val0x=save_min_x_data-save_min_x*ratiox;
	val1x=val0x+ratiox;	

	double ratioy=(save_max_y_data-save_min_y_data)/(save_max_y-save_min_y); // range of time for the whole page width
	val0y=save_min_y_data-save_min_y*ratioy;
	val1y=val0y+ratioy;

	// We set up the new viewport for the whole page
	plvpor(0., 1., 0., 1.);
	plwind(val0x, val1x, val0y, val1y);
	
	if(val0x > val1x) backwards = true;

	plschr(QPUI_CHAR_HEIGHT,eventTab->_char_size_fld->value());
	// set font
	PLINT label_colour_index = getPGPColourIndex(_event_label_colour);
	plcol0(label_colour_index);


	PLINT family, style, weight;
#if QT_VERSION >= 0x040300
	eventTab->_event_label_font_fld->getFontParams(family, style, weight);
	plsfont(family, style, weight);
#else
    family=eventTab->_event_label_font_fld->getFontIndex();
    plfont(family);
#endif
	page()->setSmooth(true);

    if(backwards) plptex(x_data, y_data, -1., 0.0, eventTab->_event_label_just_fld->value(), SafePGPString(eventTxtObj->strS().c_str()).c_str());
	else  plptex(x_data, y_data, 1., 0.0, eventTab->_event_label_just_fld->value(), SafePGPString(eventTxtObj->strS().c_str()).c_str());
	page()->setSmooth(false);

	// revert to original viewport
	plvpor(save_min_x, save_max_x, save_min_y, save_max_y);
	plwind(save_min_x_data, save_max_x_data, save_min_y_data, save_max_y_data);

	return QPUI_SUCCESS;
}


// -----------------------------------------------------

QDomElement EventItem::toXMLElement(QDomDocument& doc) const
{
    QDomElement xml=doc.createElement("Plot_Item");
    xml.setAttribute("Type", "Event");
    xml.appendChild(createTextXMLNode(doc, "Name", getName()));
    
    xml.appendChild(eventTab->toXMLElement(doc));

    xml.appendChild(createTextXMLNode(doc, "Label_Red", _event_label_colour.red()));
    xml.appendChild(createTextXMLNode(doc, "Label_Green", _event_label_colour.green()));
    xml.appendChild(createTextXMLNode(doc, "Label_Blue", _event_label_colour.blue()));

    return xml;
}

extern QPalette QPUI_normal_palette;

void EventItem::fromXMLElement(QDomElement& el)
{
    QString s;
    
    if(getXMLNodeText(el, "Name", s))
    {
        setName(s);
    }
    
    QDomElement eventElement=el.firstChildElement("Event_Tab");
    if(!eventElement.isNull())
    {
        eventTab->fromXMLElement(eventElement);
    }

    int r(0), g(0), b(0);
    if(getXMLNodeText(el, "Label_Red", s))
    {
        r=s.toInt();
    }
    if(getXMLNodeText(el, "Label_Green", s))
    {
        g=s.toInt();
    }
    if(getXMLNodeText(el, "Label_Blue", s))
    {
        b=s.toInt();
    }
    _event_label_colour = QColor(r,g,b);
    
    QPalette line_palette = QPalette(QPUI_normal_palette);
    line_palette.setColor(QPalette::Active ,QPalette::ButtonText, _event_label_colour);
    line_palette.setColor(QPalette::Inactive, QPalette::ButtonText, _event_label_colour);
    eventTab->_event_label_colour_fld->setPalette(line_palette);

    eventTab->_y_data_fld->reValidate();
    eventTab->_x_data_fld->reValidate();

}
