// qpui_globals.cc
//
// Plot Interface Implementation - Global functions and variables
//
// Janet Barnes
//
// July 2002

#include "qpui.h"
#include <QPushButton>

#if QT_VERSION < 0x040000
#include <qdatetime.h>
#else
#include <QDateTime>
#endif

QpWindow *QPUI_widget=static_cast<QpWindow*>(NULL);

// global variable for keeping track of colour indices used
int QPUI_colour_index;

// global variables for storing colour palettes for showing
// active and inactive fields
QPalette QPUI_normal_palette;

ColourTableButton::ColourTableButton(QWidget* parent):
        QPushButton(parent)
{
    menu=new QMenu(this);
    setMenu(menu);
    updateMenuEntries();
    setCurrentTable(stdPalette[0]);
    connect(QPUI_widget, SIGNAL(refreshColourTableButtonsSignal()), this, SLOT(updateMenuEntries()));
}

void ColourTableButton::updateMenuEntries()
{
    menu->clear();

    profilePalette.clear();
    foreach(ColourTable table, QSAS_profiles->current()->palette)
    {
        profilePalette.push_back(table);
    }

    // Add standard palette

    for(int i=0; i<stdPalette.size(); ++i)
    {
        QAction* action=menu->addAction(stdPalette[i].icon, stdPalette[i].name);
        QList<QVariant> data;
        data.push_back(0);
        data.push_back(i);
        action->setData(data);
    }

    connect(menu, SIGNAL(triggered(QAction*)), this, SLOT(selected(QAction*)));

    if(profilePalette.size()>0)
    {
        menu->addSeparator();

        // Add profile palette
        QMenu* profileMenu=menu->addMenu("Profile tables");

        for(int i=0; i<profilePalette.size(); ++i)
        {
            QAction* action=profileMenu->addAction(profilePalette[i].icon, profilePalette[i].name);
            QList<QVariant> data;
            data.push_back(1);
            data.push_back(i);
            action->setData(data);
        }

    }

    if(importedPalette.size()>0)
    {
        menu->addSeparator();
        // Add imported palette

        QMenu* sessionMenu=menu->addMenu("Imported tables");

        for(int i=0; i<importedPalette.size(); ++i)
        {
            QAction* action=sessionMenu->addAction(importedPalette[i].icon, importedPalette[i].name/*, this, SLOT(stdSelected(int))*/);
            QList<QVariant> data;
            data.push_back(2);
            data.push_back(i);
            action->setData(data);
        }

//         connect(sessionMenu, SIGNAL(triggered(QAction*)), this, SLOT(sessionSelected(QAction*)));
    }

    if(!stdPalette.contains(currentTable) && !profilePalette.contains(currentTable) && !importedPalette.contains(currentTable))
    {
        setCurrentTable(stdPalette[0]);
    }
}

void ColourTableButton::setStdPalette(int index)
{
    if(index>=0 && index<stdPalette.size())
    {
        setCurrentTable(stdPalette[index]);
    }
    else
    {
        setCurrentTable(stdPalette[0]);
    }
}

void ColourTableButton::selected(QAction* action)
{
    QList<QVariant> list=action->data().toList();
    int type=list.first().toInt();
    list.pop_front();
    int index=list.first().toInt();
    switch(type)
    {
        case 1:
            setProfilePalette(index);
            emit selectedProfile(index);
            break;

        case 2:
            setSessionPalette(index);
            emit selectedSession(index);
            break;

        default:
            setStdPalette(index);
            emit selectedStd(index);
            break;

    }
}

void ColourTableButton::setProfilePalette(int index)
{
    if(index>=0 && index<profilePalette.size())
    {
        setCurrentTable(profilePalette[index]);
    }
    else
    {
        setCurrentTable(stdPalette[0]);
    }
}


void ColourTableButton::setSessionPalette(int index)
{
    if(index>=0 && index<importedPalette.size())
    {
        setCurrentTable(importedPalette[index]);
    }
    else
    {
        setCurrentTable(stdPalette[0]);
    }
}

void ColourTableButton::setCurrentTable(const ColourTable& table)
{
    this->currentTable=table;
    setText(currentTable.name);
//     table.updateIcon();
    setIcon(currentTable.icon);
}

ColourTable ColourTableButton::getCurrentTable() const
{
    return currentTable;
}



FontButton::FontButton(QWidget* parent, int family, bool italic, bool bold):
	QPushButton(parent)
{
	this->family=family;
	this->italic=italic;
	this->bold=bold;
	
	menu=new QMenu(this);

#if QT_VERSION>=0x040300
	sansAction=menu->addAction("Sans font", this, SLOT(setSansFont())); sansAction->setCheckable(true); sansAction->setChecked(family==PL_FCI_SANS);
	serifAction=menu->addAction("Serif font", this, SLOT(setSerifFont())); serifAction->setCheckable(true); serifAction->setChecked(family==PL_FCI_SERIF);
	monotypeAction=menu->addAction("Monotype font", this, SLOT(setMonotypeFont())); monotypeAction->setCheckable(true); monotypeAction->setChecked(family==PL_FCI_MONO);
	decorativeAction=menu->addAction("Decorative font", this, SLOT(setDecorativeFont())); decorativeAction->setCheckable(true); decorativeAction->setChecked(family==PL_FCI_SCRIPT);
	
	menu->addSeparator();
	
	italicAction=menu->addAction("Italic", this, SLOT(checkItalic())); italicAction->setCheckable(true); italicAction->setChecked(italic);
	boldAction=menu->addAction("Bold", this, SLOT(checkBold())); boldAction->setCheckable(true); boldAction->setChecked(bold);

    setText("Font");
#else
    normalFontAction=menu->addAction("Normal", this, SLOT(setNormalFont()));
    romanFontAction=menu->addAction("Roman", this, SLOT(setRomanFont()));
    italicFontAction=menu->addAction("Italic", this, SLOT(setItalicFont()));

    setText("Normal font");
#endif
	
	setMenu(menu);

	update();
}

FontButton::~FontButton()
{
	delete menu;
}

void FontButton::getFontParams(PLINT& family, PLINT& style, PLINT& weight)
{
	family=this->family;	
	style=(italic?PL_FCI_ITALIC:PL_FCI_UPRIGHT);
	weight=(bold?PL_FCI_BOLD:PL_FCI_MEDIUM);
}

void FontButton::setFontParams(PLINT family, PLINT style, PLINT weight)
{
	switch(family)
	{
		case PL_FCI_SERIF:
			setSerifFont();
			break;
		case PL_FCI_MONO:
			setMonotypeFont();
			break;
		case PL_FCI_SCRIPT:
			setDecorativeFont();
			break;
		default:
			setSansFont();
			break;
	}
	
	setItalic(style==PL_FCI_ITALIC);
	setBold(weight==PL_FCI_BOLD);
}

void FontButton::update()
{
	QFont f=font();
	
	switch(family)
	{
		case PL_FCI_SERIF:
			f.setStyleHint(QFont::Serif);
			f.setFamily(QSAS_profiles->current()->SerifFont);
			break;
		case PL_FCI_MONO:
			f.setStyleHint(QFont::TypeWriter);
			f.setFamily(QSAS_profiles->current()->MonotypeFont);
			break;
		case PL_FCI_SCRIPT:
			f.setStyleHint(QFont::Decorative);
			f.setFamily(QSAS_profiles->current()->DecorativeFont);
			break;
		default:
			f.setStyleHint(QFont::SansSerif);
			f.setFamily(QSAS_profiles->current()->SansFont);
			break;
	}
	
	f.setBold(bold);
	f.setItalic(italic);
	
	setFont(f);
	this->setFlat(false);

}

void FontButton::setSansFont()
{
	family=PL_FCI_SANS;
#if QT_VERSION >= 0x040300
	sansAction->setChecked(true);
	serifAction->setChecked(false);
	monotypeAction->setChecked(false);
	decorativeAction->setChecked(false);
#endif
	update();
	emit familyChanged(family);
}

void FontButton::setSerifFont()
{
	family=PL_FCI_SERIF;
#if QT_VERSION >= 0x040300
	sansAction->setChecked(false);
	serifAction->setChecked(true);
	monotypeAction->setChecked(false);
	decorativeAction->setChecked(false);
#endif
	update();
	emit familyChanged(family);
}

void FontButton::setMonotypeFont()
{
	family=PL_FCI_MONO;
#if QT_VERSION >= 0x040300
    sansAction->setChecked(false);
    serifAction->setChecked(false);
    monotypeAction->setChecked(true);
    decorativeAction->setChecked(false);
#endif
	update();
	emit familyChanged(family);
}

void FontButton::setDecorativeFont()
{
	family=PL_FCI_SCRIPT;
#if QT_VERSION >= 0x040300
	sansAction->setChecked(false);
	serifAction->setChecked(false);
	monotypeAction->setChecked(false);
	decorativeAction->setChecked(true);
#endif
	update();
	emit familyChanged(family);
}

void FontButton::setItalic(bool i)
{
	italic=i;
#if QT_VERSION >= 0x040300
	italicAction->setChecked(i);
#endif

	update();
	emit styleChanged(i);
}
void FontButton::checkItalic()
{
	setItalic(!italic);
}

void FontButton::setBold(bool i)
{
	bold=i;
#if QT_VERSION >= 0x040300
	boldAction->setChecked(i);
#endif
	update();
	emit weightChanged(i);
}
void FontButton::checkBold()
{
	setBold(!bold);
}

void FontButton::setRomanFont()
{
    setSerifFont();
    setItalic(false);
    setBold(false);
    setText("Roman font");
}

void FontButton::setNormalFont()
{
    setSansFont();
    setItalic(false);
    setBold(false);
    setText("Normal font");
}

void FontButton::setItalicFont()
{
    setSerifFont();
    setItalic(true);
    setBold(false);
    setText("Italic font");
}

#if QT_VERSION < 0x040300
int FontButton::getFontIndex()
{
    if(italic)
        return 3;
    if(family==PL_FCI_SERIF)
        return 2;
    return 1;
}
#endif

double getSpacingMedian(DvObject_var D0)
{	
	int n_data=D0->seqSize();

	if(n_data<3) return -1.;
    
    // compute differences

    QList<double> diff;
	
	if(D0->is_time()){
		for(int i=0; i<n_data-1; ++i)
		{
			//	diff.push_back(fabs(timeseq[i+1].getDiffSecs(timeseq[i])));
			diff.push_back(fabs(D0->time(i+1) - D0->time(i) ));
		}
	}
	else if(D0->is_dbl()){
		for(int i=0; i<n_data-1; ++i)
		{
			//	diff.push_back(fabs(timeseq[i+1].getDiffSecs(timeseq[i])));
			diff.push_back(fabs(D0->dbl(i+1) - D0->dbl(i) ));
		}
	}
	else return -1;
	
    // sort differences

    qSort(diff);

    return diff.at(n_data/2);
}


// Function generating the dash styles
// ----------------------
void setDashStyle(QpLineStyle style, double scale) //QP_NOLINE should be catched before calling this
{
	PLINT * mark;
	PLINT * space;
    
	switch(style)
	{
		case QP_SOLID:
			mark=new PLINT;
			*mark=0;
			space=new PLINT;
			*space=0;
			plstyl(0, mark, space);
			delete mark;
			delete space;
			break;
		
		case QP_DASHED:
			mark=new PLINT;
			*mark=(int)(45000*scale);//1500;

			space=new PLINT;
			*space=(int)(45000*scale);//1500;
			
			plstyl(1, mark, space);
			delete mark;
			delete space;
			break;
		
		case QP_DOTDASH:
			mark=new PLINT[2];
			mark[0]=(int)(6000*scale); 
			mark[1]=(int)(45000*scale);
			space=new PLINT[2];
			space[0]=(int)(30000*scale); 
			space[1]=(int)(30000*scale);
			plstyl(2, mark, space);
			delete[] mark;
			delete[] space;
			break;
		
		case QP_DOTTED:
			mark=new PLINT;
			*mark=(int)(6000*scale);
			space=new PLINT;
			*space=(int)(30000*scale);
			plstyl(1, mark, space);
			delete mark;
			delete space;
			break;
		
		case QP_DASHDOTDOTDOT:
			mark=new PLINT[4];
			mark[0]=(int)(45000*scale); 
			mark[1]=(int)(6000*scale); 
			mark[2]=(int)(6000*scale); 
			mark[3]=(int)(6000*scale);
			space=new PLINT[4];
			space[0]=(int)(30000*scale); 
			space[1]=(int)(30000*scale); 
			space[2]=(int)(30000*scale); 
			space[3]=(int)(30000*scale);
			plstyl(4, mark, space);
			delete[] mark;
			delete[] space;
		break;
		
		default:
			break;
	}
}

PLFLT getCharHeight()
{
	PLFLT def, cur;
	
	plgchr(&def, &cur);
	
	// cur is given in mm. We specified the device as 4pix/mm, for a 32768 (PIXELS_Y) pix height
	
	return cur*4./32768.;
}

int getPGPColourIndex(const QColor& c, bool reset)
{
	PLINT iR,iG,iB;
	static int nextAvailable=16;
	static vector <int> userR(40);
	static vector <int> userG(40);
	static vector <int> userB(40);
	
	if(reset)
	{
		userR.clear();
		userG.clear();
		userB.clear();
		nextAvailable=16;
		for( int i=0;i<16;++i)
		{
			// set up local integer copy of fixed pgplot table
			plgcol0(i,&iR,&iG,&iB);
			userR[i]=iR;
			userG[i]=iG;
			userB[i]=iB;
		}
		return 0;
	}
		
	// find PGPlot colour index closest to this colour
	int R=c.red();
	int G=c.green();
	int B=c.blue();
	int min_diff=200000; // bigger than 255*255*3
	int colIndex=0;
		
	int indexNearest=0;
		
	for( int i=0;i<nextAvailable;i++)
	{
		int diff=(R-userR[i])*(R-userR[i]) + (G-userG[i])*(G-userG[i]) + (B-userB[i])*(B-userB[i]);
		if(diff==0)
		{
			return i; // found in existing table
		}
		else if(diff<min_diff)
		{
			min_diff = diff;
			indexNearest=i;
		}
	}
		
	if( nextAvailable < 40)
	{
		// create new colour and set in local copy
		
		userR[nextAvailable] = R;
		userG[nextAvailable] = G;
		userB[nextAvailable] = B;
		colIndex = nextAvailable;
		nextAvailable++;
	}
	else
	{
		// use nearest colour
		return indexNearest;
	}
	
	// set up colour table as new colour added
	plscmap0n(colIndex+1);
	plscol0(colIndex, R, G, B);
	
	return colIndex;
}

void print_iden() // replaces cpgiden in pgplot
{
	PLFLT xmin, xmax, ymin, ymax;
	PLFLT chrsz1, chrsz2;
	
	plgvpd(&xmin, &xmax, &ymin, &ymax);
	
	plvpor(0., 1., 0., 1.);
	
	plcol0(getPGPColourIndex(QColor(0, 0, 0), false));
	plgchr(&chrsz1, &chrsz2);
	plschr(QPUI_CHAR_HEIGHT, 0.55);
#if QT_VERSION >= 0x040300
	plsfont(PL_FCI_SANS, PL_FCI_UPRIGHT, PL_FCI_MEDIUM);
#else
    plfont(1);
#endif
	
	QString format("dd-MMM-yyyy hh:mm");
	
	QString date=QDateTime::currentDateTime().toString(format);
	QString user(getenv("USER"));
	
	plmtex("b", -0.7, 0.99, 1., (user+QString(" ")+date).toUtf8().data());

	plschr(chrsz1, chrsz2);
	plvpor(xmin, xmax, ymin, ymax);
}

PLINT symbolIndexConversion(int index) // Index in combo box to plplot index
{
	switch(index)
	{
		case 1:
			return (int)QPUI_POINT;
			break;
		case 2:
			return (int)QPUI_PLUS;
			break;
		case 3:
			return (int)QPUI_ASTERISK;
			break;
		case 4:
			return (int)QPUI_CIRCLE;
			break;
		case 5:
			return (int)QPUI_CROSS;
			break;
		case 6:
			return (int)QPUI_SQUARE;
			break;
		case 7:
			return (int)QPUI_TRIANGLE;
			break;
		case 8:
			return (int)QPUI_CIRCLED_POINT;
			break;
		case 9:
			return (int)QPUI_LENS_FLARE;
			break;
		case 10:
			return (int)QPUI_DIAMOND;
			break;
		case 11:
			return (int)QPUI_STAR;
			break;
		case 12:
			return (int)QPUI_WIDECROSS;
			break;
		case 13:
			return (int)QPUI_FILLEDSQUARE;
			break;
		case 14:
			return (int)QPUI_FILLEDCIRCLE;
			break;
		case 15:
			return (int)QPUI_FILLEDSTAR;
			break;
		case 16:
			return (int)QPUI_LEFT_ARROW;
			break;
		case 17:
			return (int)QPUI_RIGHT_ARROW;
			break;
		case 18:
			return (int)QPUI_UP_ARROW;
			break;
		case 19:
			return (int)QPUI_DOWN_ARROW;
			break;
		case 20:
			return (int)QPUI_CUSTOM_SHAPE;
			break;
		case 0:
		default:
			return (int)QPUI_NOSYMBOL;
			break;
	}
}


string SafePGPString(const string& str)
{
	// return "safe" string that can be correctly displayed by pgplot
	// Also ensures it is not too long as pgplot can crash with long strings
	
	if(str=="") return str;
	
	if(QuSlotDebug())
	{
		printf("SafePGPString %s\n", str.c_str());
		fflush(stdout);
	}

	string safe_str(str);
	size_t pos;
// 	while((pos=safe_str.find('\136'))!=string::npos) safe_str[pos]='\036';
 
	if(safe_str.length() > 178){
		safe_str.replace(175, safe_str.length() -175, "");
		safe_str += "...";
	}
	
	return safe_str;
}


